function [gm f filters] = ambimodgains(order,orderlim,fc,transwidth,nf,fmax,type,scaling,nlsp,filterlength)

%function [gm f filters] = ambimodgains(order,orderlim,fc,transwidth,nf,fmax,type,scaling,nlsp[,filterlength])
%
% Calculates frequency domain shelving filters for modified 
% high-frequency Ambisonics decoding. Can also design corresponding
% odd order windowed FIR filters. (Only 2D for now.)
%
% Parameters:
%   'order'........ Ambisonics order
%   'orderlim'..... force basic decoding up to 'orderlim'
%   'fc'........... center frequency for shelving filters.
%                   can be a two element vector
%   'transwidth'... transition width [octaves]
%   'nf'........... number of frequency components
%   'fmax'......... highest simulation frequency [Hz] 
%                   (should be equal to fs/2 for 'nf' even FIR filter design)
%   'type'......... 'maxre' or 'inphase'.
%                   can be a two element cell array, related to 'fc'.
%   'scaling'...... 'pressure' or 'energy' are valid values
%   'nlsp'......... number of loudspeakers (assumes regular 2D layout)
%   'filterlength'. length of each FIR filter [samples]
%
% The calculated frequency domain shelving filters are returned in the 
% 'nf'-by-'order'+1 matrix 'gm', with 'f' holding the corresponding 
% frequncies. If 'filterlength' is given, the corresponding FIR filters
% is returned in the 'filterlength'-by-'order'+1 matrix 'filters'.
%
% Note: 'order' can also be specified as distict modes using a
%       vector of length 'nmodes'.
%
% References:
% J. Daniel, Reprsentation de Champs Acoustiques, Application  la Transmission et  la 
%   Reproduction de Scnes Sonores Complexes dans un Contexte Multimdia, Ph.D. Thesis, 
%   University of Paris 6, Paris, France, 2000, Section 3.3.2. 
%   (http://gyronymo.free.fr/audio3D/download_Thesis_PwPt.html)
%
% - Brd Stfringsdal 2006 -

% set modes
if length(order) > 1
  % flag
  expmodes = 1;
  % if modes given as input parameter
  % fix orientation
  order = order(:).';
  % select distinctive modes
  [m throw modesfullindex] = unique(order);
  % set equivalent order
  order = max(m);
else
  % flag
  expmodes = 0;
  % set modes
  m = 0:order;
end

% linear frequency range
fnorm = (0:nf-1).'/(nf-1);
f = fnorm*fmax;

% modal gains
gm = ones(nf,order+1);

% run through for all transition bands
for ii = 1:length(fc)
  % modal gains, affected frequencies
  if iscell(type)
    thistype = type{ii}
  else
    thistype = type
  end
  switch thistype
    case 'maxre'
      gmaf = maxre(order,orderlim);
    case 'inphase'
      m = 0:order;
      gmaf = factorial(order)^2./(factorial(order+m).*factorial(order-m)); 
  end

  % scaling
  switch scaling
    case 'energy'
      scalefac = sqrt(nlsp/(gmaf(1).^2+2*sum(gmaf(2:end).^2)));
    case 'pressure'
      scalefac = 1/gmaf(1);
  end

  gmaf = scalefac*gmaf;

  % transition limits
  fmin = fc(ii)*2^(-transwidth/2);
  fmax = fc(ii)*2^(transwidth/2);
  % find entries
  fminindex = min(find(f >= fmin));
  fmaxindex = max(find(f <= fmax));
  select = (1:(fmaxindex-fminindex+1)).';

  % expand gmaf to all affected frequencies
  nfaffected = nf-(fminindex-1);
  gmaf = gmaf(ones(nfaffected,1),:);
  % cosine transition
  if fmin == fmax
    % set to interpolated value
    transfunc = (gmaf+1)/2;
  else
    transfunc = (cos(select/length(select)*pi)+1)/2;
    % expand
    transfunc = transfunc(:,ones(order+1,1));
  end
  gmaf(select,:) = gm(select+fminindex-1,:).*transfunc + gmaf(select,:).*(1-transfunc);
  gm(fminindex:end,:) = gmaf;
end

% create fir filters
if nargin > 9
  filterorder = filterlength - 1;
  % fix filterlength
  filterorder = 2*floor(filterorder/2);
  filters = zeros(filterlength,order+1);
  for ii = 1:order+1
    % use window method (hamming window)
    filters(1:filterorder+1,ii) = fir2(filterorder,fnorm,gm(:,ii));
  end
end
    
% expand to match input modes
if expmodes
  gm = gm(:,modesfullindex);
end
