/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _ambisonics_h
#define _ambisonics_h

// GSL will look for this.
#define HAVE_INLINE

#include <gsl/gsl_linalg.h>
#include "Defs.hpp"
#include "Matrix.hpp"
#include "PositionVector.hpp"
#include <cmath>

/**
 * \brief Class for common Ambisonics functions.
 *
 * Includes common utility functions implemented as static methods.
 */
class Ambisonics {
public:
	enum Normalization { SemiNormalized = 0, Normalized, N2DCompatible,
	                     FurseMalham };
private:
	std::vector<double> _pm, _pm1, _pm2; // Temp arrays for EncodingCoefs
	std::vector<double> _beta;
	unsigned _max_order;
	Normalization _norm;
	unsigned _order;
	bool _is3d;
	double _regparm;

	static double fact(unsigned n)
	{
		double t = 1.;

		while (n > 1) {
			t *= double(n);
			--n;
		}
		return t;
	}

	/**
	 * Calculate 3D Ambionics normalising coefficients for the current order
	 * and store them in an internal table.
	 */
	void UpdateBetaTable()
	{
		// _max_order is the next order we need to precalc
		if (_max_order > _order) // We don't let the table shrink
			return;

		_beta.resize((_order + 2)*(_order + 1)/2);
		unsigned pos = (_max_order + 1)*_max_order/2;

		for (unsigned m = _max_order; m <= _order; ++m) {
			for (unsigned n = 0; n <= m; ++n)
				_beta[pos++] = Beta(m, n);
		}
		_max_order = _order + 1;
	}
public:
	/**
	 * Class constructor.
	 * @param order Ambisonics order.
	 * @param is3d boolean describing if Ambisonics is three dimensional or not.
	 */
	Ambisonics(unsigned order, bool is3d) : _max_order(0), _norm(Normalized),
	                                        _order(order), _is3d(is3d),
	                                        _regparm(1./1000.)
	{
		UpdateBetaTable();
	}

	/**
	 * Sets what Ambisonics order to use.
	 * @param order Ambisonics order.
	 */
	void SetOrder(unsigned order)
	{
		_order = order;
		_pm.resize(order + 1);
		_pm1.resize(order + 1);
		_pm2.resize(order + 1);
		UpdateBetaTable();
	}

	/**
	 * Sets dimensionality of Ambisonics.
	 * @param is3d specifies whether 2D or 3D Ambisonics is to be used.
	 */
	void SetIs3D(bool is3d)
	{
		_is3d = is3d;
	}

	/**
	 * Calculate the number of total Ambisonics components for a given setup.
	 * @param order order of Ambisonics used.
	 * @param is3d specifies whether 2D or 3D Ambisonics is in use.
	 * @return The number of components.
	 */
	static unsigned NumComponents(unsigned order, bool is3d)
	{
		if (!is3d)
			return 2*order + 1;
		else
			return (order + 1)*(order + 1);
	}

	/**
	 * Calculate the number of total Ambisonics components for current setup.
	 * @return The number of components.
	 */
	unsigned NumComponents()
	{
		return NumComponents(_order, _is3d);
	}

	/**
	 * Calculate the number of Ambisonics components the given order
	 * contributes.
	 * @param order Ambisonics order for which one wants to find number of
	 * corresponding components.
	 * @param is3d specifies whether 2D or 3D Ambisonics is in use.
	 * @return Number of new components for given order.
	 */
	static unsigned OrderComponents(unsigned order, bool is3d)
	{
		if (!is3d)
			// Zeroeth order contributes one component, all others two
			return order != 0 ? 2 : 1;
		else
			return 2*order + 1;
	}

	/**
	 * Calculate the number of Ambisonics components the given order
	 * contributes.
	 * @param order Ambisonics order for which one wants to find number of
	 * corresponding components.
	 * @return Number of new components for given order.
	 */
	unsigned OrderComponents(unsigned order)
	{
		return OrderComponents(order, _is3d);
	}

	/**
	 * Calculate N3D compliant scaling coefficients \f$\beta_{m, n}\f$.
	 * @param m degree of Legendre function.
	 * @param n order of Legendre function.
	 */
	double Beta(unsigned m, unsigned n)
	{
		// Table for SN3D -> Furse-Malham conversion
		const double fmh_conv[] = {
		    // 1/sqrt(2.),
		    0.707106781,                              // m = 0
		    // 1., 1.,
		    1., 1.,                                   // m = 1
		    // 1., 2./sqrt(3.), 2./sqrt(3.),
		    1., 1.154700538, 1.154700538,             // m = 2
		    // 1., sqrt(45.)/32., 3./sqrt(5.), sqrt(8.)/5.
		    1., 0.209631373, 1.341640786, 0.565685425 // m = 3
		};
		double t = n == 0 ? 1. : 2.;
		double r = 0.;

		switch (_norm) {
		case SemiNormalized:
		case FurseMalham:
			r = sqrt(t*fact(m - n)/fact(m + n));
			break;
		case Normalized:
			r = sqrt((2.*m + 1.)*t*fact(m - n)/fact(m + n));
			break;
		case N2DCompatible:
			r = sqrt(pow(2., 2.*m)*fact(m)*fact(m)/(fact(2*m + 1)))
			    *sqrt((2.*m + 1.)*t*fact(m - n)/fact(m + n));
			break;
		}
		// Furse-Malham coefs just use semi-normalised coefs as a base,
		// multiplied with a modifier.
		if (_norm == FurseMalham && m <= 3)
			r *= fmh_conv[m*(m + 1)/2 + n];
		return r;
	}

	// TODO remove
	void DumpBetaTable()
	{
		std::vector<double>::iterator i = _beta.begin();

		for (unsigned m = 0; m < _max_order; ++m) {
			for (unsigned n = 0; n <= m; ++n)
				std::cout << *i++ << ", ";
			std::cout << std::endl;
		}
	}

	/**
	 * Change normalization method.
	 * @param norm normalization method to be used.
	 */
	void SetNormalization(Normalization norm)
	{
		_norm = norm;
		// Rebuild beta table with new normalization method
		_max_order = 0;
		UpdateBetaTable();
	}

	/**
	 * Calculate Ambisonics encoding coefficients. This method can also be
	 * used for calculating decoding coefficients for a regular speaker setup.
	 * Note that the matrix is transposed as opposed to the ordinary
	 * definition of an Ambisonics encoding matrix.
	 * This variation of the method only updates one row of the matrix.
	 * @param scalefactor constant by which all coefficients will be scaled.
	 * @param azimuth azimuth of source/speaker, in radians.
	 * @param elev elevation of source/speaker, in radians.
	 * @param mt a matrix in which to store the resulting coefficients.
	 * @param row which row of the matrix to update.
	 */
	template<class T, class F, class A>
	void EncodingCoefs(double scalefactor, double azimuth, double elev,
	                   matrix<T, F, A>& mt, size_t row)
	{
		// These recurrences are the same as found in Daniel's NFC article, but
		// modified to use indices in the range n to n - 2 and m to m - 2,
		// instead of n + 1 to n - 1 and m + 1 to m - 1.
		if (!_is3d) {
			double scale;
			size_t col = 0;
			// State variables for quick cos()/sin() calculation
			double c1 = cos(azimuth), s1 = sin(azimuth);
			double cn1 = 1., cn2 = c1, sn1 = 0.;

			if (_norm == Normalized)
				scale = scalefactor*sqrt(2.);
			else
				scale = scalefactor;

			// Multiply with beta(0, 0) in case it's not 1.0 (FMH)
			mt(row, col++) = scalefactor*_beta[0]; // y_{0, 0}^{+1}, always 1.0
			for (size_t i = 1; i <= _order; ++i) {
				// cn is cos(m*azimuth)
				const double cn = 2.*c1*cn1 - cn2;
				// sn is sin(m*azimuth)
				const double sn = c1*sn1 + s1*cn1;
				cn2 = cn1;
				cn1 = cn;
				sn1 = sn;

				// mt(row, col++) = scale*cos(m*p[row].azimuth);
				mt(row, col++) = scale*cn;
				// mt(row, col++) = scale*sin(m*p[row].azimuth);
				mt(row, col++) = scale*sn;
			}
		} else {
			double c1 = cos(azimuth), s1 = sin(azimuth);
			double uz = sin(elev), r = cos(elev);
			std::vector<double>::iterator beta = _beta.begin();

			// State variables/arrays.
			double cn1 = 1., cn2 = c1, sn1 = 0.;
			
			// Compute a table for the sin/cos(n*azimuth) recursion computation,
			// which would otherwise have to be computed over and over again.
			double trigtab[_order*2];
			for (unsigned i = 0; i < _order; ++i) {
				// cn is cos(m*azimuth)
				const double cn = 2.*c1*cn1 - cn2;
				// sn is sin(m*azimuth)
				const double sn = c1*sn1 + s1*cn1;
				
				cn2 = cn1;
				cn1 = cn;
				sn1 = sn;
				trigtab[2*i] = cn;
				trigtab[2*i + 1] = sn;
			}

			std::fill_n(_pm.begin(), _order + 1, 0.);
			std::fill_n(_pm1.begin(), _order + 1, 0.);
			std::fill_n(_pm2.begin(), _order + 1, 0.);

			// Compute omnidirectional zeroeth order component.
			_pm1[0] = 1.; // p_{0, 0} is always 1.0 
			mt(row, 0) = scalefactor*(*beta++); // y_{0, 0}^{+1}, always 1.0
		
			for (unsigned m = 1; m <= _order; ++m) {
				// Index to first coef of current order.
				unsigned col = m*m;
				// Compute p_{m, 0} special case term.
				double pm_0 = ((2*m - 1)*uz*_pm1[0] - (m - 1)*_pm2[0])/double(m);

				if (m == 1)
					// Hack for placing Z component at column 3
					col = 3;

				mt(row, col++) = scalefactor*(*beta++)*pm_0; // y_{m, 0}^{1}
				_pm[0] = pm_0;

				// Hack for placing X and Y components at columns 1 and 2
				if (m == 1)
					col = 1;
				unsigned tabpos = 0; // Index into trigtab.
				// Now compute the other p_{m, n} terms.
				for (unsigned n = 1; n <= m; ++n) {
					double pmn = _pm2[n] + (2.*m - 1.)*r*_pm1[n - 1];
					double scale = scalefactor*(*beta++)*pmn;

					_pm[n] = pmn;
					mt(row, col++) = scale*trigtab[tabpos++]; // y_{m, n}^{1}
					mt(row, col++) = scale*trigtab[tabpos++]; // y_{m, n}^{-1}
				}
				// Swap state arrays, so that p_{m - 2} array gets p_{m - 1}
				// array values, p_{m - 1} gets p_m values, and p_m reuses
				// p_{m - 2} array storage space.
				_pm2.swap(_pm1);
				_pm1.swap(_pm);
			}
		}
	}

	/**
	 * Calculate Ambisonics encoding coefficients. This method can also be
	 * used for calculating decoding coefficients for a regular speaker setup.
	 * Note that the matrix is transposed as opposed to the ordinary
	 * definition of an Ambisonics encoding matrix.
	 * This variation of the method only updates the entire matrix.
	 * @param scalefactor constant by which all coefficients will be scaled.
	 * @param p a PositionVector containing either source or loudspeaker
	 * coordinates.
	 * @param m a matrix in which to store the resulting coefficients.
	 */
	template<class T, class F, class A>
	void EncodingCoefs(double scalefactor, PositionVector& p,
	                          matrix<T, F, A>& m)
	{
		for (size_t row = 0; row < p.size(); ++row)
			EncodingCoefs(scalefactor, p[row].azimuth, p[row].elevation, m, row);
	}

	/**
	 * Sets the regularisation parameter used for regularising the decoding
	 * matrix produced by DecodingCoefs.
	 * @param regparm regularisation parameter. A value of 0 gives no
	 * regularisation.
	 */
	void SetRegularisationParam(double regparm)
	{
		_regparm = regparm;
	}

	/**
	 * Calculate coefficients for an ambisonics decoder without
	 * assumptions about the loudspeaker array setup. This involves some costly
	 * matrix operations.
	 * @param scalefactor constant by which all coefficients will be scaled.
	 * @param p a PositionVector containing either source or loudspeaker
	 * coordinates.
	 * @param m a matrix in which to store the resulting coefficients.
	 * @return Boolean value signifying whether the matrix could be correctly
	 * calculated. This might be false in cases where the matrix inversion could
	 * not be performed.
	 */
	template<class T, class F, class A>
	bool DecodingCoefs(double scalefactor, PositionVector& p,
	                   matrix<T, F, A>& m)
	{
		DoubleMatrix me_t(m);
		bool transposed = false;

#if 0		
		fprintf(stderr, "azimuths:\n");
		for (size_t i = 0; i < p.size(); ++i)
			fprintf(stderr, "%f, ", 180./M_PI*p[i].azimuth);
		fprintf(stderr, "\nelevations:\n");
		for (size_t i = 0; i < p.size(); ++i)
			fprintf(stderr, "%f, ", 180./M_PI*p[i].elevation);
#endif

		// Get the transposed encoding matrix
		EncodingCoefs(1.0/scalefactor, p, me_t);
		// Now, let's compute a Tikhonov regularised version of the pseudo-
		// inverse of this matrix.
		// SVD routines are only defined for size1() >= size2(). If this is not
		// true, we transpose the matrix and SVD that. U and V will swap place,
		// which we will fix later.
		if (me_t.size2() > me_t.size1()) {
			me_t = trans(me_t);
			transposed = true;
		}
		// We allocate the small memory amount we need for the SVD sigma array
		// on the stack here instead of using the slower gsl_vector_alloc.
		double sigma_diag[me_t.size2()];
		double work[me_t.size2()];
		DoubleMatrix V(me_t.size2(), me_t.size2());
		// Set up GSL proxies for our Matrix data
		gsl_matrix_view Up = gsl_matrix_view_array(&me_t(0, 0),
		                                           me_t.size1(), me_t.size2());
		gsl_matrix_view Vp = gsl_matrix_view_array(&V(0, 0),
		                                           V.size1(), V.size2());
		gsl_vector_view S = gsl_vector_view_array(sigma_diag, me_t.size2());
		gsl_vector_view w = gsl_vector_view_array(work, me_t.size2());

		// First do the singular value decomposition
		if (gsl_linalg_SV_decomp(&Up.matrix, &Vp.matrix, &S.vector, &w.vector))
			return false;
		if (transposed)
			swap(me_t, V);

		const double max_sigma = sigma_diag[0];
		// Regularisation parameter
		const double beta = _regparm*max_sigma*max_sigma;
		// TODO perhaps we need to check for some bad sigma matrices here?
		for (size_t i = 0; i < me_t.size2(); ++i) {
			double x = sigma_diag[i];

			// Prevent divide by zero
			if (x > 1e-20)
				sigma_diag[i] = x*x/(x*x + beta)/x;
			else
				sigma_diag[i] = 0.; // Zero limit of above expression is 0
		}
		// First multiply matrix U by the diagonal matrix S. We do this by hand
		// since no BLAS routines for this setup seems to exist.
		for (size_t i = 0; i < me_t.size1(); ++i)
			for (size_t j = 0; j < me_t.size2(); ++j)
				me_t(i, j) *= sigma_diag[j];
		// Then multiply the result of previous operation with V transposed.
		// The result of this is the Tikhonov-regularised pseudo-inverse.
		noalias(m) = prod(me_t, trans(V));

#if 0		
		fprintf(stderr, "\ntikhonov-regularisert:\n[");
		for (size_t i = 0; i < m.size1(); ++i) {
			for (size_t j = 0; j < m.size2(); ++j)
				fprintf(stderr, "%f%s", m(i, j), j == m.size2() - 1 ? ";\n" : ", ");
		}

		fprintf(stderr, "]\n");
#endif
		return true;
	}

	/**
	 * Fill a vector with encoding coefficients for a given order and component.
	 * @param coefs reference to a vector which is to contain the encoding
	 * coefficients.
	 * @param azimuth reference to a vector containing the position azimuths
	 * for each corresponding coefficient to be generated.
	 * @param elevs reference to a vector containing the position elevations
	 * for each corresponding coefficient to be generated.
	 * @param order order for which to generate coefficients. Must be higher
	 * than 0 and lower than or equal to 5.
	 * @param component component number within current order for which to
	 * generate coefficients. Range is 0 <= component < order.
	 * @param num number of coefficients to generate.
	 */
	void EncodeComponent(FloatVector& coefs, FloatVector& azimuths,
	                     FloatVector& elevs, unsigned order, unsigned component,
	                     unsigned num)
	{
		// TODO: obviously, all the direct use of cos/sin in here isn't optimal
		// at all. This is a very nice place to use a vectorised sin/cos
		// approximation, this is all very parallelizable code.
		if (!_is3d) {
			float mul = order;
			float scale;

			if (_norm == Normalized)
				scale = 1.414213562f; // sqrtf(2.f);
			else
				scale = 1.f;

			if (component == 0) {
				for (unsigned i = 0; i < num; ++i)
					coefs[i] = scale*cosf(mul*azimuths[i]);
			} else if (component == 1) {
				for (unsigned i = 0; i < num; ++i)
					coefs[i] = scale*sinf(mul*azimuths[i]);
			}
		} else {
			unsigned n;
			bool sigma; // Positive sigma?
			unsigned i = 0;

			if (order > 5)
				return;
			// For order = 1, n = 0 component (Z) comes last. For all other
			// orders, n components come in ascending order.
			if (order == 1) {
				n = (3 - component)/2;
				sigma = component != 1;
			} else {
				n = (component + 1)/2;
				sigma = component & 1;
			}
			// I'd rather have a nicer method of calculating coefs in 3D without
			// having to hard code the functions, but I can't currently think of
			// anything that'd be fast enough while interpolating positions
			// per-sample. The following functions are really just associated
			// Legendre polynomials of order m and degree n (P_n^m(sin(\phi)).
			switch (order) {
			// 1st order components
			case 1:
				switch (n) {
				case 0: // Z
					for (; i < num; ++i)
						coefs[i] = _beta[1]*sinf(elevs[i]);
					break;
				case 1: // X & Y
					for (; i < num; ++i)
						coefs[i] = _beta[2]*cosf(elevs[i]);
					break;
				}
				break;

			// 2nd order components
			case 2:
				switch (n) {
				case 0:
					for (; i < num; ++i) {
						const float sn = sinf(elevs[i]);
						coefs[i] = _beta[3]*(1.5f*sn*sn - 0.5f);
					}
					break;
				case 1:
					for (; i < num; ++i) {
						const float sn = sinf(elevs[i]);
						const float cs = cosf(elevs[i]);
						coefs[i] = _beta[4]*3.f*sn*cs;
					}
					break;
				case 2:
					for (; i < num; ++i) {
						const float cs = cosf(elevs[i]);
						coefs[i] = _beta[5]*3.f*cs*cs;
					}
					break;
				}
				break;

			// 3rd order components
			case 3:
				switch (n) {
				case 0:
					for (; i < num; ++i) {
						const float sn = sinf(elevs[i]);
						coefs[i] = _beta[6]*(2.5f*sn*sn - 1.5f)*sn;
					}
					break;
				case 1:
					for (; i < num; ++i) {
						const float cs = cosf(elevs[i]);
						coefs[i] = -_beta[7]*(7.5f*cs*cs - 6.f)*cs;
					}
					break;
				case 2:
					for (; i < num; ++i) {
						const float sn = sinf(elevs[i]);
						const float cs = cosf(elevs[i]);
						coefs[i] = _beta[8]*15.f*sn*cs*cs;
					}
					break;
				case 3:
					for (; i < num; ++i) {
						const float cs = cosf(elevs[i]);
						coefs[i] = _beta[9]*15.f*cs*cs*cs;
					}
					break;
				}
				break;

			// 4th order components
			case 4:
				switch (n) {
				case 0:
					for (; i < num; ++i) {
						const float sn = sinf(elevs[i]);
						const float sn2 = sn*sn;
						coefs[i] = _beta[10]*(4.375f*sn2*sn2 - 3.75f*sn2
						                      + 0.375f);
					}
					break;
				case 1:
					for (; i < num; ++i) {
						const float sn = sinf(elevs[i]);
						const float cs = cosf(elevs[i]);
						coefs[i] = -_beta[11]*(17.5f*cs*cs - 10.f)*cs*sn;
					}
					break;
				case 2:
					for (; i < num; ++i) {
						const float cs = cosf(elevs[i]);
						coefs[i] = -_beta[12]*(52.5f*cs*cs - 45.f)*cs*cs;
					}
					break;
				case 3:
					for (; i < num; ++i) {
						const float sn = sinf(elevs[i]);
						const float cs = cosf(elevs[i]);
						coefs[i] = _beta[13]*105.f*sn*cs*cs*cs;
					}
					break;
				case 4:
					for (; i < num; ++i) {
						const float cs = cosf(elevs[i]);
						const float cs2 = cs*cs;
						coefs[i] = _beta[14]*105.f*cs2*cs2;
					}
					break;
				}
				break;

			// 5th order components
			case 5:
				switch (n) {
				case 0:
					for (; i < num; ++i) {
						const float sn = sinf(elevs[i]);
						const float sn2 = sn*sn;
						coefs[i] = _beta[15]*(7.875f*sn2*sn2 - 8.75f*sn2
						                      + 1.875f)*sn;
					}
					break;
				case 1:
					for (; i < num; ++i) {
						const float cs = cosf(elevs[i]);
						const float cs2 = cs*cs;
						coefs[i] = _beta[16]*(39.375f*cs2*cs2 - 52.5f*cs2
						                      + 15.f)*cs;
					}
					break;
				case 2:
					for (; i < num; ++i) {
						const float sn = sinf(elevs[i]);
						const float cs = cosf(elevs[i]);
						const float cs2 = cs*cs;
						coefs[i] = -_beta[17]*(157.5f*cs2 - 105.f)*cs2*sn;
					}
					break;
				case 3:
					for (; i < num; ++i) {
						const float cs = cosf(elevs[i]);
						const float cs2 = cs*cs;
						coefs[i] = -_beta[18]*(472.5f*cs2 - 420.f)*cs2*cs;
					}
					break;
				case 4:
					for (; i < num; ++i) {
						const float sn = sinf(elevs[i]);
						const float cs = cosf(elevs[i]);
						const float cs2 = cs*cs;
						coefs[i] = _beta[19]*945.f*cs2*cs2*sn;
					}
					break;
				case 5:
					for (; i < num; ++i) {
						const float cs = cosf(elevs[i]);
						const float cs2 = cs*cs;
						coefs[i] = _beta[20]*945.f*cs2*cs2*cs;
					}
					break;
				}
				break;
			}

			if (n > 0) {
				if (sigma) {
					for (i = 0; i < num; ++i)
						coefs[i] *= cosf(n*azimuths[i]);
				} else {
					for (i = 0; i < num; ++i)
						coefs[i] *= sinf(n*azimuths[i]);
				}
			}
		}
	}
};

#endif

