/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "AmbisonicsDecoder.hpp"

#ifdef __SSE__
#include <xmmintrin.h>
#endif

AmbisonicsDecoder::AmbisonicsDecoder()
	: _convolver(512), _order(1), _is3d(false), _gain(1.f),
	  _input_filters_enabled(false), _ambi(_order, _is3d)
{
#ifdef __SSE__
	// If there's a chance we'll be using SSE for calculations, let's disable
	// denormal processing.
	_MM_SET_FLUSH_ZERO_MODE(_MM_FLUSH_ZERO_ON);
#endif
	_math = SIMDMath::Create();
	_pos.resize(1);
}

void AmbisonicsDecoder::InitMatrix()
{
	size_t num;

	num = _ambi.NumComponents();
	_matrix.resize(_pos.size(), num);
	_input_filters_enabled = false;
	UpdateAllSpeakers();  // Fill decoding matrix with valid values again
}

void AmbisonicsDecoder::SetNumSpeakers(size_t num)
{
	_pos.resize(num);
	InitMatrix();
}

void AmbisonicsDecoder::SetIs3D(bool is3d)
{
	_is3d = is3d;
	_ambi.SetIs3D(is3d);
	InitMatrix();
}

void AmbisonicsDecoder::SetOrder(unsigned order)
{
	_order = order;
	_ambi.SetOrder(order);
	InitMatrix();
}

void AmbisonicsDecoder::SetGain(float gain)
{
	// Set slide time to 100 ms and target gain to one that combined with
	// current matrix gain will yield the wanted final gain
	_outputgainer.SetGain(gain/_matrixgain, _sample_rate/10);
	_gain = gain;
}

void AmbisonicsDecoder::UpdateSpeaker(size_t spk)
{
	// The matrix _must_ be correctly dimensioned by this point.
	std::cout << "Updating coefs for speaker " << spk << std::endl;
	// No way to update only one speaker's set of coefs this way, so
	// calculate the entire matrix again
	_ambi.DecodingCoefs(_gain, _pos, _matrix);
}

void AmbisonicsDecoder::UpdateAllSpeakers()
{
	_ambi.DecodingCoefs(_gain, _pos, _matrix);
	_matrixgain = _gain;
	// Abort any volume fade in progress (though it's usually finished anyway).
	// Also make sure it starts at 1. gain next time.
	_outputgainer.SetGain(1.f, 0);
}

void AmbisonicsDecoder::SetFilters(std::vector<FloatVector>& filters)
{
	// Try for a default block size of half the filter length or at most
	// the current buffer size. This ensures that we'll have no filtering
	// latency thanks to the block processing and should also work well for
	// the smaller filters anyway.
	const unsigned N = unsigned(pow(2., ceil(log2(double(filters[0].size()/2))))); 
	const unsigned block_size = std::min(N, _buffer_size);

	_convolver = FFTConvolver(block_size);
	_firinput = InputBuffer(GetNumInputs(), block_size);
	_firoutput = OutputBuffer(GetNumInputs(), block_size);
	// Adaptor for use in Process()
	_firoutputptrs.resize(GetNumInputs());
	for (size_t i = 0; i < GetNumInputs(); ++i)
		_firoutputptrs[i] = &_firoutput[i][0];

	_filters.clear();
	for (size_t i = 0; i < filters.size(); ++i)
		_filters.push_back(PartitionedFilter(block_size, filters[i]));
	// All filters should share same size
	_firinput.ResetHistory(_filters[0]);
	_input_filters_enabled = true;
}

bool AmbisonicsDecoder::InputFiltersEnabled() const
{
	return _input_filters_enabled;
}

void AmbisonicsDecoder::SetNormalization(Ambisonics::Normalization norm)
{
	_ambi.SetNormalization(norm);
	UpdateAllSpeakers();
}

void AmbisonicsDecoder::SetRegularisationParam(double regparm)
{
	_ambi.SetRegularisationParam(regparm);
	UpdateAllSpeakers();
}

void AmbisonicsDecoder::SetAzimuth(size_t spk, double azimuth)
{
	_pos[spk].azimuth = azimuth;
	UpdateAllSpeakers();
}

void AmbisonicsDecoder::SetElevation(size_t spk, double elevation)
{
	_pos[spk].elevation = elevation;
	UpdateAllSpeakers();
}

void AmbisonicsDecoder::SetSampleRate(unsigned sr)
{
	AudioProcessor::SetSampleRate(sr);
//	_input_filters_enabled = false;
}

void AmbisonicsDecoder::SetBufferSize(unsigned size)
{
	AudioProcessor::SetBufferSize(size);
	_gainramp.resize(size);
	_input_filters_enabled = false;
	// TODO: recalc filters here in case buffer size became too big
}

void AmbisonicsDecoder::Process(float** inputs, float** outputs, unsigned num)
{
	if (_input_filters_enabled) {
		unsigned samples = num;
		
		// The use of the OutputBuffer class is somewhat unnecessary as is
		// right now, but will come in handy if we ever lose the 2^x buffer
		// size restriction.
		while (samples) {
			unsigned offset = num - samples;

			samples -= _firinput.Input(inputs, samples, offset);
			// We have a full input buffer of data, let's do some filtering
			if (_firinput.Full()) {
				unsigned comp = 0;

				_convolver.Input(_firinput);
				for (unsigned order = 0; order <= _order; ++order) {
					// Each order has one filter
					_convolver.SetFilter(_filters[order]);
					// Filter all the components of current order
					while (comp < _ambi.NumComponents(order, _is3d))
						_convolver.Work(_firinput, _firoutput, comp++);
				}
				// Now let's mix the filtered chunk
				_math->MatrixMix(&_firoutputptrs[0], outputs, _matrix,
				                 _firoutput.bufsize(), offset);
				_firinput.Overlap();
			}
		}
	} else {
		// Input filters are not enabled, so just a matrix mix will do
		_math->MatrixMix(inputs, outputs, _matrix, num);
	}
	// To properly fade the volume without clicks, we ramp the volume linearly
	// from one amplitude to another, then after the ramp is done, recalculate
	// the mixing matrix to use the new target gain.
	if (_outputgainer.Interpolating()) {
		// Create the volume ramp
		_outputgainer.MakeGainVector(&_gainramp[0], num);
		// Apply the ramp to outputs
		for (size_t i = 0; i < GetNumOutputs(); ++i)
			_math->VVMul(outputs[i], outputs[i], &_gainramp[0], num);
		// Check if fade is done
		if (!_outputgainer.Interpolating())
			UpdateAllSpeakers();
	}
}

