/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <cmath>
#include <gtkmm/aboutdialog.h>
#include <gtkmm/box.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/table.h>
#include <gtkmm/toggleaction.h>
#include "AmbisonicsDecoderGUI.hpp"
#include "DecoderSettingsWindow.hpp"
#include "MeterWindow.hpp"
#include "Widgets.hpp"

// XML formatted string describing the menu layout
const Glib::ustring AmbisonicsDecoderGUI::_menu_info =
	"<ui>"
	"  <menubar name='MenuBar'>"
	"    <menu action='FileMenu'>"
	"      <menuitem action='FileOpen'/>"
	"      <menuitem action='FileSave'/>"
	"      <menuitem action='FileLoadFilters'/>"
	"      <separator/>"
	"      <menuitem action='FileQuit'/>"
	"    </menu>"
	"    <menu action='ViewMenu'>"
	"      <menuitem action='ViewLoudspeakerLayout'/>"
	"      <menuitem action='ViewSettings'/>"
	"      <menuitem action='ViewLevelMeters'/>"
	"    </menu>"
	"    <menu action='RoutingMenu'>"
	"      <menuitem action='RoutingInputs'/>"
	"      <menuitem action='RoutingOutputs'/>"
	"    </menu>"
	"    <menu action='HelpMenu'>"
	"      <menuitem action='HelpAbout'/>"
	"    </menu>"
	"  </menubar>"
	"</ui>";

AmbisonicsDecoderGUI::AmbisonicsDecoderGUI(AmbisonicsDecoderModel& dec,
                                           JACKInterface& snd) 
	: _adj_numspeakers(1., 1., 32., 1., 10.),
	  _adj_order(1., 0., 15., 1., 3.), _btn_numspeakers(_adj_numspeakers, 0., 0),
	  _btn_order(_adj_order, 0., 0), _btn_3d("3D"), _sld_volume(-60., 12.),
	  _decoder(dec), _wnd_speakers(dec), _wnd_settings(dec), _wnd_meters(snd),
	  _snd(snd)
{
	// Init window
	set_title("Ambisonics decoder");
	set_default_size(350, 200);
	// Make containers
	Gtk::Table* table = Gtk::manage(new Gtk::Table(3, 3));
	Gtk::VBox* vbox1 = Gtk::manage(new Gtk::VBox());
	// Start constructing our GUI. This will consist of a VBox containing a
	// MenuBar, a table containing other widgets, and a StatusBar.
	add(*vbox1);

	// Construct our menu. This corresponds to the XML string at the top.
	_actiongroup = Gtk::ActionGroup::create();
	// File menu
	_actiongroup->add(Gtk::Action::create("FileMenu", "_File"));
	_actiongroup->add(Gtk::Action::create(
		"FileOpen", Gtk::Stock::OPEN, "_Open", "Open a decoder setup"),
		sigc::mem_fun(*this, &AmbisonicsDecoderGUI::on_open));
	_actiongroup->add(Gtk::Action::create(
		"FileSave", Gtk::Stock::SAVE, "_Save", "Save a decoder setup"),
		sigc::mem_fun(*this, &AmbisonicsDecoderGUI::on_save));
	_actiongroup->add(Gtk::Action::create(
		"FileLoadFilters", "_Load filters", "Load a set of filters"),
		sigc::mem_fun(*this, &AmbisonicsDecoderGUI::on_load_filters));
	_actiongroup->add(Gtk::Action::create(
		"FileQuit", Gtk::Stock::QUIT, "_Quit", "Quit"),
		sigc::mem_fun(*this, &AmbisonicsDecoderGUI::on_quit));

	// View menu
	_actiongroup->add(Gtk::Action::create("ViewMenu", "_View"));
	_act_speakerlayout = Gtk::ToggleAction::create(
		"ViewLoudspeakerLayout", "_Loudspeaker layout",
		"View the loudspeaker layout");
	_act_settings = Gtk::ToggleAction::create(
		"ViewSettings", "_Settings", "View settings");
	_act_meters = Gtk::ToggleAction::create(
		"ViewLevelMeters", "Level _meters", "View level meters");
	_actiongroup->add(_act_speakerlayout, Gtk::AccelKey("<control>L"),
		sigc::mem_fun(*this, &AmbisonicsDecoderGUI::on_view_speakers));
	_actiongroup->add(_act_settings, Gtk::AccelKey("<control>E"),
		sigc::mem_fun(*this, &AmbisonicsDecoderGUI::on_view_settings));
	_actiongroup->add(_act_meters, Gtk::AccelKey("<control>M"),
		sigc::mem_fun(*this, &AmbisonicsDecoderGUI::on_view_meters));

	// Routing menu
	_actiongroup->add(Gtk::Action::create("RoutingMenu", "_Routing"));
	_actiongroup->add(Gtk::Action::create(
		"RoutingInputs", "Auto-connect _inputs", "Auto-connect JACK inputs"),
		sigc::mem_fun(*this, &AmbisonicsDecoderGUI::on_route_inputs));
	_actiongroup->add(Gtk::Action::create(
		"RoutingOutputs", "Auto-connect _outputs", "Auto-connect JACK outputs"),
		sigc::mem_fun(*this, &AmbisonicsDecoderGUI::on_route_outputs));

	// Help menu
	_actiongroup->add(Gtk::Action::create("HelpMenu", "_Help"));
	_actiongroup->add(Gtk::Action::create(
		"HelpAbout", "_About", "About program"),
		sigc::mem_fun(*this, &AmbisonicsDecoderGUI::on_view_aboutbox));

	_uimanager = Gtk::UIManager::create();
	_uimanager->insert_action_group(_actiongroup);
	add_accel_group(_uimanager->get_accel_group());
	_uimanager->add_ui_from_string(_menu_info);
	
	Gtk::Widget* menu = _uimanager->get_widget("/MenuBar");
	vbox1->pack_start(*menu, Gtk::PACK_SHRINK);
	
	table->set_col_spacings(20);
	table->set_row_spacings(20);
	table->set_border_width(16);
	Gtk::Label* lsp_label = Gtk::manage(new Gtk::Label("Speakers:"));
	lsp_label->set_alignment(Gtk::ALIGN_RIGHT);
	Gtk::Label* order_label = Gtk::manage(new Gtk::Label("Order:"));
	order_label->set_alignment(Gtk::ALIGN_RIGHT);
	table->attach(*lsp_label, 0, 1, 0, 1);
	table->attach(*order_label, 0, 1, 1, 2);
	table->attach(_btn_numspeakers, 1, 2, 0, 1, Gtk::EXPAND);
	table->attach(_btn_order, 1, 2, 1, 2, Gtk::EXPAND);
	table->attach(_btn_3d, 1, 2, 2, 3, Gtk::EXPAND);
	table->attach(_sld_volume, 2, 3, 0, 3);
	vbox1->pack_start(*table);
	vbox1->pack_start(_statusbar, Gtk::PACK_SHRINK);

	// Connect signal handlers
	_btn_numspeakers.signal_value_changed().connect(sigc::mem_fun(
		*this, &AmbisonicsDecoderGUI::on_numspeakers_changed));
	_btn_order.signal_value_changed().connect(sigc::mem_fun(
		*this, &AmbisonicsDecoderGUI::on_order_changed));
	_btn_3d.signal_clicked().connect(sigc::mem_fun(
		*this, &AmbisonicsDecoderGUI::on_3d_clicked));
	_sld_volume.signal_value_changed().connect(sigc::mem_fun(
		*this, &AmbisonicsDecoderGUI::on_volume_changed));

	_wnd_speakers.signal_delete_event().connect(sigc::mem_fun(
			*this, &AmbisonicsDecoderGUI::on_hide_speakers));
	_wnd_settings.signal_delete_event().connect(sigc::mem_fun(
		*this, &AmbisonicsDecoderGUI::on_hide_settings));
	_wnd_meters.signal_delete_event().connect(sigc::mem_fun(
		*this, &AmbisonicsDecoderGUI::on_hide_meters));

	_decoder.AddView(this);
	_decoder.AddView(&_wnd_speakers);
	_decoder.AddView(&_wnd_settings);
	_decoder.AddView(&_wnd_meters);
	show_all_children();
}

void AmbisonicsDecoderGUI::Notify(int what)
{
	switch (what) {
	case AmbisonicsDecoderModel::ChangedNumSpeakers:
		_btn_numspeakers.set_value(_decoder.NumSpeakers());
		break;
	case AmbisonicsDecoderModel::ChangedOrder:
		_btn_order.set_value(_decoder.Order());
		break;
	case AmbisonicsDecoderModel::ChangedIs3D: 
		_btn_3d.set_active(_decoder.Is3D());
		break;
	case AmbisonicsDecoderModel::ChangedGain:
		_sld_volume.set_value(_decoder.Gain());
		break;
	case AmbisonicsDecoderModel::ChangedInputFilters:
		_statusbar.pop();
		_statusbar.push(_decoder.InputFiltersEnabled() ?
		                "Input filters loaded" : "No input filters");
	case AmbisonicsDecoderModel::ChangedNormalization:
		if (_decoder.Normalization() == Ambisonics::FurseMalham)
			_btn_order.set_range(0, 3);
		else
			_btn_order.set_range(0, 15);
	case AmbisonicsDecoderModel::ChangedAll:
		_btn_numspeakers.set_value(_decoder.NumSpeakers());
		_btn_order.set_value(_decoder.Order());
		_btn_3d.set_active(_decoder.Is3D());
		_sld_volume.set_value(_decoder.Gain());
		if (_decoder.Normalization() == Ambisonics::FurseMalham)
			_btn_order.set_range(0, 3);
		else
			_btn_order.set_range(0, 15);
		_statusbar.pop();
		_statusbar.push(_decoder.InputFiltersEnabled() ?
		                "Input filters loaded" : "No input filters");
		break;
	}
}

void AmbisonicsDecoderGUI::on_numspeakers_changed()
{
	_decoder.NumSpeakers(_btn_numspeakers.get_value_as_int());	
}

void AmbisonicsDecoderGUI::on_order_changed()
{
	_decoder.Order(_btn_order.get_value_as_int());
}

void AmbisonicsDecoderGUI::on_3d_clicked()
{
	_decoder.Is3D(_btn_3d.get_active());
}

void AmbisonicsDecoderGUI::on_volume_changed()
{
	_decoder.Gain(_sld_volume.get_value());
}

void AmbisonicsDecoderGUI::on_quit()
{
	hide();
}

void AmbisonicsDecoderGUI::on_load_filters()
{
	Gtk::FileChooserDialog dlg("Open filters", Gtk::FILE_CHOOSER_ACTION_OPEN);
	Gtk::FileFilter adf_filter, all_filter;
	
	dlg.set_transient_for(*this);
	dlg.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	dlg.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);
	adf_filter.set_name("Filter files");
	adf_filter.add_pattern("*.adf");
	adf_filter.add_pattern("*.ADF");
	all_filter.set_name("All files");
	all_filter.add_pattern("*");
	dlg.add_filter(adf_filter);
	dlg.add_filter(all_filter);
	int result = dlg.run();
	if (result == Gtk::RESPONSE_OK) {
		std::string errmsg;

		if (!_decoder.LoadFilters(dlg.get_filename(), &errmsg)) {
			Gtk::MessageDialog msg(dlg, errmsg, false, Gtk::MESSAGE_ERROR,
			                       Gtk::BUTTONS_OK, true);
			
			msg.run();
		}
	}
}

void AmbisonicsDecoderGUI::on_save()
{
	Gtk::FileChooserDialog dlg("Save setup", Gtk::FILE_CHOOSER_ACTION_SAVE);
	Gtk::FileFilter ambidec_filter;
	
	dlg.set_transient_for(*this);
	dlg.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	dlg.add_button(Gtk::Stock::SAVE, Gtk::RESPONSE_OK);
	ambidec_filter.set_name("Ambidec files");
	ambidec_filter.add_pattern("*.ambidec");
	//dlg.set_do_overwrite_confirmation();
	dlg.add_filter(ambidec_filter);
	int result = dlg.run();
	if (result == Gtk::RESPONSE_OK) {
		Glib::ustring fn = dlg.get_filename();
		std::string errmsg;
		// Check if user has specified an extension
		if (fn.find('.') == Glib::ustring::npos)
			// If not, append the default
			fn += ".ambidec";
		
		if (!_decoder.Save(fn, &errmsg)) {
			Gtk::MessageDialog msg(dlg, errmsg, false, Gtk::MESSAGE_ERROR,
			                       Gtk::BUTTONS_OK, true);

			msg.run();
		}
	}
}

void AmbisonicsDecoderGUI::on_open()
{
	Gtk::FileChooserDialog dlg("Open setup", Gtk::FILE_CHOOSER_ACTION_OPEN);
	Gtk::FileFilter ambidec_filter, all_filter;

	dlg.set_transient_for(*this);
	dlg.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	dlg.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);
	ambidec_filter.set_name("Ambidec files");
	ambidec_filter.add_pattern("*.ambidec");
	all_filter.set_name("All files");
	all_filter.add_pattern("*");
	dlg.add_filter(ambidec_filter);
	dlg.add_filter(all_filter);
	int result = dlg.run();
	if (result == Gtk::RESPONSE_OK) {
		std::string errmsg;

		if (!_decoder.Load(dlg.get_filename(), &errmsg)) {
			Gtk::MessageDialog msg(dlg, errmsg, false, Gtk::MESSAGE_ERROR,
			                       Gtk::BUTTONS_OK, true);

			msg.run();
		}
	}
}

void AmbisonicsDecoderGUI::on_view_speakers()
{
	if (_act_speakerlayout->get_active())
		_wnd_speakers.show();
	else
		_wnd_speakers.hide();
}

bool AmbisonicsDecoderGUI::on_hide_speakers(GdkEventAny* e)
{
	_act_speakerlayout->set_active(false);
	return false; // Allow window to close
}

void AmbisonicsDecoderGUI::on_view_settings()
{
	if (_act_settings->get_active())
		_wnd_settings.show();
	else
		_wnd_settings.hide();
}

bool AmbisonicsDecoderGUI::on_hide_settings(GdkEventAny* e)
{
	_act_settings->set_active(false);
	return false; // Allow window to close
}

void AmbisonicsDecoderGUI::on_view_meters()
{
	if (_act_meters->get_active())
		_wnd_meters.show();
	else
		_wnd_meters.hide();
}

bool AmbisonicsDecoderGUI::on_hide_meters(GdkEventAny* e)
{
	_act_meters->set_active(false);
	return false; // Allow window to close
}

#include "LicenseText.hpp"

void AmbisonicsDecoderGUI::on_view_aboutbox()
{
	Gtk::AboutDialog dlg;
	std::vector<Glib::ustring> authors;

	dlg.set_skip_taskbar_hint();
	authors.push_back("Thom Johansen");
	authors.push_back("Bård støfringsdal");
	authors.push_back("Peter Svensson");
	dlg.set_comments("An HOA Ambisonics decoder for JACK");
	dlg.set_version("0.2");
	dlg.set_authors(authors);
	dlg.set_copyright("Copyright © 2006-2007 Q2S, "
	                  "Norwegian University of Science and Technology");
	dlg.set_website("http://www.q2s.ntnu.no");
	dlg.set_license(license_text);
	// logo?
	dlg.run();
}

void AmbisonicsDecoderGUI::on_route_inputs()
{
	const char* str[] = { "ambienc:output_", "ldc:output_",
	                      "alsa_pcm:capture_" };
	StringInputDialog input("Target ports", str, sizeof(str)/sizeof(char*));

	if (input.Run())
		_snd.ConnectInputs(input.GetString());
}

void AmbisonicsDecoderGUI::on_route_outputs()
{
	const char* str[] = { "alsa_pcm:playback_" };
	StringInputDialog input("Target ports", str, sizeof(str)/sizeof(char*));

	if (input.Run())
		_snd.ConnectOutputs(input.GetString());
}

