/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _ambisonicsdecodermodel_h
#define _ambisonicsdecodermodel_h

#include <boost/lexical_cast.hpp>
#include <cmath>
#include <cstring>
#include <fstream>
#include <iostream>
#include <string>
#include "AmbisonicsDecoder.hpp"
#include "ConfigFile.hpp"
#include "AudioProcessor.hpp"
#include "Model.hpp"
#include "PositionVector.hpp"

class AmbisonicsDecoderModel : public Model {
public:
	enum SpeakerLayoutPreset { CustomLayout, CircleLayout };
private:
	PositionVector _speakers;
	AmbisonicsDecoder _dec;
	AudioInterface& _snd;
	bool _3d;
	SpeakerLayoutPreset _speakerlayout;
	unsigned _order;
	double _gain, _first_speaker_azimuth;
	Ambisonics::Normalization _normalization;
	double _regparm;

	void UpdateSpeakerLayout()
	{
		if (_speakerlayout == CircleLayout)
			_speakers.MakeCircle(_first_speaker_azimuth, 1., -360.);
		for (size_t i = 0; i < _speakers.size(); ++i) {
			_dec.SetAzimuth(i, _speakers[i].azimuth*M_PI/180.);
			_dec.SetElevation(i, _speakers[i].elevation*M_PI/180.);
		}
	}

	void UpdateDecoder()
	{
		AudioInterface::Disable lock(_snd);
		_dec.SetNumSpeakers(_speakers.size());
		_dec.SetIs3D(_3d);
		_dec.SetOrder(_order);
		_dec.SetGain(powf(10.f, _gain/20.f));
		_dec.SetNormalization(_normalization);
		_dec.SetRegularisationParam(_regparm);
		UpdateSpeakerLayout();
		_snd.UpdatePorts();
	}
public:
	AudioProcessor* GetAudioProcessor()
	{
		return &_dec;
	}

	enum {
		ChangedAll = 0, ChangedIs3D, ChangedOrder, ChangedGain,
		ChangedSpeakerLayout, ChangedFirstSpeakerAzimuth,
		ChangedNumSpeakers, ChangedSpeakerPosition, ChangedInputFilters,
		ChangedNormalization, ChangedRegularisationParam
	};

	AmbisonicsDecoderModel(AudioInterface& ai) : _snd(ai)
 	{
		DefaultSettings();
		_speakers.resize(16);
		_snd.SetProcess(_dec);
		UpdateDecoder();
		_snd.EnableProcessing();
	}

	void DefaultSettings()
	{
		_3d = false;
		_speakerlayout = CircleLayout;
		_order = 1;
		_gain = 0.;
		_first_speaker_azimuth = 0.;
		_normalization = Ambisonics::Normalized;
		_regparm = 1./1000.;
	}

	void Is3D(bool is3d)
	{
		{
			AudioInterface::Disable lock(_snd);
			if (is3d == _3d) return;
			_3d = is3d;
			_dec.SetIs3D(is3d);
			_snd.UpdatePorts(); // Number of inputs has changed
		}
		NotifyViews(ChangedIs3D);
		NotifyViews(ChangedInputFilters);
	}
	
	bool Is3D() const
	{
		return _3d;
	}
	
	void Order(unsigned order)
	{
		{
			AudioInterface::Disable lock(_snd);
			if (order == _order) return;
			_order = order;
			_dec.SetOrder(order);
			_snd.UpdatePorts(); // Number of inputs has changed
		}
		NotifyViews(ChangedOrder);
		NotifyViews(ChangedInputFilters);
	}
	
	unsigned Order() const
	{
		return _order;
	}

	void Gain(double gain)
	{
		{
			AudioInterface::Disable lock(_snd);
			if (gain == _gain) return;
			_gain = gain;
			float linval = powf(10.f, gain/20.f);
			_dec.SetGain(linval);
		}
		NotifyViews(ChangedGain);
	}

	double Gain() const
	{
		return _gain;
	}

	void SpeakerLayout(SpeakerLayoutPreset layout)
	{
		{
			AudioInterface::Disable lock(_snd);
			if (layout == _speakerlayout) return;
			_speakerlayout = layout;
			UpdateSpeakerLayout();
		}
		NotifyViews(ChangedSpeakerLayout);
	}
	
	SpeakerLayoutPreset SpeakerLayout() const
	{
		return _speakerlayout;
	}
	
	void FirstSpeakerAzimuth(double azimuth)
	{
		{
			AudioInterface::Disable lock(_snd);
			if (azimuth == _first_speaker_azimuth) return;
			_first_speaker_azimuth = azimuth;
			UpdateSpeakerLayout();
		}
		NotifyViews(ChangedFirstSpeakerAzimuth);
	}
	
	double FirstSpeakerAzimuth() const
	{
		return _first_speaker_azimuth;
	}
	
	void NumSpeakers(unsigned numspeakers)
	{
		{
			AudioInterface::Disable lock(_snd);
			if (numspeakers == _speakers.size()) return;
			_dec.SetNumSpeakers(numspeakers);
			_speakers.resize(numspeakers);
			UpdateSpeakerLayout();
			_snd.UpdatePorts(); // Number of outputs changed
		}
		NotifyViews(ChangedNumSpeakers);
		NotifyViews(ChangedInputFilters);
	}

	size_t NumSpeakers() const
	{
		return _speakers.size();
	}

	void Normalization(Ambisonics::Normalization norm)
	{
		{
			AudioInterface::Disable lock(_snd);
			if (norm == _normalization) return;
			_normalization = norm;
			if (norm == Ambisonics::FurseMalham && _order > 3) {
				_order = 3;
				_dec.SetOrder(3);
				NotifyViews(ChangedOrder);
			}
			_dec.SetNormalization(norm);
		}
		NotifyViews(ChangedNormalization);
	}

	Ambisonics::Normalization Normalization() const
	{
		return _normalization;
	}

	void SpeakerAzimuth(size_t spk, double azimuth)
	{
		{
			AudioInterface::Disable lock(_snd);
			if (_speakers[spk].azimuth == azimuth) return;
			_speakers[spk].azimuth = azimuth;
			_dec.SetAzimuth(spk, azimuth*M_PI/180.);
		}
		NotifyViews(ChangedSpeakerPosition);
	}

	double SpeakerAzimuth(size_t spk) const
	{
		return _speakers[spk].azimuth;
	}

	void SpeakerElevation(size_t spk, double elev)
	{
		{
			AudioInterface::Disable lock(_snd);
			if (_speakers[spk].elevation == elev) return;
			_speakers[spk].elevation = elev;
			_dec.SetElevation(spk, elev*M_PI/180.);
		}
		NotifyViews(ChangedSpeakerPosition);
	}

	double SpeakerElevation(size_t spk) const
	{
		return _speakers[spk].elevation;
	}

	void RegularisationParam(double regparm)
	{
		{
			AudioInterface::Disable lock(_snd);
			if (_regparm == regparm) return;
			_regparm = regparm;
			_dec.SetRegularisationParam(regparm);
		}
		NotifyViews(ChangedRegularisationParam);
	}
	
	double RegularisationParam() const
	{
		return _regparm;
	}

	bool InputFiltersEnabled() const
	{
		return _dec.InputFiltersEnabled();
	}

	bool LoadFilters(std::string filename, std::string* errmsg = 0)
	{
		// TODO: currently, we assume the file was written with the same endian
		// order as we're currently executing on
		AudioInterface::Disable lock(_snd);
		std::ifstream f(filename.c_str(), std::ios::in);
		char buf[4];
		unsigned int samplerate, order, filterlength, nlsp;
	
		if (!f) {
			if (errmsg)
				errmsg->assign("Could not open file");
			return false;
		}
		f.exceptions(std::ifstream::eofbit | std::ifstream::failbit
		             | std::ifstream::badbit);
		try {
			f.read(buf, 3); // Read signature
			if (strncmp(buf, "ADF", 3)) {
				if (errmsg)
					errmsg->assign("File is not a valid ambidec filter file");
				return false;
			}
			f.read(reinterpret_cast<char*>(&samplerate), 4);
			f.read(reinterpret_cast<char*>(&order), 4);
			f.read(reinterpret_cast<char*>(&filterlength), 4);
			f.read(reinterpret_cast<char*>(&nlsp), 4);
		
			if (order != Order()) {
				if (errmsg)
					errmsg->assign("Filter file intended for Ambisonics order "
					               + boost::lexical_cast<std::string>(order));
				return false;
			}
			if (nlsp != NumSpeakers()) {
				if (errmsg)
					errmsg->assign("Filter file intended for "
					               + boost::lexical_cast<std::string>(nlsp)
					               + " loudspeakers");
				return false;
			}
			std::vector<FloatVector> filters(order + 1,
			                                 FloatVector(filterlength));
			for (size_t i = 0; i < order + 1; ++i)
				f.read(reinterpret_cast<char*>(&filters[i][0]), 4*filterlength);
			_dec.SetFilters(filters);
		}
		catch (std::ifstream::failure& e) {
			if (errmsg)
				errmsg->assign("Read error");
			return false;
		}
		NotifyViews(ChangedInputFilters);
		return true;
	}

	bool Save(std::string filename, std::string* errmsg = 0)
	{
		try {
			ConfigWriter c(filename);

			c.SetItem("ambidec_cfg_version", 2);
			c.SetItem("order", _order);
			c.SetItem("3d", _3d);
			c.SetItem("first_speaker_azimuth", _first_speaker_azimuth);
			c.SetItem("gain", _gain);
			c.SetItem("speaker_layout", _speakerlayout);
			c.SetItem("normalization", _normalization);
			c.SetItem("regularization_parameter", _regparm);
			c.Comment("order is \"azimuth, elevation\"");
			c.StartMatrixItem("loudspeakers", 2, _speakers.size());
			for (size_t i = 0; i < _speakers.size(); ++i) {
				c.SetItemValue(_speakers[i].azimuth);
				c.SetItemValue(_speakers[i].elevation);
			}
		}
		catch (std::runtime_error& e) {
			if (errmsg)
				errmsg->assign(e.what());
			return false;
		}
		return true;
	}

	bool Load(std::string filename, std::string* errmsg = 0)
	{
		try {
			ConfigReader c(filename, true);
			int t;

			c.GetValue("ambidec_cfg_version", t);
			if (t != 2) {
				if (errmsg)
					errmsg->assign("Incompatible configuration file version");
				return false;
			}
			c.GetValue("order", _order);
			c.GetValue("3d", _3d);
			c.GetValue("first_speaker_azimuth", _first_speaker_azimuth);
			c.GetValue("gain", _gain);
			c.GetValue("speaker_layout", t);
			_speakerlayout = AmbisonicsDecoderModel::SpeakerLayoutPreset(t);
			c.GetValue("normalization", t);
			_normalization = Ambisonics::Normalization(t);
			c.GetValue("regularization_parameter", _regparm);
			ConfigReader::Item* item = c.GetItem("loudspeakers");
			_speakers.resize(item->ColSize(1));
			for (size_t i = 0; i < _speakers.size(); ++i) {
				c.GetValue("loudspeakers", _speakers[i].azimuth, 1, i + 1);
				c.GetValue("loudspeakers", _speakers[i].elevation, 2, i + 1);
			}
		}
		catch (std::runtime_error& e) {
			if (errmsg)
				errmsg->assign(e.what());
			// State might be inconsistent here, reset to a good default.
			// User tried to load new settings anyway, so shouldn't care too
			// much if the old ones got lost.
			DefaultSettings();
			UpdateDecoder();
			NotifyViews();
			return false;
		}
		UpdateDecoder();
		NotifyViews();
		return true;
	}
};

#endif

