/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "AmbisonicsEncoder.hpp"
#include <cstring>

#ifdef __SSE__
#include <xmmintrin.h>
#endif
	
AmbisonicsEncoder::AmbisonicsEncoder(double max_radius)
    : _nfcradiuslimiter(2.f),
	  _srcradiuslimiter(6.f), _max_radius(max_radius),
	  _amplimiter(_srcradiuslimiter, 0.f, max_radius, int(max_radius)*10),
	  _nfc_samples_to_recalc(0), _is3d(false), _order(1), _lsp_radius(2.),
	  _ambi(_order, _is3d)
{
#ifdef __SSE__
	// If there's a chance we'll be using SSE for calculations, let's disable
	// denormal processing.
	_MM_SET_FLUSH_ZERO_MODE(_MM_FLUSH_ZERO_ON);
#endif
	_math = SIMDMathPtr(SIMDMath::Create());
	_ambi.SetNormalization(Ambisonics::Normalized);
}

void AmbisonicsEncoder::UpdateNFCFilters(size_t source)
{
	Source& s = _sources[source];
	double source_rad;

	if (s.planewave)
		source_rad = NFCFilter::MaxDistance + 10;
	else if (s.interp_radius)
		source_rad = _interpolator.GetRadius(source);
	else
		source_rad = _sourcepos[source].radius;
	
	if (!s.planewave && !s.nfc_highpass)
		source_rad = _nfcradiuslimiter(source_rad);
	for (size_t order = 0; order < s.nfcfilters.size(); ++order) {
		s.nfcfilters[order].SetDistances(source_rad, _lsp_radius);
		//printf("order %zu:\n", order + 1);
		//s.nfcfilters[order].DumpSpectrum(_sample_rate);
		if (s.nfc_highpass)
			s.nfclimiters[order].SetDistance(source_rad, _is3d);
	}
}

void AmbisonicsEncoder::InitSource(size_t source)
{
	Source& s = _sources[source];

	// Create NFC filter for new source
	s.nfcfilters.resize(_order, NFCFilter(_sample_rate));
	s.nfclimiters.resize(_order, NFCHighpass(_sample_rate, 15));
	 // Filters have increasing orders starting from order 1
	for (size_t i = 0; i < _order; ++i) {
		s.nfcfilters[i].SetOrder(i + 1);
		s.nfclimiters[i].SetOrder((i + 1));
	}
	// Give filter valid coefs.
	UpdateNFCFilters(source);
	s.delay.SetDistance(_amplimiter(_sourcepos[source].radius));
	// Fill decoding matrix entry with valid values
	_ambi.EncodingCoefs(1.0, _sourcepos[source].azimuth,
	                    _sourcepos[source].elevation, _matrix, source);
}

void AmbisonicsEncoder::SetNumSources(size_t num)
{
	size_t old_num = _sources.size();
	size_t comps = _ambi.NumComponents();

	_sourcepos.resize(num);
	_sources.resize(num, Source(_sample_rate));
	_interpolator.SetNumSources(num);
	_matrix.resize(num, comps);
	for (size_t source = old_num; source < num; ++source)
		InitSource(source);
}

void AmbisonicsEncoder::SetIs3D(bool is3d)
{
	_is3d = is3d;
	_ambi.SetIs3D(is3d);
	size_t num = _ambi.NumComponents();

	_matrix.resize(_sources.size(), num);
	for (size_t source = 0; source < _sources.size(); ++source)
		InitSource(source);
}

void AmbisonicsEncoder::SetOrder(unsigned order)
{
	_order = order;
	_ambi.SetOrder(order);
	size_t num = _ambi.NumComponents();

	_matrix.resize(_sources.size(), num);
	for (size_t source = 0; source < _sources.size(); ++source)
		InitSource(source);
}

void AmbisonicsEncoder::SetMaxRadius(double max_radius)
{
	_max_radius = max_radius;
	_amplimiter.SetRange(0.f, max_radius, int(max_radius)*10);
}

void AmbisonicsEncoder::SetLoudspeakerRadius(double radius)
{
	_lsp_radius = radius;
	_nfcradiuslimiter.Min(radius);
	for (size_t source = 0; source < _sources.size(); ++source)
		UpdateNFCFilters(source);
}

void AmbisonicsEncoder::SetMaxSourceAmplification(double amp)
{
	_srcradiuslimiter.Max(amp);
	// Recalc interpolation table
	_amplimiter.SetRange(0.f, _max_radius, int(_max_radius)*10);
}

void AmbisonicsEncoder::SetDistanceSimulation(size_t source, bool enable)
{
	Source& s = _sources[source];

	if (!s.planewave)
		s.gain_xfader.Start(enable, _sample_rate/5);
	s.distance_sim = enable;
}

void AmbisonicsEncoder::SetDelaySimulation(size_t source, bool enable)
{
	Source& s = _sources[source];
	unsigned delay = 0;

	if (enable && s.delay_xfader.Finished()) {
		// Delay the onset of the crossfade by the max delay time so that the
		// delay line will have time to fill with new data (the current contents
		// will be old data), avoiding any artifacts
		delay = unsigned(_max_radius/344.*_sample_rate);
	}
	if (!s.planewave)
		s.delay_xfader.Start(enable, _sample_rate/5, delay);
	s.delay_sim = enable;
}

void AmbisonicsEncoder::SetPlaneWave(size_t source, bool enable)
{
	Source& s = _sources[source];
	unsigned delay = 0;

	if (!enable && s.delay_xfader.Finished()) {
		// See comment in SetDelaySimulation()
		delay = unsigned(_max_radius/344.*_sample_rate);
	}
	if (s.delay_sim)
		s.delay_xfader.Start(!enable, _sample_rate/5, delay);
	if (s.distance_sim)
		s.gain_xfader.Start(!enable, _sample_rate/5);
	s.planewave = enable;
	UpdateNFCFilters(source);
}

void AmbisonicsEncoder::SetNFC(size_t source, bool enable)
{
	Source& s = _sources[source];

	// NFC filters are not running while disabled, so we should clear their
	// histories when reenabling them
	if (enable && s.nfc_xfader.Finished()) {
		for (size_t order = 0; order < s.nfcfilters.size(); ++order) {
			s.nfcfilters[order].ClearHistory();
			s.nfclimiters[order].ClearHistory();
		}
	}
	s.nfc_xfader.Start(enable, _sample_rate/5);
	s.nfc = enable;
}

void AmbisonicsEncoder::SetNFCHighpass(size_t source, bool enabled)
{
	_sources[source].nfc_highpass = enabled;
	UpdateNFCFilters(source);
}

void AmbisonicsEncoder::SetInterpolationHalfTime(double time)
{
	_interpolator.SetScale(time*double(_sample_rate));
}

void AmbisonicsEncoder::SetNormalization(Ambisonics::Normalization norm)
{
	_ambi.SetNormalization(norm);
	_ambi.EncodingCoefs(1.0, _sourcepos, _matrix);
}

void AmbisonicsEncoder::SetAzimuth(size_t source, double azimuth, bool interp)
{
	_sourcepos[source].azimuth = azimuth;
	_interpolator.SetAzimuth(source, azimuth, interp);
	_ambi.EncodingCoefs(1.0, azimuth, _sourcepos[source].elevation, _matrix,
	                    source);
	_sources[source].interp_azimuth = interp;
}

void AmbisonicsEncoder::SetElevation(size_t source, double elev, bool interp)
{
	_sourcepos[source].elevation = elev;
	_interpolator.SetElevation(source, elev, interp);
	_ambi.EncodingCoefs(1.0, _sourcepos[source].azimuth, elev, _matrix, source);
	_sources[source].interp_elev = interp;
}

void AmbisonicsEncoder::SetRadius(size_t source, double radius, bool interp)
{
	Source& s = _sources[source];

	_sourcepos[source].radius = radius;
	_interpolator.SetRadius(source, radius, interp);
	s.delay.SetDistance(radius);
	s.interp_radius = interp;
	for (size_t order = 0; order < s.nfclimiters.size(); ++order) {
		s.nfclimiters[order].SetDistance(radius, _is3d);
		s.nfclimiters[order].ClearHistory();
	}
}

void AmbisonicsEncoder::SetGain(size_t source, float gain, bool interp)
{
	Source& s = _sources[source];
	unsigned samples = interp ? _sample_rate/10 : 0;

	s.gain = gain;
	s.gainer.SetGain(gain, samples);
}

void AmbisonicsEncoder::SetSampleRate(unsigned sr)
{
	AudioProcessor::SetSampleRate(sr);
	// Update filters 400 times per sec
	_nfc_recalc_delta = sr/400;
	for (size_t src = 0; src < _sources.size(); ++src) {
		Source& s = _sources[src];

		for (size_t i = 0; i < s.nfcfilters.size(); ++i) {
			s.nfcfilters[i].SetSampleRate(sr);
			s.nfclimiters[i].SetSampleRate(sr);
		}
		UpdateNFCFilters(src);
		s.delay.SetSampleRate(sr);
		s.delay.SetDistance(_sourcepos[src].radius);
	}
}

void AmbisonicsEncoder::SetBufferSize(unsigned size)
{
	AudioProcessor::SetBufferSize(size);
	_source_buf.resize(size);
	_nfc_source_buf.resize(size);
	_azimuth_buf.resize(size);
	_elev_buf.resize(size);
	_radius_buf.resize(size);
	_temp_buf.resize(size);
}

void AmbisonicsEncoder::SourceSimProcessing(float* input, float* output,
                                            size_t src, unsigned num)
{
	Source& s = _sources[src];
	const bool gain_xfade = !s.gain_xfader.Finished();
	const bool delay_xfade = !s.delay_xfader.Finished();
	const bool xfading = gain_xfade || delay_xfade;
	float* amp = gain_xfade ? output : &_temp_buf[0];
	float* rad = &_radius_buf[0];
	float* in;

	// Check if we want to do source distance simulation at all
	if (!xfading && (s.planewave || (!s.distance_sim && !s.delay_sim))) {
   		// No source distance simulation
		memcpy(output, input, num*sizeof(float));
		return;
	} 

	// Check if we'll be doing distance attenuation
	if (s.distance_sim || gain_xfade) {
		// If so, check if we need to generate a whole array of amplitude
		// limited 1/radius values, or just one
		if (s.interp_radius) {
			float rcp_lsp_radius = 1.f/float(_lsp_radius);

			// TODO: find out if such an elaborate shaping function is really
			// needed! this thing is expensive.
			for (unsigned i = 0; i < num; ++i)
				amp[i] = _srcradiuslimiter(_radius_buf[i]*rcp_lsp_radius);
		} else {
			*amp = _srcradiuslimiter(_sourcepos[src].radius/_lsp_radius);
		}
	}

	// The signal chain looks like this:
	//         _______       _______      ______       _______
	//        |       |     |       |    |      |     |       |
	// In ____| Delay |_____| Xfade |____| Gain |_____| Xfade |____ Out
	//     |  |_______|  ___|       | |  |______|  ___|       |
	//     |____________|   |_______| |___________|   |_______|
	//
	
	// Delay simulation processing
	if (s.delay_sim || delay_xfade) {
		// If distance attenuation is currently crossfading, use _temp_buf as a
		// mix buffer, since it's not used yet. If it's not crossfading, just
		// write straight to the output buffer.
		in = gain_xfade ? &_temp_buf[0] : output;
		if (s.interp_radius) 
			s.delay.Process(input, in, rad, num);
		else
			s.delay.Process(input, in, num);
		if (delay_xfade)
			s.delay_xfader.Process(in, input, in, num);
	} else {
		// Delay simulation is disabled, make distance attenuation take its
		// input from the input array
		in = input;
	}

	// Distance attenuation processing
	if (s.distance_sim || gain_xfade) {
		if (s.interp_radius)
			_math->VVMul(output, in, amp, num);
		else
			_math->VSMul(output, in, *amp, num);
		if (gain_xfade)
			s.gain_xfader.Process(output, in, output, num);
	}
}

// Needs _radius_buf to be valid
void AmbisonicsEncoder::NFCProcessing(float* input, float* output, size_t src,
                                      unsigned order, unsigned num)
{
	Source& s = _sources[src];

	// Recall state of crossfader at start of fade
	s.nfc_xfader.LoadState();
	const bool nfc_xfade = !s.nfc_xfader.Finished();
	float* out = nfc_xfade ? &_temp_buf[0] : output;

	if (!s.nfc && !nfc_xfade) { // Is NFC disabled?
		if (input != output)
			memcpy(output, input, num*sizeof(float));
		return;
	}
	
	if (s.interp_radius) {
		double source_rad;
		unsigned pos = 0; // Current position in the arrays

		while (num - pos > 0) {
			if (_nfc_samples_to_recalc_temp == 0) {
				_nfc_samples_to_recalc_temp = _nfc_recalc_delta;
				source_rad = s.planewave
				             ? NFCFilter::MaxDistance + 10 
				             : _radius_buf[pos];

				if (!s.planewave && !s.nfc_highpass)
					source_rad = _nfcradiuslimiter(source_rad);

				s.nfcfilters[order].SetDistance(source_rad);
				if (s.nfc_highpass)
					s.nfclimiters[order].SetDistance(source_rad, _is3d);
			}
			unsigned samples = std::min(_nfc_samples_to_recalc_temp, num - pos);
			s.nfcfilters[order].Process(input + pos, out + pos, samples);
			if (s.nfc_highpass)
				s.nfclimiters[order].Process(out + pos, out + pos,
				                             samples);

			_nfc_samples_to_recalc_temp -= samples;
			pos += samples;
		}
	} else {
		// Radius parameter is not currently being interpolated, and NFC filters
		// should already be set with correct parameters.
		s.nfcfilters[order].Process(input, out, num);
		if (s.nfc_highpass)
			s.nfclimiters[order].Process(out, out, num);
	}
	if (nfc_xfade)
		s.nfc_xfader.Process(out, input, output, num);
}

void AmbisonicsEncoder::ProcessSource(float* input, float** outputs,
                                      size_t src, unsigned num)
{
	unsigned comp = 1;
	Source& s = _sources[src];

	// See if radius parameter needs interpolating.
	if (s.interp_radius) {
		// Try to interpolate it. If we succeed, radius values are placed in
		// _radius_buf. If not, we'll mark this source as finished with
		// interpolating, then move on.
		if (!_interpolator.InterpolateRadius(src, num, _radius_buf)) {
			// Interpolation has finished
			s.interp_radius = false;
			// Set NFC filters to final value
			UpdateNFCFilters(src);
		}
	}

	// Do the gain and delay modelling
	SourceSimProcessing(input, &_source_buf[0], src, num);

	// Gain the source input, if needed
	// TODO? once the gainer is finished interpolating, the final gain value
	// can be baked into the encoding matrix at the cost of more complexity
	if (s.gain != 1.f || s.gainer.Interpolating())
		s.gainer.Work(&_source_buf[0], num);

	// Check if we need to generate any interpolated azimuth/elevation data
	if (s.interp_azimuth && !_interpolator.InterpolateAzimuth(src, num,
	                                                          _azimuth_buf))
		s.interp_azimuth = false;
	if (s.interp_elev && !_interpolator.InterpolateElevation(src, num,
	                                                         _elev_buf))
		s.interp_elev = false;

	// We need both interpolation arrays to be filled, if we're not
	// interpolating both parameters, fill one of them with static values.
	if (_is3d) {
		if (!s.interp_elev && s.interp_azimuth)
			std::fill(_elev_buf.begin(), _elev_buf.end(),
			          _sourcepos[src].elevation);
		else if (!s.interp_azimuth && s.interp_elev)
			std::fill(_azimuth_buf.begin(), _azimuth_buf.end(),
			          _sourcepos[src].azimuth);
	}

	// If we're on the first source, we should overwrite the values already in
	// the output arrays, but if we're on higher numbered sources, we should
	// accumulate our results to the arrays. We do this to avoid having to clear
	// all arrays in a separate pass, which would be slower. We also avoid the
	// read and add operations for the first pass.

	// Encode 0th order component first. This should usually need no gaining at
	// all, so we check for unity gain to avoid wasting cycles.
	if (_matrix(src, 0) == 1.0) {
		if (src == 0)
			memcpy(outputs[0], &_source_buf[0], num*sizeof(float));
		else
			_math->VVAcc(outputs[0], &_source_buf[0], num);
	} else {	
		if (src == 0)
			_math->VSMul(outputs[0], &_source_buf[0], _matrix(0, 0), num);
		else
			_math->VSMulAcc(outputs[0], &_source_buf[0], _matrix(src, 0), num);
	}
	
	// In case we're crossfading NFC, save the crossfader state now, since it
	// will be recalled once per order in NFCProcessing()
	s.nfc_xfader.SaveState();

	for (unsigned order = 0; order < _order; ++order)  {
		_nfc_samples_to_recalc_temp = _nfc_samples_to_recalc;
		NFCProcessing(&_source_buf[0], &_nfc_source_buf[0], src, order, num);

		for (unsigned i = 0; i < _ambi.OrderComponents(order + 1); ++i) {
			// Check if we're interpolating azimuth or elevation
			if (s.interp_azimuth || s.interp_elev) {
				// We're interpolating, fill _temp_buf with encoding coefs...
				_ambi.EncodeComponent(_temp_buf, _azimuth_buf, _elev_buf,
				                      order + 1, i, num);
				// then multiply outputs by coefs
				if (src == 0) {
					_math->VVMul(outputs[comp], &_nfc_source_buf[0],
					             &_temp_buf[0], num);
				} else {
					_math->VVMulAcc(outputs[comp], &_nfc_source_buf[0],
					                &_temp_buf[0], num);
				}
			} else {
				// We're not interpolating, use matrix values
				if (src == 0) {
					_math->VSMul(outputs[comp], &_nfc_source_buf[0],
					             _matrix(0, comp), num);
				} else {
					_math->VSMulAcc(outputs[comp], &_nfc_source_buf[0],
					                 _matrix(src, comp), num);
				}
			}
			++comp;
		}
	}
}

void AmbisonicsEncoder::Process(float** inputs, float** outputs, unsigned num)
{
	for (size_t src = 0; src < _sources.size(); ++src)
		ProcessSource(inputs[src], outputs, src, num);

	// Set samples left to NFC filter recalculation to new value.
	_nfc_samples_to_recalc = _nfc_samples_to_recalc_temp;
}

