/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _ambisonicsencoder_h
#define _ambisonicsencoder_h

#include "Ambisonics.hpp"
#include "AudioProcessor.hpp"
#include "Defs.hpp"
#include "DelayLine.hpp"
#include "DSPTools.hpp"
#include "InterpolatedFunction.hpp"
#include "Matrix.hpp"
#include "NFCFilter.hpp"
#include "NFCHighpass.hpp"
#include "PositionInterpolator.hpp"
#include "PositionVector.hpp"
#include "SIMDMath.hpp"
#include <vector>

/**
 * \brief Implementation of an HOA Ambionics encoder.
 */
class AmbisonicsEncoder : public AudioProcessor {
	struct Source {
		float gain;
		bool planewave, delay_sim, distance_sim, nfc, nfc_highpass;
		bool interp_azimuth, interp_elev, interp_radius;
		DelayLine delay;
		std::vector<NFCFilter> nfcfilters;
		std::vector<NFCHighpass> nfclimiters;
		Crossfader delay_xfader, gain_xfader, nfc_xfader;
		Gainer gainer;
		Source(unsigned sr) : delay(sr, 0.5f) // Max 0.5 sec delay
		{
			gain = 1.f;
			planewave = false;
			delay_sim = distance_sim = nfc = true;
			nfc_highpass = false;
			interp_azimuth = interp_elev = interp_radius = false;
		}
	};

	/**
	 * Functor for calculating a 1/x curve limited with output limited to a
	 * given max amplitude.
	 */
	class SourceRadiusLimiter {
		float _maxdb, _t;
	public:
		SourceRadiusLimiter(float maxdb)
		{
			Max(maxdb);
		}

		void Max(float maxdb)
		{
			_maxdb = maxdb;
			_t = powf(10.f, -maxdb/20.f);
		}

		float operator()(float x)
		{
			const float t = x*_maxdb/5.f;
			const float w = _t*expf(-0.5f*t*t);
			return 1.f/sqrtf(x*x + w*w);
		}
	};

	/**
	 * Functor for soft-limiting a radius to a given minimum value.
	 */
	class NFCRadiusLimiter {
		float _mindist;
	public:
		NFCRadiusLimiter(float mindist)
		{
			Min(mindist);
		}

		void Min(float mindist)
		{
			_mindist = mindist;
		}

		float operator()(float x)
		{
			const float t = 0.99f*x/_mindist;
			const float w = _mindist*expf(-0.5f*t*t);
			return sqrtf(x*x + w*w);
		}
	};

	Matrix _matrix;
	PositionVector _sourcepos;
	std::vector<Source> _sources;
	// These will contain interpolated azimuth, elevation and radius values
	FloatVector _azimuth_buf, _elev_buf, _radius_buf;
	// This will contain the processed source samples, before and after NFC
	FloatVector _source_buf, _nfc_source_buf;
	// This will be used for temporary storage, Ambisonics coefs among others
	FloatVector _temp_buf;
	SIMDMathPtr _math;
	PositionInterpolator _interpolator;
	NFCRadiusLimiter _nfcradiuslimiter;
	SourceRadiusLimiter _srcradiuslimiter;
	double _max_radius;
	InterpolatedFunction<SourceRadiusLimiter> _amplimiter;
	unsigned _nfc_samples_to_recalc, _nfc_samples_to_recalc_temp;
	unsigned _nfc_recalc_delta;
	bool _is3d;
	unsigned _order;
	double _lsp_radius;
	Ambisonics _ambi;
	/**
	 * Updates an NFC filter set with the current source data.
	 * @param source source number.
	 */
	void UpdateNFCFilters(size_t source);
	/**
	 * Intialises and/or updates a source's state to reflect current settings.
	 * @param source source number.
	 */
	void InitSource(size_t source);
	/**
	 * Perform required source distance simulation processing on a given source,
	 * taking into account possible interpolation.
	 * @param input float array containing input source samples.
	 * @param output float array to store result of source distance simulation
	 * in. May not be the same as input.
	 * @param src source number.
	 * @param num number of samples to process.
	 */
	void SourceSimProcessing(float* input, float* output, size_t src,
	                         unsigned num);
	/**
	 * Performing NFC filtering on a given source, taking into account possible
	 * interpolation.
	 * @param input float array containing input source samples.
	 * @param output array to store result of filtering in. May not be an alias
	 * of input.
	 * @param order order for which to perform filtering, where 0 is the first
	 * order (0th order never needs filtering).
	 * @param num number of samples to filter.
	 */
	void NFCProcessing(float* input, float* output, size_t src, unsigned order,
	                   unsigned num);
	/**
	 * Perform Ambisonics encoding for a given source.
	 * @param input float array containing source samples to encode.
	 * @param outputs pointer to a number of float arrays to contain encoded
	 * samples. Can be the same as input.
	 * @param src source number to encode.
	 * @param num number of samples to encode.
	 */
	void ProcessSource(float* input, float** outputs, size_t src, unsigned num);
public:
	/**
	 * Class constructor.
	 * @param max_radius maximum allowable source radius, specified in meters.
	 */
	AmbisonicsEncoder(double max_radius = 50.);
	/**
	 * Set number of sources to encode. The PositionVector passed to the
	 * constructor must be correctly sized prior to calling this method.
	 * @param num new number of sources.
	 */
	void SetNumSources(size_t num);
	/**
	 * Set whether Ambisonics encoder should operate in 2D or 3D mode.
	 * @param is3d decides whether encoderis 3D or not.
	 */
	void SetIs3D(bool is3d);
	/**
	 * Set order of Ambisonics encoder. For 2D Ambisonics, order is limited to
	 * 15th order, since this is the highest order supported by the NFCFilter
	 * class. For 3D Ambisonics the limit is 5th order when interpolating
	 * sources, the reason being that 3D encoding equations has to be hardcoded
	 * in Ambisonics::EncodeComponent, and this has not been done for orders
	 * higher than this.
	 * @param order order of Ambisonics.
	 */
	void SetOrder(unsigned order);
	/**
	 * Set the maximum radius to be used for a source in encoding.
	 * @param max_radius maximum radius in meters.
	 */
	void SetMaxRadius(double max_radius);
	/**
	 * Set loudspeaker array radius.
	 * @param radius loudspeaker array radius.
	 */
	void SetLoudspeakerRadius(double radius);
	/**
	 * Set maximum source amplification factor which will occur when radius
	 * approaches zero. Transition from ordinary reciprocal distance attenuation
	 * to this maximum will be smooth.
	 * @param amp max source amplification in decibels.
	 */
	void SetMaxSourceAmplification(double amp);
	/**
	 * Set distance attenuation simulation on or off.
	 * @param source source number.
	 * @param enable distance attenuation simulation enable status.
	 */
	void SetDistanceSimulation(size_t source, bool enable);
	/**
	 * Set delay time simulation  on or off.
	 * @param source source number.
	 * @param enable delay simulation enable status.
	 */
	void SetDelaySimulation(size_t source, bool enable);
	/**
	 * Set plane wave status flag for a given source. When enabled, this will
	 * give a source unity gain and set the NFC filters to infinite source
	 * radius.
	 * @param source source number.
	 * @param source plane wave enable status.
	 */
	void SetPlaneWave(size_t source, bool enable);
	/**
	 * Set whether to enable near-field correction filters or not.
	 * @param source source number.
	 * @param enabled NFC filter enable status.
	 */
	void SetNFC(size_t source, bool enabled);
	/**
	 * Sets whether to enable highpass filtering of the NFC filters.
	 * @param source source number.
	 * @param enabled NFC filter highpass enable status.
	 */
	void SetNFCHighpass(size_t source, bool enabled);
	/**
	 * Set source position interpolation half-time. This value specifies at
	 * which point the interpolated position reaches halfway between the old
	 * position and the new. Note that this does not mean that the interpolation
	 * is done at double this time.
	 * @param interpolation half-time in seconds.
	 */
	void SetInterpolationHalfTime(double time);
	/**
	 * Set normalization scheme used for output format.
	 * @param norm normalization scheme to be used.
	 */
	void SetNormalization(Ambisonics::Normalization norm);
	/**
	 * Set source azimuth to value given in position vector. Value should be in
	 * radians.
	 * @param source source number.
	 * @param azimuth source azimuth in radians.
	 * @param interp whether to interpolate the source radius from current
	 * value to new value.
	 */
	void SetAzimuth(size_t source, double azimuth, bool interp = true);
	/**
	 * Set source elevation to value given in position vector. Value should be
	 * in radians.
	 * @param source source number.
	 * @param elevation source elevation in radians.
	 * @param interp whether to interpolate the source elevation from current
	 * value to new value.
	 */
	void SetElevation(size_t source, double elev, bool interp = true);
	/**
	 * Set source radius to value given in position vector. Value should be
	 * given in meters.
	 * @param source source number.
	 * @param radius source radius in meters.
	 * @param interp whether to interpolate the source radius from current
	 * value to new value.
	 */
	void SetRadius(size_t source, double radius, bool interp = true);
	/**
	 * Set source gain.
	 * @param source source number.
	 * @param gain source gain.
	 * @param interp whether to smooth the gain change.
	 */
	void SetGain(size_t source, float gain, bool interp = true);
	virtual void SetSampleRate(unsigned sr);
	virtual void SetBufferSize(unsigned size);
	virtual void Process(float** inputs, float** outputs, unsigned num);
	virtual size_t GetNumInputs() const
	{
		return _matrix.size1();
	}
	virtual size_t GetNumOutputs() const
	{
		return _matrix.size2();
	}
};

#include "Mutex.hpp"
#include <queue>
#include <boost/pool/pool_alloc.hpp>

/**
 * \brief Thread-safe version of AmbisonicsEncoder.
 * 
 * Use this class instead of AmbisonicsEncoder if Process() and the other
 * methods are not called from the same thread.
 * @see AmbisonicsEncoder
 */
class ThreadsafeAmbisonicsEncoder : public AmbisonicsEncoder {
	struct Param {
		enum Type { MaxRadius, LoudspeakerRadius, MaxSourceAmp, DistanceSim,
		            DelaySim, PlaneWave, NFC, NFCHighpass, Azimuth, Elevation,
		            Radius, InterpolationHalftime, Normalization
		};
		Type type;
		size_t index;
		bool flag;
		double value;
		Param(Type t) : type(t)
		{
		}
	};
	// Use pool allocation to ensure realtime-safe deallocation of objects 
	typedef std::deque< Param, boost::pool_allocator<Param> > ParamDeque;
	std::queue<Param, ParamDeque> _param_queue;
	Mutex _queue_mutex;
	void IndexParam(Param::Type t, size_t i, bool f = false, double v = 0.)
	{
		ScopedLock lock(_queue_mutex);
		Param p(t);
		p.index = i;
		p.flag = f;
		p.value = v;
		_param_queue.push(p);
	}

	void ValueParam(Param::Type t, double v)
	{
		ScopedLock lock(_queue_mutex);
		Param p(t);
		p.value = v;
		_param_queue.push(p);
	}
public:
	ThreadsafeAmbisonicsEncoder(double max_radius = 50.)
	: AmbisonicsEncoder(max_radius)
	{
	}

	// We handle all parameter setting by pushing a Param object onto a deque.
	// Process() will pop these objects off the deque, insepect them, and
	// execute a corresponding method in the AmbisonicsEncoder base class.
	// For this to make sense, only methods of low complexity can be called.
	// The rest will have to be called after locking the encoder object somehow.
	// This makes sure the encoder state is always consistent, even if accessed
	// by multiple threads.
	void SetMaxRadius(double max_radius)
	{
		ValueParam(Param::MaxRadius, max_radius);
	}

	void SetLoudspeakerRadius(double radius)
	{
		ValueParam(Param::LoudspeakerRadius, radius);
	}

	void SetMaxSourceAmplification(double amp)
	{
		ValueParam(Param::MaxSourceAmp, amp);
	}

	void SetDistanceSimulation(size_t source, bool enable)
	{
		IndexParam(Param::DistanceSim, source, enable);
	}

	void SetDelaySimulation(size_t source, bool enable)
	{
		IndexParam(Param::DelaySim, source, enable);
	}

	void SetPlaneWave(size_t source, bool enable)
	{
		IndexParam(Param::PlaneWave, source, enable);
	}

	void SetNFC(size_t source, bool enabled)
	{
		IndexParam(Param::NFC, source, enabled);
	}

	void SetNFCHighpass(size_t source, bool enabled)
	{
		IndexParam(Param::NFCHighpass, source, enabled);
	}

	void SetAzimuth(size_t source, double azimuth, bool interp = true)
	{
		IndexParam(Param::Azimuth, source, interp, azimuth);
	}

	void SetElevation(size_t source, double elev, bool interp = true)
	{
		IndexParam(Param::Elevation, source, interp, elev);
	}

	void SetRadius(size_t source, double radius, bool interp = true)
	{
		IndexParam(Param::Radius, source, interp, radius);
	}

	void SetInterpolationHalfTime(double time)
	{
		ValueParam(Param::InterpolationHalftime, time);
	}

	void SetNormalization(Ambisonics::Normalization norm)
	{
		IndexParam(Param::Normalization, norm);
	}

	void Process(float** inputs, float** outputs, unsigned num)
	{
		ScopedLock lock(_queue_mutex);
		while (!_param_queue.empty()) {
			Param& p = _param_queue.front();
			switch (p.type) {
			case Param::MaxRadius:
				AmbisonicsEncoder::SetMaxRadius(p.value);
				break;
			case Param::LoudspeakerRadius:
				AmbisonicsEncoder::SetLoudspeakerRadius(p.value);
				break;
			case Param::MaxSourceAmp:
				AmbisonicsEncoder::SetMaxSourceAmplification(p.value);
				break;
			case Param::DistanceSim:
				AmbisonicsEncoder::SetDistanceSimulation(p.index, p.flag);
				break;
			case Param::DelaySim:
				AmbisonicsEncoder::SetDelaySimulation(p.index, p.flag);
				break;
			case Param::PlaneWave:
				AmbisonicsEncoder::SetPlaneWave(p.index, p.flag);
				break;
			case Param::NFC:
				AmbisonicsEncoder::SetNFC(p.index, p.flag);
				break;
			case Param::NFCHighpass:
				AmbisonicsEncoder::SetNFCHighpass(p.index, p.flag);
				break;
			case Param::Azimuth:
				AmbisonicsEncoder::SetAzimuth(p.index, p.value, p.flag);
				break;
			case Param::Elevation:
				AmbisonicsEncoder::SetElevation(p.index, p.value, p.flag);
				break;
			case Param::Radius:
				AmbisonicsEncoder::SetRadius(p.index, p.value, p.flag);
				break;
			case Param::InterpolationHalftime:
				AmbisonicsEncoder::SetInterpolationHalfTime(p.value);
				break;
			case Param::Normalization:
				AmbisonicsEncoder::SetNormalization(Ambisonics::Normalization(p.index));
				break;
			default:
				printf("DEBUG: unhandled case in threadsafe Process()\n");
			}
			_param_queue.pop();
		}
		AmbisonicsEncoder::Process(inputs, outputs, num);
	}

	~ThreadsafeAmbisonicsEncoder()
	{
		boost::singleton_pool<boost::pool_allocator_tag,
		    sizeof(Param)>::release_memory();
	}
};

#endif

