/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <gtkmm/aboutdialog.h>
#include <gtkmm/box.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/table.h>
#include "AmbisonicsEncoderGUI.hpp"
#include "Widgets.hpp"

const Glib::ustring AmbisonicsEncoderGUI::_menu_info =
	"<ui>"
	"  <menubar name='MenuBar'>"
	"    <menu action='FileMenu'>"
	"      <menuitem action='FileOpen'/>"
	"      <menuitem action='FileSave'/>"
	"      <separator/>"
	"      <menuitem action='FileQuit'/>"
	"    </menu>"
	"    <menu action='ViewMenu'>"
	"      <menuitem action='ViewSources'/>"
	"      <menuitem action='ViewSettings'/>"
	"      <menuitem action='ViewLevelMeters'/>"
	"    </menu>"
	"    <menu action='RoutingMenu'>"
	"      <menuitem action='RoutingInputs'/>"
	"      <menuitem action='RoutingOutputs'/>"
	"    </menu>"
	"    <menu action='HelpMenu'>"
	"      <menuitem action='HelpAbout'/>"
	"    </menu>"
	"  </menubar>"
	"</ui>";
	
AmbisonicsEncoderGUI::AmbisonicsEncoderGUI(AmbisonicsEncoderModel& enc,
                                           JACKInterface& snd)
	: _adj_numsources(1., 1., 128., 1., 5.), _adj_order(1., 0., 15., 1., 3.),
	  _adj_lspdistance(2., 0.1, 20., 0.1, 1.),
	  _btn_numsources(_adj_numsources, 0., 0), _btn_order(_adj_order, 0., 0),
	  _btn_3d("3D"), _sld_lspdistance(_adj_lspdistance), _encoder(enc),
	  _wnd_sources(enc), _wnd_settings(enc), _wnd_meters(snd), _snd(snd)
{
	set_title("Ambisonics encoder");
	set_default_size(400, 200);

	// Make containers
	Gtk::Table* table = Gtk::manage(new Gtk::Table(4, 2));
	Gtk::VBox* vbox1 = Gtk::manage(new Gtk::VBox());
	// Construct our menu. This corresponds to the XML string at the top.
	_actiongroup = Gtk::ActionGroup::create();
	// File menu
	_actiongroup->add(Gtk::Action::create("FileMenu", "_File"));
	_actiongroup->add(Gtk::Action::create(
		"FileOpen", Gtk::Stock::OPEN, "_Open", "Open an encoder setup"),
		sigc::mem_fun(*this, &AmbisonicsEncoderGUI::on_open));
	_actiongroup->add(Gtk::Action::create(
		"FileSave", Gtk::Stock::SAVE, "_Save", "Save an encoder setup"),
		sigc::mem_fun(*this, &AmbisonicsEncoderGUI::on_save));
	_actiongroup->add(Gtk::Action::create(
		"FileQuit", Gtk::Stock::QUIT, "_Quit", "Quit"),
		sigc::mem_fun(*this, &AmbisonicsEncoderGUI::on_quit));

	// View menu
	_actiongroup->add(Gtk::Action::create("ViewMenu", "_View"));
	_act_sources = Gtk::ToggleAction::create(
		"ViewSources", "_Sources", "View sources");
	_act_settings = Gtk::ToggleAction::create(
		"ViewSettings", "S_ettings", "View settings");
	_act_meters = Gtk::ToggleAction::create(
		"ViewLevelMeters", "_Level meters", "View level meters");
	_actiongroup->add(_act_sources, Gtk::AccelKey("<control>U"),
		sigc::mem_fun(*this, &AmbisonicsEncoderGUI::on_view_sources));
	_actiongroup->add(_act_settings, Gtk::AccelKey("<control>E"),
		sigc::mem_fun(*this, &AmbisonicsEncoderGUI::on_view_settings));
	_actiongroup->add(_act_meters, Gtk::AccelKey("<control>M"),
		sigc::mem_fun(*this, &AmbisonicsEncoderGUI::on_view_meters));

	// Routing menu
	_actiongroup->add(Gtk::Action::create("RoutingMenu", "_Routing"));
	_actiongroup->add(Gtk::Action::create(
		"RoutingInputs", "Auto-connect _inputs", "Auto-connect JACK inputs"),
		sigc::mem_fun(*this, &AmbisonicsEncoderGUI::on_route_inputs));
	_actiongroup->add(Gtk::Action::create(
		"RoutingOutputs", "Auto-connect _outputs", "Auto-connect JACK outputs"),
		sigc::mem_fun(*this, &AmbisonicsEncoderGUI::on_route_outputs));

	// Help menu
	_actiongroup->add(Gtk::Action::create("HelpMenu", "_Help"));
	_actiongroup->add(Gtk::Action::create(
		"HelpAbout", "_About", "About program"),
		sigc::mem_fun(*this, &AmbisonicsEncoderGUI::on_view_aboutbox));

	_uimanager = Gtk::UIManager::create();
	_uimanager->insert_action_group(_actiongroup);
	add_accel_group(_uimanager->get_accel_group());
	_uimanager->add_ui_from_string(_menu_info);
	
	Gtk::Widget* menu = _uimanager->get_widget("/MenuBar");
	vbox1->pack_start(*menu, Gtk::PACK_SHRINK);

	// Create widget table
	table->set_col_spacings(20);
	table->set_row_spacings(6);
	table->set_border_width(16);
	Gtk::Label* src_label = Gtk::manage(new Gtk::Label("Sources:"));
	src_label->set_alignment(Gtk::ALIGN_RIGHT);
	Gtk::Label* order_label = Gtk::manage(new Gtk::Label("Order:"));
	order_label->set_alignment(Gtk::ALIGN_RIGHT);
	Gtk::Label* lspdist_label = Gtk::manage(
		new Gtk::Label("Loudspeaker\ndistance:"));
	lspdist_label->set_alignment(Gtk::ALIGN_RIGHT);
	table->attach(*src_label, 0, 1, 0, 1, Gtk::FILL);
	table->attach(*order_label, 0, 1, 1, 2, Gtk::FILL);
	table->attach(_btn_numsources, 1, 2, 0, 1, Gtk::EXPAND);
	table->attach(_btn_order, 1, 2, 1, 2, Gtk::EXPAND);
	table->attach(_btn_3d, 1, 2, 2, 3, Gtk::EXPAND);
	table->attach(*lspdist_label, 0, 1, 3, 4, Gtk::FILL);
	table->attach(_sld_lspdistance, 1, 2, 3, 4);	
	vbox1->pack_start(*table);

	add(*vbox1);

	// Connect signal handlers
	_btn_numsources.signal_value_changed().connect(sigc::mem_fun(
		*this, &AmbisonicsEncoderGUI::on_numsources_changed));
	_btn_order.signal_value_changed().connect(sigc::mem_fun(
		*this, &AmbisonicsEncoderGUI::on_order_changed));
	_btn_3d.signal_clicked().connect(sigc::mem_fun(
		*this, &AmbisonicsEncoderGUI::on_3d_clicked));
	_sld_lspdistance.signal_value_changed().connect(sigc::mem_fun(
		*this, &AmbisonicsEncoderGUI::on_lsp_distance_changed));

	_wnd_sources.signal_delete_event().connect(sigc::mem_fun(
			*this, &AmbisonicsEncoderGUI::on_hide_sources));
	_wnd_settings.signal_delete_event().connect(sigc::mem_fun(
			*this, &AmbisonicsEncoderGUI::on_hide_settings));
	_wnd_meters.signal_delete_event().connect(sigc::mem_fun(
		*this, &AmbisonicsEncoderGUI::on_hide_meters));

	_encoder.AddView(this);
	_encoder.AddView(&_wnd_sources);
	_encoder.AddView(&_wnd_settings);
	_encoder.AddView(&_wnd_meters);
	show_all_children();
}

void AmbisonicsEncoderGUI::Notify(int what)
{
	switch (what) {
	case AmbisonicsEncoderModel::ChangedNumSources:
		_btn_numsources.set_value(_encoder.NumSources());
		break;
	case AmbisonicsEncoderModel::ChangedOrder:
		_btn_order.set_value(_encoder.Order());
		break;
	case AmbisonicsEncoderModel::ChangedIs3D:
		// Limit order spinbutton to 5 if we've got 3D operation
		_btn_order.set_range(0, _encoder.Is3D() ? 5 : 15);
		_btn_3d.set_active(_encoder.Is3D());
		break;
	case AmbisonicsEncoderModel::ChangedLoudspeakerRadius:
		_sld_lspdistance.set_value(_encoder.LoudspeakerRadius());
		break;
	case AmbisonicsEncoderModel::ChangedNormalization:
		if (_encoder.Normalization() == Ambisonics::FurseMalham)
			_btn_order.set_range(0, 3);
		else
			_btn_order.set_range(0, _encoder.Is3D() ? 5 : 15);
		break;
	case AmbisonicsEncoderModel::ChangedAll:
		_btn_numsources.set_value(_encoder.NumSources());
		_btn_order.set_value(_encoder.Order());
		_btn_3d.set_active(_encoder.Is3D());
		_sld_lspdistance.set_value(_encoder.LoudspeakerRadius());
		break;
	}
}

void AmbisonicsEncoderGUI::on_numsources_changed()
{
	_encoder.NumSources(_btn_numsources.get_value_as_int());
}

void AmbisonicsEncoderGUI::on_order_changed()
{
	_encoder.Order(_btn_order.get_value_as_int());
}

void AmbisonicsEncoderGUI::on_3d_clicked()
{
	_encoder.Is3D(_btn_3d.get_active());
}

void AmbisonicsEncoderGUI::on_lsp_distance_changed()
{
	_encoder.LoudspeakerRadius(_sld_lspdistance.get_value());
}

void AmbisonicsEncoderGUI::on_open()
{
	Gtk::FileChooserDialog dlg("Open setup", Gtk::FILE_CHOOSER_ACTION_OPEN);
	Gtk::FileFilter ambienc_filter, all_filter;

	dlg.set_transient_for(*this);
	dlg.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	dlg.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);
	ambienc_filter.set_name("Ambienc files");
	ambienc_filter.add_pattern("*.ambienc");
	all_filter.set_name("All files");
	all_filter.add_pattern("*");
	dlg.add_filter(ambienc_filter);
	dlg.add_filter(all_filter);
	int result = dlg.run();
	if (result == Gtk::RESPONSE_OK) {
		std::string errmsg;

		if (!_encoder.Load(dlg.get_filename(), &errmsg)) {
			Gtk::MessageDialog msg(dlg, errmsg, false, Gtk::MESSAGE_ERROR,
			                       Gtk::BUTTONS_OK, true);

			msg.run();
		}
	}
}

void AmbisonicsEncoderGUI::on_save()
{
	Gtk::FileChooserDialog dlg("Save setup", Gtk::FILE_CHOOSER_ACTION_SAVE);
	Gtk::FileFilter ambienc_filter;
	
	dlg.set_transient_for(*this);
	dlg.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	dlg.add_button(Gtk::Stock::SAVE, Gtk::RESPONSE_OK);
	ambienc_filter.set_name("Ambienc files");
	ambienc_filter.add_pattern("*.ambienc");
//	dlg.set_do_overwrite_confirmation();
	dlg.add_filter(ambienc_filter);
	int result = dlg.run();
	if (result == Gtk::RESPONSE_OK) {
		Glib::ustring fn = dlg.get_filename();
		std::string errmsg;
		// Check if user has specified an extension
		if (fn.find('.') == Glib::ustring::npos)
			// If not, append the default
			fn += ".ambienc";
		
		if (!_encoder.Save(fn, &errmsg)) {
			Gtk::MessageDialog msg(dlg, errmsg, false, Gtk::MESSAGE_ERROR,
			                        Gtk::BUTTONS_OK, true);

			msg.run();
		}
	}
}

void AmbisonicsEncoderGUI::on_quit()
{
	hide();
}

void AmbisonicsEncoderGUI::on_view_sources()
{
	if (_act_sources->get_active())
		_wnd_sources.show();
	else
		_wnd_sources.hide();
}

bool AmbisonicsEncoderGUI::on_hide_sources(GdkEventAny* e)
{
	_act_sources->set_active(false);
	return false; // allow window to close
}

void AmbisonicsEncoderGUI::on_view_settings()
{
	if (_act_settings->get_active())
		_wnd_settings.show();
	else
		_wnd_settings.hide();
}

bool AmbisonicsEncoderGUI::on_hide_settings(GdkEventAny* e)
{
	_act_settings->set_active(false);
	return false; // allow window to close
}

void AmbisonicsEncoderGUI::on_view_meters()
{
	if (_act_meters->get_active())
		_wnd_meters.show();
	else
		_wnd_meters.hide();
}

bool AmbisonicsEncoderGUI::on_hide_meters(GdkEventAny* e)
{
	_act_meters->set_active(false);
	return false; // Allow window to close
}

#include "LicenseText.hpp"

void AmbisonicsEncoderGUI::on_view_aboutbox()
{
	Gtk::AboutDialog dlg;
	std::vector<Glib::ustring> authors;

	dlg.set_skip_taskbar_hint();
	authors.push_back("Thom Johansen");
	authors.push_back("Bård støfringsdal");
	authors.push_back("Peter Svensson");
	dlg.set_comments("An HOA Ambisonics encoder for JACK");
	dlg.set_version("0.2");
	dlg.set_authors(authors);
	dlg.set_copyright("Copyright © 2006-2007 Q2S, "
	                  "Norwegian University of Science and Technology");
	dlg.set_website("http://www.q2s.ntnu.no");
	dlg.set_license(license_text);
	dlg.run();
}

void AmbisonicsEncoderGUI::on_route_inputs()
{
	const char* str[] = { "alsa_pcm:capture_" };
	StringInputDialog input("Target ports", str, sizeof(str)/sizeof(char*));

	if (input.Run())
		_snd.ConnectInputs(input.GetString());
}	

void AmbisonicsEncoderGUI::on_route_outputs()
{
	const char* str[] = { "ambidec:input_", "ldc:input_",
	                      "alsa_pcm:playback_" };
	StringInputDialog input("Target ports", str, sizeof(str)/sizeof(char*));

	if (input.Run())
		_snd.ConnectOutputs(input.GetString());
}

