/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _ambisonicsencodermodel_h
#define _ambisonicsencodermodel_h

#include <cmath>
#include <fstream>
#include <vector>
#include <string>
#include "AmbisonicsEncoder.hpp"
#include "ConfigFile.hpp"
#include "AudioProcessor.hpp"
#include "MIDIReceiver.hpp"
#include "Model.hpp"
#include "PositionVector.hpp"

class AmbisonicsEncoderModel : public Model, public MIDIReceiver {
	struct SourceData {
		double gain;
		bool planewave;
		SourceData() : gain(0.f), planewave(false)
		{
		}
	};

	PositionVector _sources;
	ThreadsafeAmbisonicsEncoder _enc;
	AudioInterface& _snd;
	std::vector< std::vector<char> > _ctrlvalues;
	std::vector<SourceData> _sourcedata;
	bool _3d;
	unsigned _order;
	double _lsp_radius;
	bool _nfc, _distance_sim, _delay_sim;
	bool _nfc_highpass;
	double _max_source_amp, _interphalftime;
	Ambisonics::Normalization _normalization;

	bool _seen_lsb_controller;

	void UpdateEncoder()
	{
		AudioInterface::Disable lock(_snd);
		_enc.SetNumSources(_sources.size());
		_enc.SetLoudspeakerRadius(_lsp_radius);
		_enc.SetOrder(_order);
		_enc.SetIs3D(_3d);
		for (size_t i = 0; i < _sources.size(); ++i) {
			// Tell encoder to not interpolate as we're initialising
			_enc.SetAzimuth(i, _sources[i].azimuth*M_PI/180., false);
			_enc.SetElevation(i, _sources[i].elevation*M_PI/180., false);
			_enc.SetRadius(i, _sources[i].radius, false);
			_enc.SetGain(i, pow(10., _sourcedata[i].gain/20.), false);
			_enc.SetPlaneWave(i, _sourcedata[i].planewave);
			_enc.SetDistanceSimulation(i, _distance_sim);
			_enc.SetDelaySimulation(i, _delay_sim);
			_enc.SetNFC(i, _nfc);
			_enc.SetNFCHighpass(i, _nfc_highpass);
		}
		_enc.SetMaxSourceAmplification(_max_source_amp);
		_enc.SetInterpolationHalfTime(_interphalftime);
		_enc.SetNormalization(_normalization);
		_snd.UpdatePorts();
	}
public:
	AudioProcessor* GetAudioProcessor()
	{
		return &_enc;
	}
	enum {
		ChangedAll = 0, ChangedIs3D, ChangedOrder, ChangedNumSources,
		ChangedLoudspeakerRadius, ChangedDistanceSimulation,
		ChangedDelaySimulation, ChangedNFC, ChangedPlaneWave,
		ChangedNFCHighpass, ChangedSourcePosition, ChangedSourceGain,
		ChangedMaxSourceAmplification, ChangedInterpolationHalfTime,
		ChangedNormalization
	};
	AmbisonicsEncoderModel(AudioInterface& ai)
		: _enc(50.), _snd(ai),
		  _ctrlvalues(5, std::vector<char>(64, 0)), _seen_lsb_controller(false)
	{
		_snd.SetProcess(_enc);
		DefaultSettings();
		UpdateEncoder();
		_snd.EnableProcessing();
	}

	void DefaultSettings()
	{
		_3d = false;
		_order = 1;
		_sources.resize(1);
		_sourcedata.resize(1);
		_lsp_radius = 2.;
		_nfc = true;
		_distance_sim = true;
		_delay_sim = true;
		_nfc_highpass = false;
		_max_source_amp = 12.;
		_interphalftime = 1.;
		_normalization = Ambisonics::Normalized;
	}

	void Is3D(bool is3d)
	{
		bool changed_order = false;

		{
			AudioInterface::Disable lock(_snd);
			if (is3d == _3d) return;
			_3d = is3d;
			// Prevent using higher orders than 5 if 3D.
			if (is3d && _order > 5) {
				_order = 5;
				_enc.SetOrder(5);
				changed_order = true;
			}
			_enc.SetIs3D(is3d);
			_snd.UpdatePorts(); // Number of outputs has changed
		}
		NotifyViews(ChangedIs3D);
		if (changed_order)
			NotifyViews(ChangedOrder);
	}
	
	bool Is3D() const
	{
		return _3d;
	}
	
	void Order(unsigned order)
	{
		{
			AudioInterface::Disable lock(_snd);
			if (order == _order) return;
			if (_3d && order > 5)
				order = 5;
			_order = order;
			_enc.SetOrder(order);
			_snd.UpdatePorts(); // Number of outputs has changed
		}
		NotifyViews(ChangedOrder);
	}
	
	unsigned Order() const
	{
		return _order;
	}

	void NumSources(size_t numsources)
	{
		const size_t old_size = _sources.size();
		{
			AudioInterface::Disable lock(_snd);
			if (numsources == _sources.size()) return;
			_sources.resize(numsources);
			_sourcedata.resize(numsources);
			_enc.SetNumSources(numsources);
			// Set new source settings
			for (size_t i = old_size; i < _sources.size(); ++i) {
				_enc.SetAzimuth(i, _sources[i].azimuth, false);
				_enc.SetElevation(i, _sources[i].elevation, false);
				_enc.SetRadius(i, _sources[i].radius, false);
				_enc.SetDistanceSimulation(i, _distance_sim);
				_enc.SetDelaySimulation(i, _delay_sim);
				_enc.SetNFC(i, _nfc);
				_enc.SetNFCHighpass(i, _nfc_highpass);
			}
			_snd.UpdatePorts(); // Number of inputs changed
		}
		NotifyViews(ChangedNumSources);
	}
	
	size_t NumSources() const
	{
		return _sources.size();
	}

	double MaxRadius() const
	{
		// Hard coded for now, in case we want to add support for changing it.
		return 50.;
	}

	void LoudspeakerRadius(double radius)
	{
		if (radius == _lsp_radius) return;
		_lsp_radius = radius;
		_enc.SetLoudspeakerRadius(radius);
		NotifyViews(ChangedLoudspeakerRadius);
	}
	
	double LoudspeakerRadius() const
	{
		return _lsp_radius;
	}

	void DistanceSimulation(bool distsim)
	{
		if (distsim == _distance_sim) return;
		_distance_sim = distsim;
		for (size_t i = 0; i < _sources.size(); ++i)
			_enc.SetDistanceSimulation(i, distsim);
		NotifyViews(ChangedDistanceSimulation);
	}
	
	bool DistanceSimulation() const
	{
		return _distance_sim;
	}

	void DelaySimulation(bool delaysim)
	{
		if (delaysim == _delay_sim) return;
		_delay_sim = delaysim;
		for (size_t i = 0; i < _sources.size(); ++i)
			_enc.SetDelaySimulation(i, delaysim);
		NotifyViews(ChangedDelaySimulation);
	}

	bool DelaySimulation() const
	{
		return _delay_sim;
	}

	void NFC(bool nfc)
	{
		if (nfc == _nfc) return;
		_nfc = nfc;
		for (size_t i = 0; i < _sources.size(); ++i)
			_enc.SetNFC(i, nfc);
		NotifyViews(ChangedNFC);
	}

	bool NFC() const
	{
		return _nfc;
	}

	void NFCHighpass(bool nfc_highpass)
	{
		if (nfc_highpass == _nfc_highpass) return;
		_nfc_highpass = nfc_highpass;
		for (size_t i = 0; i < _sources.size(); ++i)
			_enc.SetNFCHighpass(i, nfc_highpass);
		NotifyViews(ChangedNFCHighpass);
	}

	bool NFCHighpass() const
	{
		return _nfc_highpass;
	}

	void MaxSourceAmplification(double amp)
	{
		if (amp == _max_source_amp) return;
		_max_source_amp = amp;
		_enc.SetMaxSourceAmplification(amp);
		NotifyViews(ChangedMaxSourceAmplification);
	}

	double MaxSourceAmplification() const
	{
		return _max_source_amp;
	}

	void PlaneWave(size_t src, bool pw)
	{
		if (pw == _sourcedata[src].planewave) return;
		_sourcedata[src].planewave = pw;
		_enc.SetPlaneWave(src, pw);
		NotifyViews(ChangedPlaneWave);
	}

	bool PlaneWave(size_t src) const
	{
		return _sourcedata[src].planewave;
	}
	
	void InterpolationHalfTime(double time)
	{
		if (time == _interphalftime) return;
		_interphalftime = time;
		_enc.SetInterpolationHalfTime(time);
		NotifyViews(ChangedInterpolationHalfTime);
	}

	double InterpolationHalfTime() const
	{
		return _interphalftime;
	}

	void Normalization(Ambisonics::Normalization norm)
	{
		if (norm == _normalization) return;
		_normalization = norm;
		if (norm == Ambisonics::FurseMalham && _order > 3) {
			_order = 3;
			_enc.SetOrder(3);
			NotifyViews(ChangedOrder);
		}
		_enc.SetNormalization(norm);
		NotifyViews(ChangedNormalization);
	}

	Ambisonics::Normalization Normalization() const
	{
		return _normalization;
	}

	void SourceAzimuth(size_t src, double azimuth)
	{
		if (_sources[src].azimuth == azimuth) return;
		_sources[src].azimuth = azimuth;
		_enc.SetAzimuth(src, azimuth*M_PI/180.);
		NotifyViews(ChangedSourcePosition);
	}

	double SourceAzimuth(size_t src) const
	{
		return _sources[src].azimuth;
	}

	void SourceElevation(size_t src, double elev)
	{
		if (_sources[src].elevation == elev) return;
		_sources[src].elevation = elev;
		_enc.SetElevation(src, elev*M_PI/180.);
		NotifyViews(ChangedSourcePosition);
	}

	double SourceElevation(size_t src) const
	{
		return _sources[src].elevation;
	}

	void SourceRadius(size_t src, double radius)
	{
		if (_sources[src].radius == radius) return;
		_sources[src].radius = radius;
		_enc.SetRadius(src, radius);
		NotifyViews(ChangedSourcePosition);
	}

	double SourceRadius(size_t src) const
	{
		return _sources[src].radius;
	}

	void SourceGain(size_t src, double gain)
	{
		if (_sourcedata[src].gain == gain) return;
		_sourcedata[src].gain = gain;
		_enc.SetGain(src, pow(10., gain/20.));
		NotifyViews(ChangedSourceGain);
	}

	double SourceGain(size_t src) const
	{
		return _sourcedata[src].gain;
	}

	bool Save(std::string filename, std::string* errmsg = 0)
	{
		try {
			ConfigWriter c(filename);

			c.SetItem("ambienc_cfg_version", 2);
			c.SetItem("order", _order);
			c.SetItem("3d", _3d);
			c.SetItem("lsp_array_radius", _lsp_radius);
			c.SetItem("nfc_enabled", _nfc);
			c.SetItem("nfc_highpass_enabled", _nfc_highpass);
			c.SetItem("distance_sim_enabled", _distance_sim);
			c.SetItem("delay_sim_enabled", _delay_sim);
			c.SetItem("max_source_amplification", _max_source_amp);
			c.SetItem("interpolation_halfpoint_time", _interphalftime);
			c.SetItem("normalization", _normalization);
			c.Comment("order is \"azimuth, elevation, radius, planewave, gain\"");
			c.StartMatrixItem("sources", 5, _sources.size());
			for (size_t i = 0; i < _sources.size(); ++i) {
				c.SetItemValue(_sources[i].azimuth);
				c.SetItemValue(_sources[i].elevation);
				c.SetItemValue(_sources[i].radius);
				c.SetItemValue(_sourcedata[i].planewave);
				c.SetItemValue(_sourcedata[i].gain);
			}
		}
		catch (std::runtime_error& e) {
			if (errmsg)
				errmsg->assign(e.what());
			return false;
		}
		return true;
	}

	bool Load(std::string filename, std::string* errmsg = 0)
	{
		try {
			int t;
			ConfigReader c(filename, true);

			c.GetValue("ambienc_cfg_version", t);
			if (t != 2) {
				if (errmsg)
					errmsg->assign("Incompatible configuration file version");
				return false;
			}
			c.GetValue("order", _order);
			c.GetValue("3d", _3d);
			c.GetValue("lsp_array_radius", _lsp_radius);
			c.GetValue("nfc_enabled", _nfc);
			c.GetValue("nfc_highpass_enabled", _nfc_highpass);
			c.GetValue("distance_sim_enabled", _distance_sim);
			c.GetValue("delay_sim_enabled", _delay_sim);
			c.GetValue("max_source_amplification", _max_source_amp);
			c.GetValue("interpolation_halfpoint_time", _interphalftime);
			c.GetValue("normalization", t);
			_normalization = Ambisonics::Normalization(t);
			ConfigReader::Item* item = c.GetItem("sources");
			_sources.resize(item->ColSize(1));
			_sourcedata.resize(item->ColSize(1));
			for (unsigned i = 0; i < _sources.size(); ++i) {
				c.GetValue("sources", _sources[i].azimuth, 1, i + 1);
				c.GetValue("sources", _sources[i].elevation, 2, i + 1);
				c.GetValue("sources", _sources[i].radius, 3, i + 1);
				c.GetValue("sources", _sourcedata[i].planewave, 4, i + 1);
				c.GetValue("sources", _sourcedata[i].gain, 5, i + 1);
			}
		}
		catch (std::runtime_error& e) {
			if (errmsg)
				errmsg->assign(e.what());
			// State might be inconsistent now, reset to a good default
			DefaultSettings();
			UpdateEncoder();
			NotifyViews();
			return false;
		}
		UpdateEncoder();
		NotifyViews();
		return true;
	}

	void MIDIMessage(MessageType msg, char chan, char num, char val)
	{
		double normval; // Controller value in the range 0 to 1
		const size_t source_num = num & 0x1f;

		// We only want controller messages
		if (msg != Controller)
			return;
		// Make sure controller number is lower than number of sources and
		// channel number lower than number of parameters controllable.
		if (source_num >= NumSources() || chan > 4)
			return;
		// Check if controller number is in our range at all
		if (num > 63)
			return;

		_ctrlvalues[chan][num] = val;
		// Check if we received high or low part of double precision ctrl. value
		if (num > 31) {
			_seen_lsb_controller = true; // Flag 14 bit MIDI mode
			normval = ((_ctrlvalues[chan][num - 32] << 7) | val)/16383.;
		} else {
			// If we haven't seen an LSB controller message yet, we assume we're
			// being used in ordinary 7 bit MIDI controller mode.
			if (_seen_lsb_controller)
				normval = ((val << 7) | (_ctrlvalues[chan][num + 32]))/16383.;
			else
				normval = val/127.;
		}

		switch (chan) {
		case 0: // Azimuth
			SourceAzimuth(source_num, -180. + normval*360.); // -180 dg .. 180 dg
			break;
		case 1: // Elevation
			SourceElevation(source_num, -90. + normval*180); // -90 dg .. 90 dg
			break;
		case 2: // Radius 
			SourceRadius(source_num, 0.1 + normval*49.9); // 0.1 m .. 50 m
			break;
		case 3: // Plane wave switch
			PlaneWave(source_num, normval > 0.5);
			break;
		case 4: // Gain
			SourceGain(source_num, -60. + normval*72.); // -60 dB .. 12 dB
			break;
		}
	}
};

#endif

