/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _audioprocessor_h
#define _audioprocessor_h

#include "Mutex.hpp"
#include <cstddef> 

/**
 * \brief Common interface class for an audio driver or host to communicate with
 * a class capable of processing floating point audio in blocks. 
 *
 * It provides an interface for calling the processor with arrays
 * of input and output data, setting the sample rate and buffer size of the
 * audio host. 
 * If an audio processor needs notification of audio events, it should override
 * the members of this class that are needed.
 */
class AudioProcessor
{
protected:
	unsigned _sample_rate, _buffer_size;
public:
	/**
	 * Constructs an audio processor.
	 */
	AudioProcessor() : _sample_rate(0)
	{
	}
	/**
	 * This method notifies the audio processor that a change in sample rate
	 * has happened in the host object.
	 * @param sr sample rate.
	 */
	virtual void SetSampleRate(unsigned sr)
	{
		_sample_rate = sr;
	}
	/**
	 * This method notifies the audio processor that a change in buffer size
	 * has happened in the host object. This is usable if an audio processor
	 * uses block based processing and needs to allocate internal buffers for
	 * internal processing.
	 * @param size buffer size in samples.
	 */
	virtual void SetBufferSize(unsigned size)
	{
		_buffer_size = size;
	}
	/**
	 * This method must be overriden to provide the host with information as
	 * to how many input channels it expects to be passed.
	 * @return Number of input channels audio processor expects to be passed.
	 */
	virtual size_t GetNumInputs() const = 0;
	/**
	 * This method must be overriden to provide the host with information as
	 * to how many output channels it expects to be passed.
	 * @return Number of output channels audio processor expects to be passed
	 */
	virtual size_t GetNumOutputs() const = 0;
	/**
	 * This method will be called by the host whenever the audio processor
	 * is to process data. The number of input and output arrays provided
	 * are decided by the values the audio processor class returns with the
	 * GetNumInputs() and GetNumOutputs() methods.
	 * @param inputs array of float pointers pointing to input data.
	 * @param outputs array of float pointers pointing to output data.
	 * @param num number of samples to be processed.
	 */
	virtual void Process(float** inputs, float** outputs, unsigned num) = 0;
	/**
	 * Class destructor.
	 */
	virtual ~AudioProcessor()
	{
	}
};

class AudioInterface
{
protected:
	AudioProcessor* _process;
	Mutex _process_mutex;
public:
	/**
	 * \brief Convenience class used to disable processing in an audio
	 * interface for the current scope.
	 *
	 * @see AudioInterface
	 */
	class Disable
	{
	private:
		AudioInterface& _aif;
	public:
		/**
		 * Class constructor. This will disable audio processing until end of
		 * current scope.
		 * @param aif reference to a AudioInterface object.
		 */
		Disable(AudioInterface& aif) : _aif(aif)
		{
			_aif.DisableProcessing();
		}
		/**
		 * Class destructor. This reenables audio processing.
		 */
		~Disable()
		{
			_aif.EnableProcessing();
		}
	};

	/**
	 * Class constructor. Throws exceptions of type std::runtime_error when
	 * errors occur.
	 */
	AudioInterface() : _process(0), _process_mutex(Mutex::Recursive)
	{
//		DisableProcessing();
	}
	/**
	 * Calls class set with SetProcess() and queries number of inputs and
	 * outputs, then updates number of input and output port if needed.
	 */
	virtual void UpdatePorts() = 0;
	/**
	 * Sets AudioProcessor object that will process all requests for audio.
	 * @param o reference to an AudioProcessor object.
	 */
	virtual void SetProcess(AudioProcessor& o) = 0;
	/**
	 * Get the maximum peak value for a given channel during the last frame
	 * processed.
	 * @param chan channel number.
	 * @param peak reference to float variable where peak value is stored.
	 * @param output true for output channel, false for input channel.
	 * @return false if no new audio frame has occured since last peak readout.
	 */
	virtual bool GetPeak(size_t chan, float& peak, bool output = true) = 0;
	/**
	 * Tell the audio driver if it should bother scanning for peaks or not.
	 * Disabling this should save some processing power.
	 * @param enable true to enable peak scanning.
	 */
	virtual void FindPeaks(bool enable = true) = 0;
	/**
	 * Turn off all forwarding of incoming audio process requests. Use this
	 * when the AudioProcessor object is in a volatile/changing state and can
	 * not be called without risk of crashing.
	 * @see EnableProcessing
	 */
	void DisableProcessing()
	{
		_process_mutex.Lock();
	}
	/**
	 * Enable forwarding of incoming audio process requests after a call to
	 * DisableProcessing(). This method must be called once in every program
	 * to start the initial processing.
	 * @see DisableProcessing
	 */
	void EnableProcessing()
	{
		_process_mutex.Unlock();
	}	
	/**
	 * Get number of input ports for current audio processor object assigned to
	 * this object.
	 * @return number of input ports.
	 */
	size_t GetNumInputs() const
	{
		if (_process)
			return _process->GetNumInputs();
		else
			return 0;
	}
	/**
	 * Get number of output ports for current audio processor object assigned to
	 * this object.
	 * @return number of output ports.
	 */
	size_t GetNumOutputs() const
	{
		if (_process)
			return _process->GetNumOutputs();
		else
			return 0;
	}
	/**
	 * Class destructor.
	 */
	virtual ~AudioInterface()
	{
	}
};

#endif

