/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "ButterworthFilter.hpp"

std::vector<double> ButterworthFilter::_designcoefs;
unsigned ButterworthFilter::_precalcorder = 0;

void ButterworthFilter::CalcCoefTable(unsigned order)
{
	size_t cursize = _designcoefs.size();
	size_t newsize = ((order + 1)/2)*(order/2);

	_designcoefs.resize(newsize);
	// Make coef sets for each order we haven't, up to the specified order.
	for (size_t curorder = _precalcorder + 1; curorder <= order; ++curorder) {
		for (size_t k = 1; k <= curorder/2; ++k)
			_designcoefs[cursize++] =
				-cos((2*k + curorder - 1)/double(curorder)*M_PI/2.);
	}
	_precalcorder = order;
}

void ButterworthFilter::SetOrder(unsigned order)
{
	// Check if we have pre-calculated the coefs we need up to this order.
	if (_precalcorder < order)
		CalcCoefTable(order);
	_num_second_order_sec = order/2;
	_first_order_sec = order & 1 ? true : false;
	_coef.resize(_num_second_order_sec*5 + (order & 1)*3);
	ResizeHistory();
	_order = order;
}

void ButterworthFilter::SetCutoff(double cutoff)
{
	std::vector<CoefType>::iterator c = _coef.begin();
	double* dcoefs = &_designcoefs[(_order/2)*((_order - 1)/2)];
	const double x = cutoff*_cutoff_mult;
	// Third order taylor approximation of tan(), useful for cutoffs up to
	// around 0.75*Nyquist.
	//const double w = x*(x*x*(2./15.*x*x + 1./3.) + 1.);
	// Exact cutoff calculation
	const double w = tan(x);
	double b0, b1, b2;

	// The FIR part of the filter is fixed, calculate it here.
	switch (_type) {
	case Lowpass:
		b0 = b2 = w*w;
		b1 = b0 + b0; // 2*w*w
		break;
	default:
	case Highpass:
		b0 = b2 = 1.;
		b1 = -2.;
	}

	// Calculate the second order sections
	for (size_t i = 0; i < _order/2; ++i) {
		const double coef = dcoefs[i];
		double rcp_a0 = 1./(1. + w*(2.*coef + w));
		double a1 = 2.*(w*w - 1.);
		double a2 = 1. + w*(w - 2.*coef);

		*c++ = b0*rcp_a0;
		*c++ = b1*rcp_a0;
		*c++ = b2*rcp_a0;
		*c++ = a1*rcp_a0;
		*c++ = a2*rcp_a0;
	}

	// If order is odd, calculate a first order section too
	if (_order&1) {
		switch (_type) {
		case Lowpass:
			b0 = b1 = w;
			break;
		case Highpass:
			b0 = 1.;
			b1 = -1.;
			break;
		}
		double rcp_a0 = 1./(1. + w);
		double a1 = w - 1.;

		*c++ = b0*rcp_a0;
		*c++ = b1*rcp_a0;
		*c++ = a1*rcp_a0;
	}
}

