/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "ConfigFile.hpp"

void Tokenizer::Tokenize(std::string& s)
{
	_tokbounds.first = _tokbounds.second = std::string::npos;
	_curline = &s;
}

Tokenizer::TokenType Tokenizer::GetToken()
{
	std::string& s = *_curline;
	std::string::size_type p;
	std::string::size_type start = _tokbounds.second;
	
	// If we're at the start of a new line, start searching at pos 0, if
	// we're continuing from another token, start at pos behind token.
	if (start == std::string::npos)
		start = 0;
	else
		start++;

	if (_tokbounds.second + 1 >= s.size())
		// We're past the end of the string
		return Tokenizer::End;
	p = s.find_first_not_of(" \t\n\r", start);
	if (p == std::string::npos) {
		// We found nothing but whitespace
		return Tokenizer::End;
	}

	// Look for recognizable tokens
	if (s[p] == '#') {
		// Rest of the string is a comment
		return Tokenizer::End;
	} else if (s[p] == ',') {
		// We found a value delimiter
		_tokbounds.first = p;
		_tokbounds.second = p;
		return Tokenizer::ValueSeparator;
	} else if (s[p] == '=') {
		// We found a key-value delimiter
		_tokbounds.first = p;
		_tokbounds.second = p;
		return Tokenizer::KeySeparator;
	} else if (s[p] == '\"' || s[p] == '\'') {
		// We found the start of a string
		const char quote = s[p];
		std::string::size_type e = p;

		// Now scan for the end of the string, disregarding escaped quotes
		do {
			e = s.find(quote, e + 1);
		} while (e != std::string::npos && s[e - 1] == '\\');
		if (e == std::string::npos)
			throw std::runtime_error("Could not find end quote in value");
		_tokbounds.first = p;
		_tokbounds.second = e;
		return Tokenizer::String;
	} else {
		// We can't find any special characters, so assume we have a literal
		// token
		std::string::size_type e = s.find_first_of(" #\t,=\n\r", p + 1);
		if (e == std::string::npos)
			e = s.size();
		_tokbounds.first = p;
		_tokbounds.second = e - 1;
		return Tokenizer::Literal;
	}
}

void ConfigReader::read_key(std::ifstream& f, std::string& s, std::string& key)
{
	Tokenizer::TokenBoundaries bounds;

	// Scan for a key token until we find one or we reach the end of the file
	do {
		Tokenizer::TokenType t = _tok.GetToken();
		if (t == Tokenizer::End) {
			std::getline(f, s);
			_tok.Tokenize(s);
		}
		if (t == Tokenizer::Literal) {
			// We found a possible key, check if the next token is a "=" to make
			// sure.
			bounds = _tok.GetTokenBoundaries();
			key.assign(s, bounds.first, bounds.second - bounds.first + 1);
			t = _tok.GetToken();
			if (t != Tokenizer::KeySeparator)
				throw std::runtime_error("Expected key/value separator");
			if (_cfgitems.find(key) != _cfgitems.end())
				throw std::runtime_error("Item \"" + key
				                         + "\" already exists");
			_cfgitems[key] = Item();
			break;
		}
	} while (!f.eof());
}

ConfigReader::ConfigReader(std::string fn, bool excpt)
    : _exceptions_enabled(excpt)
{
	std::ifstream f(fn.c_str(), std::ios::in);
	std::string curline;

	if (!f)
		throw std::runtime_error("Couldn't open configuration file");

	while (!f.eof()) {
		Tokenizer::TokenType t;
		Tokenizer::TokenBoundaries bounds;
		std::string key;

		std::getline(f, curline);
		_tok.Tokenize(curline);
		// Look for a "key = " sequence
		read_key(f, curline, key);

		// We allow the "key = " line to be empty for prettier matrix
		// configuration entries
		bool skipped_key_line = false;
		bool seen_valsep = true;
		unsigned row = 1, col = 1;
		Item& cfgitem = _cfgitems[key];
		do {
			std::string val;

			t = _tok.GetToken();
			if (t == Tokenizer::Literal || t == Tokenizer::String) {
				if (!seen_valsep)
					throw std::runtime_error("Expected value separator in "
					                         "item " + key);
				bounds = _tok.GetTokenBoundaries();
				val.assign(curline, bounds.first,
				           bounds.second - bounds.first + 1);
				cfgitem.Set(val, col, row);
				++col;
				seen_valsep = false;
				t = _tok.GetToken();
				if (t == Tokenizer::End) {
					// We're finished, break out of value loop 
					break;
				} else if (t == Tokenizer::ValueSeparator) {
					seen_valsep = true;
				}
			} else if (t == Tokenizer::End) {
				if (!skipped_key_line && col == 1 && row == 1) {
					// We haven't seen any values yet, and skip the key
					// line in anticipation of some data
					skipped_key_line = true;
				} else if (skipped_key_line && col == 1) {
					throw std::runtime_error("Expected value in item "
					                         + key);
				} else if (seen_valsep) {
					col = 1;
					++row;
					skipped_key_line = true;
				}
				std::getline(f, curline);
				_tok.Tokenize(curline);
			}
		} while (!f.eof());
	}
}

ConfigReader::Item* ConfigReader::GetItem(std::string key)
{
	std::map<std::string, Item>::iterator i = _cfgitems.find(key);
	if (i == _cfgitems.end()) {
		if (!_exceptions_enabled)
			return 0;
		throw std::runtime_error("Item \"" + key + "\" does not exist in "
		                         "configuration file");
	}
	return &i->second;
}

ConfigWriter::ConfigWriter(std::string fn) : _f(fn.c_str(), std::ios::out)
{
	if (!_f)
		throw std::runtime_error("Could not open configuration file for "
		                         "writing");
}

void ConfigWriter::StartMatrixItem(std::string key, unsigned cols,
                                   unsigned rows)
{
	_f << key << (rows == 1 ? " = " : " =\n");
	_col = _row = 1;
	_numcols = cols;
	_numrows = rows;
}

