/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _configfile_h
#define _configfile_h

#include <algorithm>
#include <fstream>
#include <list>
#include <map>
#include <sstream>
#include <stdexcept>
#include <string>
#include <utility>

/**
 * \brief Simple tokenizer used to split a configuration file into its
 * component parts.
 */
class Tokenizer {
public:
	enum TokenType { Literal, String, KeySeparator, ValueSeparator, End };
	typedef std::pair<std::string::size_type, std::string::size_type>
	        TokenBoundaries;
private:
	TokenBoundaries _tokbounds;
	std::string* _curline;
public:
	Tokenizer() : _tokbounds(std::string::npos, std::string::npos)
	{
	}

	/**
	 * Prepare tokenization of new line.
	 * @param s string to tokenize.
	 */
	void Tokenize(std::string& s);
	/**
	 * Parse string for another token and return type.
	 * @return Token type found.
	 */
	TokenType GetToken();

	/**
	 * Get string indices representing the boundaries of the previous token
	 * found with GetToken(). Indices are from the first character of the token
	 * to (inclusive) the last. For a one character token, they are the same.
	 * @return Token boundaries.
	 */
	TokenBoundaries GetTokenBoundaries() const
	{
		return _tokbounds;
	}
};

/**
 * \brief Class for reading simple configuration files.
 *
 * Configuration items need to be in the form "name = value", where "name" has
 * to be one word. "value" can be one of the following:
 * - one word/number/etc, for example: 123.00, 1337, cake, true
 * - a quote delimited string: "hello world", "escaped \"quotes\"", '1 2 3'
 * - a list/vector of the above: 123.3, "hello world", cake, 1, 2, 3
 * - a matrix of the above:
 *   1, 2, 3,
 *   "hello world", 5, true,
 *   7, 8, 9
 *   Note that the last entry must not have a trailing comma.
 * 
 * One newline after the "=" character is allowed before a value, and can be
 * used to make matrix items look a bit tidier.
 * Characters should be ASCII. Comments are marked with the # character and
 * can occur anywhere.
 */
class ConfigReader {
public:
	/**
	 * \brief Class containing a set of values for one configuration item.
	 */
	class Item {
		struct Value {
			std::string value;
			unsigned col, row;
			Value(std::string& val, size_t c, size_t r)
			: value(val), col(c), row(r)
			{
			}
		};
		std::list<Value> _values;
	public:
		/**
		 * Sets value in item.
		 * @param val string with value.
		 * @param col matrix column of value to set.
		 * @param row matrix row of value to set.
		 */
		void Set(std::string& val, unsigned col = 1, unsigned row = 1)
		{
			_values.push_back(Value(val, col, row));
		}

		/**
		 * Get value in item.
		 * @param col matrix column of value to get.
		 * @param row matrix row of value to get.
		 * @return String with value.
		 */
		std::string Get(unsigned col = 1, unsigned row = 1) const
		{
			std::list<Value>::const_iterator i = _values.begin();

			do {
				if (i->col == col && i->row == row)
					return i->value;
			} while (++i != _values.end());
			// If we exited the loop it means we found nothing
			throw std::runtime_error("Could not find item");
		}

		/**
		 * Find number of entries in specified row.
		 * @return Number of entries in row.
		 */
		unsigned RowSize(unsigned row) const
		{
			unsigned size = 0;
			std::list<Value>::const_iterator i = _values.begin();

			do {
				if (i->row == row)
					size = std::max(size, i->col);
			} while (++i != _values.end());
			return size;
		}

		/**
		 * Find number of entries in specified column.
		 * @return Number of entries in column.
		 */
		unsigned ColSize(unsigned col) const
		{
			unsigned size = 0;
			std::list<Value>::const_iterator i = _values.begin();

			do {
				if (i->col == col)
					size = std::max(size, i->row);
			} while (++i != _values.end());
			return size;
		}
	};

private:
	Tokenizer _tok;
	std::map<std::string, Item> _cfgitems;
	bool _exceptions_enabled;
	void read_key(std::ifstream& f, std::string& s, std::string& key);
public:
	/**
	 * Class constructor.
	 * @param fn filename from which to read configuration items.
	 * @param excpt decide whether to throw an exception if trying to fetch a
	 * configuration item that does not exist in the file.
	 */
	ConfigReader(std::string fn, bool excpt = true);
	/**
	 * Get configuration item. Depending on constructor, this method might
	 * throw an exception if specified item does not exist.
	 * @param key name of configuration item to get.
	 * @return Pointer to Item class.
	 */
	Item* GetItem(std::string key);

	/**
	 * Get configuration value and convert to desired type. Depending on
	 * constructor, this method might throw an exception if specified item
	 * does not exist.
	 * @param key name of configuration item to get.
	 * @param res reference to a variable where the configuration value will be
	 * stored after type conversion.
	 * @param col item matrix column.
	 * @param row item matrix row.
	 */
	template <typename T>
	void GetValue(std::string key, T& res, unsigned col = 1, unsigned row = 1)
	{
		std::stringstream conv;
		Item* item = GetItem(key);

		if (item == 0)
			return;
		std::string value = item->Get(col, row);
		if (!(conv << value) || !(conv >> res) || !(conv >> std::ws).eof())
			throw std::runtime_error("Invalid value in item \"" + key +
			                         "\": " + value);
	}
};

/**
 * \brief Class for writing simple configuration files.
 *
 * See ConfigReader for configuration file format.
 */
class ConfigWriter {
	std::ofstream _f;
	unsigned _numcols, _numrows;
	unsigned _col, _row;
public:
	ConfigWriter(std::string fn);
	/**
	 * Start matrix/vector configuration item.
	 * @param key name of item.
	 * @param cols number of columns in matrix/vector.
	 * @param rows number of rows in matrix (use 1 for vector).
	 */
	void StartMatrixItem(std::string key, unsigned cols, unsigned rows = 1);

	/**
	 * Set next item in configuration item started with StartMatrixItem() call.
	 * @param value configuration value.
	 */
	template <typename T>
	void SetItemValue(T value)
	{
		_f << value;
		if (_col++ < _numcols) {
			_f << ", ";
		} else {
			_col = 1;
			if (_row++ < _numrows)
				_f << "," << std::endl;
			else
				_f << std::endl;
		}
	}

	/**
	 * Create item with single value.
	 * @param key item name.
	 * @param value item value.
	 */
	template <typename T>
	void SetItem(std::string key, T value)
	{
		_f << key << " = " << value << std::endl;
	}

	void Comment(std::string comment)
	{
		_f << "# " << comment << std::endl;
	}
};

#endif

