/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _dsptools_h
#define _dsptools_h

#include <algorithm>

/**
 * \brief Simple interpolating gain class.
 */
class Gainer {
	float _gain, _delta;
	unsigned _samples;
public:
	/**
	 * Class constructor.
	 * @param gain gain factor, defaults to unity gain (1.0).
	 */
	Gainer(float gain = 1.f) : _gain(gain), _samples(0)
	{
	}

	/**
	 * Check if gain is currently interpolating.
	 * @return True if gain is interpolating, false if not.
	 */
	bool Interpolating() const
	{
		return _samples != 0;
	}

	/**
	 * Set gain factor.
	 * @param gain gain factor.
	 * @param samples over how many samples to interpolate from the current gain
	 * setting to the specified one, defaults to instantaneous change (0).
	 */
	void SetGain(float gain, unsigned samples = 0)
	{
		_samples = samples;
		if (samples > 0)
			_delta = (gain - _gain)/_samples;
		else
			_gain = gain;
	}

	/**
	 * Perform gaining.
	 * @param x array of floats to apply gain to.
	 * @param num number of elements in x to apply gain to.
	 */
	void Work(float* x, unsigned num)
	{
		unsigned samples_left = std::min(_samples, num);
		unsigned i;

		// TODO simd optimize
		for (i = 0; i < samples_left; ++i) {
			x[i] *= _gain;
			_gain += _delta;
		}
		for (; i < num; ++i)
			x[i] *= _gain;
		_samples -= samples_left;
	}

	/**
	 * Fill a vector with gain values, so an external procedure can apply the
	 * actual gaining. Mainly used to apply the same interpolated gain curve to
	 * the several data sets.
	 * @param x array of floats in which to store the gain values.
	 * @param num number of gain values to store.
	 */
	void MakeGainVector(float* x, unsigned num)
	{
		unsigned samples_left = std::min(_samples, num);
		unsigned i;

		for (i = 0; i < samples_left; ++i) {
			x[i] = _gain;
			_gain += _delta;
		}
		for (; i < num; ++i)
			x[i] = _gain;
		_samples -= samples_left;
	}
};


/**
 * \brief Class for performing (linear) crossfading between two signals.
 */
class Crossfader {
	float _pos, _delta;
	unsigned _samples, _delay_samples;
	float _state_pos;
	unsigned _state_samples, _state_delay_samples;
public:
	/**
	 * Class constructor. Crossfader is constructed in finished state.
	 */
	Crossfader() : _samples(0), _delay_samples(0)
	{
	}

	/**
	 * Initiates a crossfade if none is already started, or change a currently
	 * executing crossfeed.
	 * @param fadein whether the first signal specified in Process should be
	 * faded in or out.
	 * @param samples duration of crossfeed in samples.
	 * @param delay how many samples to delay onset of the crossfade if no
	 * crossfade is currently in action. Defaults to no delay.
	 */
	void Start(bool fadein, unsigned samples, unsigned delay = 0)
	{
		// Check if fade is already in progress
		if (_samples == 0) {
			_pos = fadein ? 0.f : 1.f;
			_delay_samples = delay;
		}
		_samples = samples;
		if (fadein)
			_delta = 1.f - _pos;
		else
			_delta = -_pos;
		_delta /= _samples;
	}

	/**
	 * Queries whether crossfader is finished fading.
	 * @return True if finished, false if not.
	 */
	bool Finished() const
	{
		return _samples == 0;
	}

	/**
	 * Saves the current state of the crossfader. Use this if you need to fade
	 * several channels of data in parallel, but want to allocate just one
	 * Crossfader object.
	 * @see LoadState
	 */
	void SaveState()
	{
		_state_pos = _pos;
		_state_samples = _samples;
		_state_delay_samples = _delay_samples;
	}

	/**
	 * Loads state saved with last call of SaveState().
	 * @see SaveState.
	 */
	void LoadState()
	{
		_pos = _state_pos;
		_samples = _state_samples;
		_state_delay_samples = _delay_samples;
	}

	/**
	 * Crossfade audio data.
	 * @param in1 array of data to be crossfaded.
	 * @param in2 array of data to be crossfaded.
	 * @param out array in which to store crossfaded result.
	 * @param num number of samples to crossfade.
	 */
	void Process(float* in1, float* in2, float* out, unsigned num)
	{
		// TODO use some equal power crossfading function
		unsigned i = 0;

		for (; i < num && _delay_samples > 0; ++i, --_delay_samples)
			out[i] = _pos*in1[i] + (1.f - _pos)*in2[i];
		// Main crossfade
		for (; i < num && _samples > 0; ++i, --_samples) {
			out[i] = _pos*in1[i] + (1.f - _pos)*in2[i];
			_pos += _delta;
		}
		// If the fade is done, but we've not produced all samples we should
		// yet, just go on as before, but stop updating _pos.
		for (; i < num; ++i)
			out[i] = _pos*in1[i] + (1.f - _pos)*in2[i];
	}
};

#endif

