/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _delayline_h
#define _delayline_h

#include <cmath>
#include <vector>

/**
 * \brief Modulatable delay line using third order hermite interpolation.
 */
class DelayLine {
private:
	float _max_delay, _sample_rate, _c;
	float _dist_conv;
	std::vector<float> _buffer;
	unsigned _mask, _writepos;
	float _curdelay;
	float* _writeptr;
	
public:
	/**
	 * Class constructor.
	 * @param sr sample rate.
	 * @param max_delay maximum delay in seconds. If any subsequent delay time
	 * is greater than this, the delay line will not function as intended, but
	 * tt should also not crash.
	 * @param c speed of sound in meters per second. Defaults to 344 m/s.
	 */
	DelayLine(unsigned sr, float max_delay, float c = 344.f)
	: _max_delay(max_delay), _c(c)
	{
		SetSampleRate(sr);
	}

	/**
	 * Clears the delay line to 0.
	 */
	void Clear()
	{
		std::fill(_buffer.begin(), _buffer.end(), 0.f);
	}

	/**
	 * Sets sample rate.
	 * @param sr sample rate.
	 */
	void SetSampleRate(unsigned sr)
	{
		_sample_rate = sr;
		_dist_conv = _sample_rate/_c;
		double samples = _max_delay*_sample_rate;
		// Round sample amount up to next power of two
		unsigned num = unsigned(pow(2., ceil(log(samples)/log(2.))));
		_mask = num - 1;
		_buffer.assign(num, 0.f);
		_writepos = 0;
	}

	void SetDistance(float distance)
	{
		_curdelay = distance*_dist_conv;
	}
	
	/**
	 * Delay with array of distance values for use when distance is changing.
	 * Ignores the internal _curdelay variable set by SetDistance.
	 * @param in input array.
	 * @param out output array.
	 * @param distance array of distances deciding the instantaneous delay.
	 * @param num number of samples.
	 */
	void Process(float* in, float* out, float* distance, unsigned num)
	{
		for (unsigned i = 0; i < num; ++i) {
			const float curdelay = distance[i]*_dist_conv;
			// Calculate read position
			unsigned r = _writepos - unsigned(curdelay);
			// Calculate fractional read position (1 - frac(curdelay))
			float frac = 1.f - (curdelay - truncf(curdelay));
			// Do masked reads of all values we need. This way the reads will
			// automatically wrap if they're out of bounds.
			const float xm1 = _buffer[(r - 1) & _mask];
		   	const float x0 = _buffer[r & _mask];
			const float xp1 = _buffer[(r + 1) & _mask];
			const float xp2 = _buffer[(r + 2) & _mask];

			// Third order hermite interpolation
			const float c = 0.5f*(xp1 - xm1);
			const float v = x0 - xp1;
			const float w = c + v;
			const float a = w + v + 0.5f*(xp2 - x0);
			const float b = w + a;

			// Calculate interpolated output.
			float output = ((a*frac - b)*frac + c)*frac + x0;
			// Save input to delay line first, in case in and out are the same.
			_buffer[_writepos++] = in[i];
			out[i] = output;
			// Mask write position in case it overflows buffer.
			_writepos &= _mask;
		}
	}

	/**
	 * Delay for use when distance is static. Uses the distance value set with
	 * SetDistance().
	 * @param in input array.
	 * @param out output array.
	 * @param num number of samples.
	 */
	void Process(float* in, float* out, unsigned num)
	{
		// See other Process for more comments.
		const unsigned read_trail = unsigned(_curdelay);
		// Calculate fractional read position (1 - frac(curdelay))
		const float frac = 1.f - (_curdelay - truncf(_curdelay));
		// Calculate read position
		unsigned r = _writepos - read_trail;
		// Do masked reads of all values we need. This way the reads will
		// automatically wrap if they're out of bounds.
		float xm1 = _buffer[(r - 1) & _mask];
	   	float x0 = _buffer[r & _mask];
		float xp1 = _buffer[(r + 1) & _mask];
		float xp2 = _buffer[(r + 2) & _mask];

		for (unsigned i = 0; i < num; ++i) {
			// Third order hermite interpolation
			const float c = 0.5f*(xp1 - xm1);
			const float v = x0 - xp1;
			const float w = c + v;
			const float a = w + v + 0.5f*(xp2 - x0);
			const float b = w + a;

			// Calculate interpolated output.
			float output = ((a*frac - b)*frac + c)*frac + x0;
			// Save input to delay line first, in case in and out are the same.
			_buffer[_writepos++] = in[i];
			out[i] = output;
			// Mask write position in case it overflows buffer.
			_writepos &= _mask;
			r = _writepos - read_trail;
			// r is incremented one sample at a time, so we only need to fetch
			// one new sample, the other three are already in x0, xp1 and xp2
			xm1 = x0;
			x0 = xp1;
			xp1 = xp2;
			xp2 = _buffer[(r + 2) & _mask];
		}
	}
};

#endif

