/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "EncoderSettingsWindow.hpp"
#include <gtkmm/notebook.h>
#include <gtkmm/stock.h>
#include <gtkmm/table.h>

EncoderSettingsWindow::EncoderSettingsWindow(AmbisonicsEncoderModel& enc)
	: _adj_maxsourcegain(10., -20., 40., 1., 3.),
	  _adj_halfpointtime(1., 0.1, 10., 0.1, 1.),
	  _btn_maxsourcegain(_adj_maxsourcegain, 0., 0),
	  _btn_halfpointtime(_adj_halfpointtime, 0., 1),
	  _enc(enc)
{
	set_title("Encoder settings");
	set_border_width(5);
	set_default_size(300, 200);

	Gtk::Notebook* tabs = Gtk::manage(new Gtk::Notebook());
	Gtk::Table* table_gen = Gtk::manage(new Gtk::Table(1, 3));
	Gtk::Table* table_src = Gtk::manage(new Gtk::Table(4, 3));
	Gtk::Table* table_nfc = Gtk::manage(new Gtk::Table(2, 3));
	// Build a GUI consisting of a NoteBook with one table in each tab.
	// First page labels
	Gtk::Label* label1 = Gtk::manage(new Gtk::Label("Distance attenuation:"));
	label1->set_alignment(Gtk::ALIGN_RIGHT);
	Gtk::Label* label2 = Gtk::manage(new Gtk::Label("Propagation delay:"));
	label2->set_alignment(Gtk::ALIGN_RIGHT);
	Gtk::Label* label3 = Gtk::manage(new Gtk::Label("Maximum source gain:"));
	label3->set_alignment(Gtk::ALIGN_RIGHT);
	Gtk::Label* label4 = Gtk::manage(new Gtk::Label("Interpolation half-point time:"));
	label4->set_alignment(Gtk::ALIGN_RIGHT);
	Gtk::Label* label5 = Gtk::manage(new Gtk::Label("dB"));
	Gtk::Label* label6 = Gtk::manage(new Gtk::Label("s"));

	// Second page labels
	Gtk::Label* label7 = Gtk::manage(new Gtk::Label("Near-field compensation:"));
	label7->set_alignment(Gtk::ALIGN_RIGHT);
	Gtk::Label* label8 = Gtk::manage(new Gtk::Label("Per-component highpass filters (EXPERIMENTAL!):"));
	label8->set_alignment(Gtk::ALIGN_RIGHT);
	
	Gtk::Label* label9 = Gtk::manage(new Gtk::Label("Output normalization:"));
	label9->set_alignment(Gtk::ALIGN_RIGHT);

	// Populate general settings page
	table_gen->set_border_width(16);
	table_gen->set_col_spacing(0, 20);
	table_gen->set_col_spacing(1, 10);
	table_gen->attach(*label9, 0, 1, 0, 1, Gtk::FILL);
	table_gen->attach(_combo_norm, 1, 2, 0, 1, Gtk::FILL | Gtk::EXPAND, Gtk::EXPAND);

	// Populate source settings page
	table_src->set_border_width(16);
	table_src->set_col_spacing(0, 20);
	table_src->set_col_spacing(1, 10);
	table_src->attach(*label1, 0, 1, 0, 1, Gtk::FILL);
	table_src->attach(*label2, 0, 1, 1, 2, Gtk::FILL);
	table_src->attach(*label3, 0, 1, 2, 3, Gtk::FILL);
	table_src->attach(*label4, 0, 1, 3, 4, Gtk::FILL);
	table_src->attach(_btn_distance, 1, 2, 0, 1);
	table_src->attach(_btn_delay, 1, 2, 1, 2);
	table_src->attach(_btn_maxsourcegain, 1, 2, 2, 3);
	table_src->attach(_btn_halfpointtime, 1, 2, 3, 4);
	table_src->attach(*label5, 2, 3, 2, 3, Gtk::FILL);
	table_src->attach(*label6, 2, 3, 3, 4, Gtk::FILL);
	
	// Populate NFC settings page
	table_nfc->set_border_width(16);
	table_nfc->set_col_spacing(0, 20);
	table_nfc->set_col_spacing(1, 10);
	table_nfc->attach(*label7, 0, 1, 0, 1, Gtk::FILL);
	table_nfc->attach(_btn_nfc, 1, 2, 0, 1);
	table_nfc->attach(*label8, 0, 1, 1, 2, Gtk::FILL);
	table_nfc->attach(_btn_nfc_highpass, 1, 2, 1, 2);

	// Connect signal handlers
	_combo_norm.signal_changed().connect(sigc::mem_fun(
		*this, &EncoderSettingsWindow::on_normalization_changed));
	_btn_distance.signal_clicked().connect(sigc::mem_fun(
		*this, &EncoderSettingsWindow::on_distance_clicked));
	_btn_delay.signal_clicked().connect(sigc::mem_fun(
		*this, &EncoderSettingsWindow::on_delay_clicked));
	_btn_maxsourcegain.signal_value_changed().connect(sigc::mem_fun(
		*this, &EncoderSettingsWindow::on_maxsourcegain_changed));
	_btn_halfpointtime.signal_value_changed().connect(sigc::mem_fun(
		*this, &EncoderSettingsWindow::on_halfpointtime_changed));
	_btn_nfc.signal_clicked().connect(sigc::mem_fun(
		*this, &EncoderSettingsWindow::on_nfc_clicked));
	_btn_nfc_highpass.signal_clicked().connect(sigc::mem_fun(
		*this, &EncoderSettingsWindow::on_nfc_highpass_clicked));

	tabs->append_page(*table_gen, "General");
	tabs->append_page(*table_src, "Source settings");
	tabs->append_page(*table_nfc, "Near-field compensation");
	add(*tabs);
	show_all_children();
}

void EncoderSettingsWindow::Notify(int what)
{
	switch (what) {
	case AmbisonicsEncoderModel::ChangedNormalization:
		_combo_norm.set_norm(_enc.Normalization());
		break;
	case AmbisonicsEncoderModel::ChangedDistanceSimulation:
		_btn_distance.set_active(_enc.DistanceSimulation());
		break;
	case AmbisonicsEncoderModel::ChangedDelaySimulation:		
		_btn_delay.set_active(_enc.DelaySimulation());
		break;
	case AmbisonicsEncoderModel::ChangedNFC:
		_btn_nfc.set_active(_enc.NFC());
		break;
	case AmbisonicsEncoderModel::ChangedNFCHighpass:
		_btn_nfc_highpass.set_active(_enc.NFCHighpass());
		break;
	case AmbisonicsEncoderModel::ChangedMaxSourceAmplification:
		_btn_maxsourcegain.set_value(_enc.MaxSourceAmplification());
		break;
	case AmbisonicsEncoderModel::ChangedInterpolationHalfTime:
		_btn_halfpointtime.set_value(_enc.InterpolationHalfTime());
		break;
	case AmbisonicsEncoderModel::ChangedAll:
		_combo_norm.set_norm(_enc.Normalization());
		_btn_distance.set_active(_enc.DistanceSimulation());
		_btn_delay.set_active(_enc.DelaySimulation());
		_btn_nfc.set_active(_enc.NFC());
		_btn_nfc_highpass.set_active(_enc.NFCHighpass());
		_btn_maxsourcegain.set_value(_enc.MaxSourceAmplification());
		_btn_halfpointtime.set_value(_enc.InterpolationHalfTime());
		break;
	}
}

void EncoderSettingsWindow::on_normalization_changed()
{
	_enc.Normalization(_combo_norm.get_norm());
}

void EncoderSettingsWindow::on_distance_clicked()
{
	_enc.DistanceSimulation(_btn_distance.get_active());
}

void EncoderSettingsWindow::on_delay_clicked()
{
	_enc.DelaySimulation(_btn_delay.get_active());
}

void EncoderSettingsWindow::on_nfc_clicked()
{
	_enc.NFC(_btn_nfc.get_active());
}

void EncoderSettingsWindow::on_nfc_highpass_clicked()
{
	_enc.NFCHighpass(_btn_nfc_highpass.get_active());
}

void EncoderSettingsWindow::on_maxsourcegain_changed()
{
	_enc.MaxSourceAmplification(_btn_maxsourcegain.get_value());
}

void EncoderSettingsWindow::on_halfpointtime_changed()
{
	_enc.InterpolationHalfTime(_btn_halfpointtime.get_value());
}

