/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _fftfilter_h
#define _fftfilter_h

#include <algorithm>
#include <cmath>
#include <cstring>
#include <vector>
#include "Defs.hpp"
#include "FFTW.hpp"
#include "SIMDMath.hpp"

// This type can be used to quickly change the FFT class, as long as the
// interface is the same.
typedef FFTW FFT;
typedef float Complex;
typedef std::vector< Complex, align16_allocator<Complex> > ComplexVector;
typedef std::vector<FloatVector> ChannelVector_t;

/**
 * \brief Class for containing one or more FloatVectors, for use with
 * multichannel signal input/output.
 */
class ChannelVector : public ChannelVector_t {
protected:
	size_t _size;
public:
	ChannelVector() : _size(0)
	{
	}

	/**
	 * Class constructor.
	 * @param number of channels.
	 * @param size number of samples in each channel.
	 */
	ChannelVector(size_t channels, size_t size = 0)
	: ChannelVector_t(channels, FloatVector(size, 0.f)), _size(size)
	{
	}

	/**
	 * Set number of samples in each channel.
	 * @param newsize new number of samples per channel
	 */
	void bufsize(size_t newsize)
	{
		for (size_t i = 0; i < size(); ++i)
			(*this)[i].resize(newsize);
		_size = newsize;
	}
	
	size_t bufsize() const
	{
		return _size;
	}	

	/**
	 * Sets number of channels.
	 * @param numchannels number of channels.
	 */	
	void channels(size_t numchannels)
	{
		resize(numchannels, FloatVector(_size, 0.f));
	}
	
	size_t channels() const
	{
		return size();
	}
};

/**
 * \brief Class for containing a partitioned filter for use with FFTConvolver.
 */
class PartitionedFilter : public ComplexVector {
protected:
	unsigned _num_parts;
public:
	/**
	 * Class constructor.
	 * @param P block size used for filtering. Must be a power of two.
	 * @param f reference to FloatVector containing filter coefficients.
	 */
	PartitionedFilter(unsigned P, FloatVector& f)
	{
		const unsigned N = P*2;
		FFT fft(N);
		FloatVector impulsepart(N);
		FloatVector::iterator partition_end = f.begin();

		// Divide filter size by partition size and round up
		_num_parts = (f.size() + P - 1)/P;
		resize(_num_parts*N);
		// Now do a separate FFT of each filter partition padded with zeros to
		// twice its length
		for (unsigned i = 0; i < _num_parts; ++i) {
			FloatVector::iterator partition_start = partition_end;
			partition_end += P;
			if (partition_end > f.end())
				partition_end = f.end();

			// Zero-pad
			std::fill(impulsepart.begin(), impulsepart.end(), 0.f);
			std::copy(partition_start, partition_end, impulsepart.begin());
			// Normalise, so we don't have to do this when convolving run-time
			fft.rescale(&impulsepart[0]);
			fft.do_fft(&(*this)[i*N], &impulsepart[0]);
		}
	}

	/**
	 * Get number of partitions filter is partitioned into.
	 * @return The number of partitions.
	 */
	unsigned GetNumPartitions() const
	{
		return _num_parts;
	}
};

/**
 * \brief Class for containing and outputting output data.
 */
class OutputBuffer : public ChannelVector {
protected:
	unsigned _pos;
public:
	OutputBuffer() : _pos(0)
	{
	}

	/**
	 * Class constructor.
	 * @param number of channels.
	 * @param size number of samples in each channel.
	 */
	OutputBuffer(size_t channels, size_t size)
		: ChannelVector(channels, size), _pos(0)
	{
	}
	
	/**
	 * Clear contents and reset buffer position.
	 */
	void Clear()
	{
		_pos = 0;
		for (size_t channel = 0; channel < channels(); ++channel)
			std::fill((*this)[channel].begin(), (*this)[channel].end(), 0.f);
	}
	
	/**
	 * Writes output a number of samples.
	 * @param out two-dimensional float array in which to place output data.
	 * @param num number of samples to output per channel.
	 * @param offset start offset for each channel in output array. Defaults
	 * to 0.
	 */
	void Output(float** out, unsigned num, unsigned offset = 0)
	{
		for (size_t channel = 0; channel < channels(); ++channel) {
			float* start = &((*this)[channel][_pos]);
			memcpy(out[channel] + offset, start, num*sizeof(float));
		}
		_pos += num;
	}
};

/**
 * \brief Class for dealing with filter history. Contains previously FFTed input
 * blocks.
 */
class History : public std::vector<ComplexVector> {
	size_t _histpos;
public:
	size_t GetPos()
	{
		return _histpos;
	}

	void ClearPos()
	{
		_histpos = 0;
	}

	void IncrementPos()
	{
		if (_histpos == 0)
			_histpos = size();
		--_histpos;
	}
};

typedef std::vector<History> HistoryVector;

/**
 * \brief Overlapped input buffer class for use with FFTConvolver.
 *
 * Contains its own filter input history.
 */
class InputBuffer : public ChannelVector {
	unsigned _buf_pos;
	// This will be where the FFT convolution algorithm saves the previous 
	// FFTed data belonging to this buffer
	HistoryVector _history;
public:
	InputBuffer()
	{
	}
	
	/**
	 * Class constructor.
	 * @param number of channels.
	 * @param size number of samples in each channel.
	 */
	InputBuffer(size_t channels, size_t size)
		: ChannelVector(channels, size*2), _buf_pos(size), _history(channels)
	{
	}

	/**
	 * Read data from two-dimensional float array until input buffer is full.
	 * @param in two-dimensional float array from which to read data.
	 * @param num maximum number of samples to read.
	 * @param offset start offset for each channel in input array. Defaults
	 * to 0.
	 * @return Number of samples actually read.
	 */
	size_t Input(float** in, unsigned num, size_t offset = 0)
	{
		size_t samples_left = std::min<size_t>(num, bufsize() - _buf_pos);

		for (size_t c = 0; c < channels(); ++c)
			memcpy(&((*this)[c][_buf_pos]), in[c] + offset,
			       samples_left*sizeof(float));
		_buf_pos += samples_left;
		return samples_left;
	}

	/**
	 * Read data from ChannelVector until input buffer is full.
	 * @param in ChannelVector from which to read data.
	 * @param num maximum number of samples to read.
	 * @param offset start offset for each channel in input array. Defaults
	 * to 0.
	 * @return Number of samples actually read.
	 */
	size_t Input(ChannelVector& in, size_t num, size_t offset = 0)
	{
		size_t samples_left = std::min(num, bufsize() - _buf_pos);
		ChannelVector& buf = *this;

		for (size_t c = 0; c < channels(); ++c)
			memcpy(&buf[c][_buf_pos], &in[c][offset],
			       samples_left*sizeof(float));
		_buf_pos += samples_left;
		return samples_left;
	}
	
	/**
	 * Copies the second half of the buffer to the first half and resets buffer
	 * position for more reading.
	 */
	void Overlap()
	{
		InputBuffer& buf = *this;
		_buf_pos = bufsize()/2;
		for (size_t c = 0; c < channels(); ++c)
			memcpy(&buf[c][0], &buf[c][_buf_pos], _buf_pos*sizeof(float));
	}
	
	/**
	 * Check if buffer is full.
	 * @return Buffer full status.
	 */
	bool Full() const
	{
		return _buf_pos == bufsize();
	}

	/**
	 * Resets internal filter history. Must always be called before using class.
	 * @param f PartitionedFilter instance to use.
	 */
	void ResetHistory(PartitionedFilter& f)
	{
		size_t num_partitions = f.GetNumPartitions();

		for (size_t i = 0; i < _history.size(); ++i) {
			_history[i].resize(num_partitions);
			_history[i].ClearPos();
			for (size_t j = 0; j < num_partitions; ++j) {
				// History blocks will always have same size as length of the
				// input buffer
				_history[i][j].assign(bufsize(), 0.f);
			}
		}
	}

	// Return reference to the filter history.
	HistoryVector& GetHistory()
	{
		return _history;
	}
};

/**
 * \brief Class for FFT based convolution.
 *
 * Uses partitioned overlap-save with support for SIMD accelerated complex
 * multiply to achieve efficient convolution.
 */
class FFTConvolver {
	boost::shared_ptr<FFT> _fft;
	boost::shared_ptr<SIMDMath> _math;
	ComplexVector _accf, _temp;
	PartitionedFilter* _ir;
	unsigned _N, _P, _L;
public:
	/**
	 * Class constructor.
	 * @param P number of samples per block to perform filtering on.
	 */
	FFTConvolver(unsigned P = 512)
	{
		_math = SIMDMath::Create();
		SetBlockSize(P);
	}

	/**
	 * Set filter to be used in next filtering operation.
	 * @param f partitioned filter to use.
	 */
	void SetFilter(PartitionedFilter& f)
	{
		_ir = &f;
	}
	
	/**
	 * Take one block of input and place it in the input history, ready to be
	 * used in filtering.
	 * @param in InputBuffer containing input block.
	 */
	void Input(InputBuffer& in)
	{
		HistoryVector& history = in.GetHistory();

		for (size_t chan = 0; chan < in.channels(); ++chan) {
			size_t histpos = history[chan].GetPos();
			_fft->do_fft(&history[chan][histpos].front(), &in[chan][0]);
		}
	}
	
	/**
	 * Performs filtering.
	 * @param in InputBuffer where the input signal is stored. Remember to call
	 * the Input() method on this before using it with this method.
	 * @param out OutputBuffer where filtered output gets written.
	 * @param chan channel index to use for input and output.
	 */ 
	void Work(InputBuffer& in, OutputBuffer& out, size_t chan)
	{
		HistoryVector& history = in.GetHistory();
		size_t hpos = history[chan].GetPos();

		_math->SpectrumMul(&_accf[0], &history[chan][hpos++].front(),
		                   &_ir->front(), _N);
		for (unsigned i = 1; i < _ir->GetNumPartitions(); ++i) {
			if (hpos >= history[chan].size())
				hpos = 0;
			_math->SpectrumMulAcc(&_accf[0], &history[chan][hpos++].front(),
			                      &_ir->front() + i*_N, _N);
		}
		_fft->do_ifft(&_accf[0], &_temp[0]);
		// Toss away the first P values
		// This copy can be eliminated if we're ok with wasting memory in the
		// OutputBuffer class. In that case, we can write straight to that.
		memcpy(&out[chan][0], &_temp[_P], (_N - _P)*sizeof(float));
		history[chan].IncrementPos();
	}
	
	/**
	 * Sets block size to use, corresponds to half the FFT size used internally
	 * for filtering. User has to repartition filter used manually.
	 * @param P block size.
	 */
	void SetBlockSize(unsigned P)
	{
		_P = _L = P;
		_N = P*2;
		_fft.reset(new FFT(_N));
		_accf.resize(_N);
		_temp.resize(_N);
	}
};

#endif

