/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "FFTW.hpp"
#include <algorithm>
#include <cstdio>
#include <cstring>
#include <iostream>

void FFTW::HandleWisdom(std::string filename)
{
	int sizes[] = { 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384 };
	fftwf_plan fft, ifft;
	char buf[1024];
	char *homedir;
	FILE* f;	

	homedir = getenv("HOME");
	if (!homedir)
		return;
	strcpy(buf, homedir);
	strcat(buf, "/");
	strcat(buf, filename.c_str());
	
	// First lets try to learn from the wisdom.
	if ((f = fopen(buf, "r")) != NULL) {
		fftwf_import_wisdom_from_file(f);
		fclose(f);
	}

	// Then let's try to improve on it.
	for (unsigned i = 0; i < sizeof(sizes)/sizeof(int); ++i) {
		int size = sizes[i];
		float* in = (float *)fftw_malloc(size*sizeof(float));
		float* out = (float *)fftw_malloc(size*sizeof(float));
		
		fft = fftwf_plan_r2r_1d(size, in, out, FFTW_R2HC,
		                        FFTW_PATIENT | FFTW_PRESERVE_INPUT);
		ifft = fftwf_plan_r2r_1d(size, out, in, FFTW_HC2R,
		                         FFTW_PATIENT | FFTW_PRESERVE_INPUT);
		fftw_free(in);
		fftw_free(out);
		fftwf_destroy_plan(fft);
		fftwf_destroy_plan(ifft);
		std::cout << ".";
		std::cout.flush();
	}
	std::cout << std::endl;

	if ((f = fopen(buf, "w")) != NULL) {
		fftwf_export_wisdom_to_file(f);
		fclose(f);
	}
}

FFTW::FFTW(const long length) : _length(length)
{
	flt_t* in;
	flt_t* out;
	
	in = (flt_t*)fftw_malloc(length*sizeof(flt_t));
	out = (flt_t*)fftw_malloc(length*sizeof(flt_t));
	// We use the half-complex output format because of the ease of implementing
	// SIMD versions of the convolution multiplication with it.
	_fft = fftwf_plan_r2r_1d(length, in, out, FFTW_R2HC,
	                         FFTW_PATIENT | FFTW_PRESERVE_INPUT);
	_ifft = fftwf_plan_r2r_1d(length, out, in, FFTW_HC2R,
	                          FFTW_PATIENT | FFTW_PRESERVE_INPUT);
	fftw_free(in);
	fftw_free(out);
}

void FFTW::do_fft(flt_t* f, flt_t* x)
{
	fftwf_execute_r2r(_fft, x, f);
	std::reverse(f + _length/2 + 1, f + _length);
}

void FFTW::do_ifft(flt_t* f, flt_t* x)
{
	std::reverse(f + _length/2 + 1, f + _length);
	fftwf_execute_r2r(_ifft, f, x);
}

void FFTW::rescale(flt_t* x) const
{
	flt_t n = flt_t(1.0/_length);

	for (long i = 0; i < _length; ++i)
		x[i] *= n;
}

FFTW::~FFTW()
{
	fftwf_destroy_plan(_fft);
	fftwf_destroy_plan(_ifft);
}

