/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _iirfilter_h
#define _iirfilter_h

#include <algorithm>
#include <cstdio>
#include <vector>

// Define this to use direct form 1 instead of transposed direct form 2
//#define DF1 1

/**
 * \brief Class for performing IIR filtering.
 * 
 * Filter must be specified in second/first order sections. Coefficient
 * generation must either be handled by derived classes or set with the
 * SetCoefs method.
 * @param T type the filter will use internally for calculations, coefs and
 * history storage. Defaults to float.
 */
template <class T = float>
class IIRFilter {
protected:
	std::vector<T> _coef, _hist; // coefs and filter history
	unsigned _num_second_order_sec;
	bool _first_order_sec;
	/**
	 * Resizes the filter history vector based on current filter coefficients
	 * set.
	 */
	void ResizeHistory()
	{
#ifndef DF1
		_hist.resize(_num_second_order_sec*2 + (_first_order_sec ? 1 : 0));
#else
		_hist.resize(_num_second_order_sec*4 + (_first_order_sec ? 2 : 0));
#endif
	}
public:
	typedef T CoefType;
	/**
	 * Constructs an empty IIR filter.
	 */
	IIRFilter() : _num_second_order_sec(0), _first_order_sec(false)
	{
	}
	/**
	 * Performs the actual IIR filtering using transposed direct form 2.
	 * @param in pointer to input array. Can be the same as the output pointer.
	 * @param out pointer to output array. Can be the same as the input pointer.
	 * @param num number of samples to be filtered.
	 */
	void Process(float* input, float* output, unsigned num)
	{
		unsigned ci = 0, hi = 0;
		
		for (unsigned s = 0; s < _num_second_order_sec; ++s) {
			const T b0 = _coef[ci]; // FIR coefficients
			const T b1 = _coef[ci + 1];
			const T b2 = _coef[ci + 2];
			const T a1 = _coef[ci + 3]; // IIR coefficients
			const T a2 = _coef[ci + 4];
			T w1 = _hist[hi];
			T w2 = _hist[hi + 1];
#ifdef DF1
			T w3 = _hist[hi + 2];
			T w4 = _hist[hi + 3];
#endif

			for (unsigned i = 0; i < num; ++i) {
#ifndef DF1
				const T x = T(input[i]);
				const T y = b0*x + w1;
				w1 = b1*x - a1*y + w2;
				w2 = b2*x - a2*y;
				output[i] = float(y);
#else
				const T y = b1*w1 + b2*w2 - a1*w3 - a2*w4;
				w2 = w1;
				w1 = T(input[i]);
				w4 = w3;
				w3 = y + w1*b0;
				output[i] = float(w3);
#endif
			}
			_hist[hi] = w1;
			_hist[hi + 1] = w2;
#ifdef DF1
			_hist[hi + 2] = w3;
			_hist[hi + 3] = w4;
#endif
			/* We want the next filter sections to use the output of the
			 * preceding section as input, so we start taking filter input
			 * from the out pointer after the first section.
			 */
			input = output;
			ci += 5;
#ifndef DF1
			hi += 2;
#else
			hi += 4;
#endif
		}
		
		if (_first_order_sec) {
			const T b0 = _coef[ci];
			const T b1 = _coef[ci + 1];
			const T a1 = _coef[ci + 2];
			T w1 = _hist[hi];
#ifdef DF1
			T w2 = _hist[hi + 1];
#endif
			for (unsigned i = 0; i < num; ++i) {
#ifndef DF1
				const T x = T(input[i]);
				const T y = b0*x + w1;
				w1 = b1*x - a1*y;
				output[i] = float(y);
#else
				T y = b1*w1 - a1*w2;
				w1 = T(input[i]);
				w2 = y + w1*b0;
				output[i] = float(w2);
#endif
			}
			_hist[hi] = w1;
#ifdef DF1
			_hist[hi + 1] = w2;
#endif
		}
	}

	/**
	 * Clears the filter history.
	 */
	void ClearHistory()
	{
		std::fill(_hist.begin(), _hist.end(), T(0.));
	}
	
	/**
	 * Sets the filter coefficients for the filter.
	 * @param c filter coefficients, FIR coefs first, then IIR coefs.
	 */
	void SetCoefs(std::vector<T>& c)
	{
		_coef = c;
		// Should have five coefficients per second order section...
		_num_second_order_sec = c.size()/5;
		// and three extra if there's a first order section tacked on as well
		_first_order_sec = c.size()%5 == 3;
		ResizeHistory();
	}

	/**
	 * Debugging aid which emits Matlab/Octave code to plot the frequency
	 * response of the current filter.
	 * @param sr sample rate.
	 * @param from frequency scale factor from which to start plotting the
	 * spectrum. Defaults to 0.
	 * @param to frequency scale factor at which to stop plotting the spectrum.
	 * Defaults to 1.
	 */
	void DumpSpectrum(unsigned sr, double from = 0, double to = 1.) const
	{
		const T* c = &_coef.front();
		char mult[] = "H.*";
		char empty[] = "";
		char* mptr = empty;
		
		printf("w = linspace(%f*pi, %f*pi, 2000);\n", from, to);
		for (size_t i = 0; i < _num_second_order_sec; ++i) {
			printf("H = %sfreqz([%.18f, %.18f, %.18f], [1, %.18f, %.18f], w);\n"
		           , mptr, c[0], c[1], c[2], c[3], c[4]);
			mptr = mult;
			c += 5;
		}
		if (_first_order_sec)
			printf("H = %sfreqz([%.18f, %.18f], [1, %.18f], w);\n",
			       mptr, c[0], c[1], c[2]);
		printf("subplot(2, 1, 1);\n"
		       "plot(w/pi*%i/2, 20*log10(abs(H)));\n"
		       "axis([%f, %f]);\n"
		       "subplot(2, 1, 2);\n"
		       "plot(w/pi*%i/2, 180/pi*unwrap(angle(H)));\n"
		       "axis([%f, %f]);\n",
		       sr, sr*0.5*from, sr*0.5*to, sr, sr*0.5*from, sr*0.5*to
		);
	}
};

#endif

