/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _interpolatedfunction_h
#define _interpolatedfunction_h

#include <vector>

/**
 * \brief Class for making a linearly interpolated representation of a function
 * with one variable.
 *
 * Should be used for speeding up slow functions. Function to be used must be
 * available as a functor class.
 * @param F class name of functor to interpolate.
 * @param T type to use.
 */
template<class F, typename T = float>
class InterpolatedFunction {
	std::vector<T> _table;
	F& _func;
	T _scale, _offset;
public:
	/**
	 * Construct an object.
	 * @param F reference to functor object to be interpolated.
	 * @param min minimum function range to interpolate.
	 * @param max maximum function range to interpolate.
	 * @param num number of points in internal function table. The bigger this
	 * number is, the more accurate the interpolation will be.
	 */
	InterpolatedFunction(F& f, T min, T max, size_t num)
		: _func(f)
	{
		SetRange(min, max, num);
	}
	/**
	 * Recalculate internal table based on range variables. Functor specified in
	 * constructor will be used. 
	 * @param min minimum function range to interpolate.
	 * @param max maximum function range to interpolate.
	 * @param num number of points in internal function table. The bigger this
	 * number is, the more accurate the interpolation will be.
	 */
	void SetRange(T min, T max, size_t num)
	{
		_table.resize(num + 1); // room for guard point
		
		// This loop also creates the guard point
		for (size_t i = 0; i <= num; ++i) {
			T curval = min + (max - min)*i/T(num - 1);
			_table[i] = _func(curval);
		}
		_offset = min;
		_scale = (num - 1)/(max - min);
	}

	/**
	 * Do linear interpolation of function and return estimated result. NOTE:
	 * no out of bounds checking is performed, if the argument is out of range,
	 * a crash will very probably occur!
	 * @param x function argumnent.
	 * @return Interpolated value.
	 */
	T operator()(T x)
	{
		T pos = (x - _offset)*_scale;
		size_t x_0 = size_t(pos);
		size_t x_1 = x_0 + 1;
		T frac = pos - x_0;
		
		return _table[x_0] + frac*(_table[x_1] - _table[x_0]);
	}
};

#endif

