/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "JACKInterface.hpp"
#include "AmbisonicsDecoderGUI.hpp"
#include <algorithm>
#include <cstdio>
#include <errno.h>
#include <stdexcept>
#include <sstream>
#include <string>

static int times_locked = 0;

void JACKInterface::PeakScan(PortData& port, float* x, size_t num)
{
	// In case a peak hasn't been read last frame, let's include last frame's
	// peak while we look for a new one, so we don't skip any peaks.
	float peak = port.peak;

	for (size_t n = 0; n < num; ++n)
		if (fabsf(x[n]) > peak) peak = fabsf(x[n]);
	port.peak = peak;
	port.peak_read = false;
}

int JACKInterface::GlobalProcessCallback(jack_nframes_t num, void* arg)
{
	return static_cast<JACKInterface*>(arg)->ProcessCallback(num);
}

int JACKInterface::ProcessCallback(jack_nframes_t num)
{
	ScopedTryLock lock(_process_mutex);

	if (!_process)
		return 0;
	// Check if the processing mutex is locked, which it might be for two
	// reasons: 1. We're currently messing with our own internals, like
	// adding/removing ports/buffers. 2. AudioProcessor object is being
	// manipulated in some non-threadsafe way and we need to wait for it to
	// finish.
	if (!lock.Locked()) {
		times_locked++;
		return 0;
	}
	for (size_t i = 0; i < _inports.size(); ++i)
		_indata[i] = (float*)jack_port_get_buffer(_inports[i].port, num);
	// JACK allows us to cache output buffer pointers between calls to the
	// buffer size callbacks. The only times we need to fetch them again, is
	// when the buffer size callback has been called, or when we've added new
	// output ports.
	if (!_cached_ports_valid) {
		for (size_t i = 0; i < _outports.size(); ++i) {
			_outdata[i] = (float*)jack_port_get_buffer(_outports[i].port, num);
		}
		_cached_ports_valid = true;
	}

	// Find peak data from input frames
	if (_scan_peaks) {
		for (size_t i = 0; i < _indata.size(); ++i)
			PeakScan(_inports[i], _indata[i], num);
	}

	// Call our processing object
	_process->Process(&_indata[0], &_outdata[0], num);

	// Find peak data from output frames
	if (_scan_peaks) {
		for (size_t i = 0; i < _outdata.size(); ++i)
			PeakScan(_outports[i], _outdata[i], num);
	}
	return 0;
}


int JACKInterface::BufferSizeCallback(jack_nframes_t num, void* arg)
{
	JACKInterface* c = static_cast<JACKInterface*>(arg);

	c->_cached_ports_valid = false;
	if (c->_process)
		c->_process->SetBufferSize(num);
	return 0;
}

JACKInterface::JACKInterface(char* clientname) : _cached_ports_valid(false),
                                                 _scan_peaks(false)
{
	DisableProcessing();
	_client = jack_client_open(clientname, JackNullOption, NULL);
	if (!_client)
		throw std::runtime_error("Could not connect to JACK server.");
	jack_set_process_callback(_client, GlobalProcessCallback, this);
	jack_set_buffer_size_callback(_client, BufferSizeCallback, this);
	jack_activate(_client);
}

void JACKInterface::SetNumPorts(size_t num, PortVector& pv, ArrayVector& av,
                                const char* prefix, unsigned long flags)
{
	char name[255];
	size_t oldsize = pv.size();
	// We disable processing so we don't get interrupted by the process
	// callback in the middle of this, and call Process() with an arbitrary
	// number of channels.
	ScopedLock lock(_process_mutex);

	if (num > oldsize) {
		pv.resize(num);
		for (size_t i = oldsize; i < num; ++i) {
			snprintf(name, sizeof(name), "%s_%zu", prefix, i + 1);
			#ifdef DEBUG
			printf("adding port %s, index:%zu\n", name, i);
			#endif
			pv[i].port = jack_port_register(_client, name,
			                                JACK_DEFAULT_AUDIO_TYPE, flags, 0);
			if (pv[i].port == NULL)
				throw std::runtime_error("jack_port_register failed! "
				    "Try to raise max port number in JACK.");
		}
	} else if (oldsize > num) {
		for (size_t i = num; i < oldsize; ++i) {
			#ifdef DEBUG
			printf("deleting port %zu\n", i + 1);
			#endif
			jack_port_unregister(_client, pv[i].port);
		}
		pv.resize(num);
	}
	av.resize(num);
}

void JACKInterface::SetNumInputs(size_t num)
{
	SetNumPorts(num, _inports, _indata, "input", JackPortIsInput);
}

void JACKInterface::SetNumOutputs(size_t num)
{
	ScopedLock lock(_process_mutex);  // Protect the _cached_ports_valid variable
	SetNumPorts(num, _outports, _outdata, "output", JackPortIsOutput);
	_cached_ports_valid = false;
}

void JACKInterface::UpdatePorts()
{
	if (!_process)
		return;
	size_t inputs = _process->GetNumInputs();
	size_t outputs = _process->GetNumOutputs();
	if (inputs != _inports.size())
		SetNumInputs(inputs);
	if (outputs != _outports.size())
		SetNumOutputs(outputs);
}

void JACKInterface::ConnectPorts(std::string name, PortVector& ports,
                                 unsigned long flags)
{
	// Find the application part of the name
	std::string::size_type colon_pos = name.find(':');
	if (colon_pos == std::string::npos) return; // No separator
	std::string app(name, 0, colon_pos);
	std::string portname(name.begin() + colon_pos, name.end());
	// Get all port names that start with our app name. This pointer
	// also needs to be freed before we exit.
	app += ".*";
	const char** p = jack_get_ports(_client, app.c_str(), NULL, flags);
	
	if (p == NULL)
		return;  // No matches, exit
	// Separate the app name from the port name
	colon_pos = 0;
	while (p[0][colon_pos] != ':') ++colon_pos;
	std::string appname(p[0], colon_pos);
	
	for (size_t portnum = 0; portnum < ports.size(); ++portnum) {
		int err;
		std::ostringstream name;
		name.str("");
		name << appname << portname << portnum + 1;
		if (flags & JackPortIsInput) {
			err = jack_connect(_client, jack_port_name(ports[portnum].port),
			                   name.str().c_str());
		} else {
			err = jack_connect(_client, name.str().c_str(),
			                   jack_port_name(ports[portnum].port));
		}
		// If connection failed, break out (already existing connection is not
		// an error)
		if (err != 0 && err != EEXIST)
			break;
	}
	free(p);
}

void JACKInterface::ConnectInputs(std::string name)
{
	ConnectPorts(name, _inports, JackPortIsOutput);
}

void JACKInterface::ConnectOutputs(std::string name)
{
	ConnectPorts(name, _outports, JackPortIsInput);
}

void JACKInterface::SetProcess(AudioProcessor& o)
{
	_cached_ports_valid = false;
	_process = &o;
	o.SetSampleRate(jack_get_sample_rate(_client));
	o.SetBufferSize(jack_get_buffer_size(_client));
	UpdatePorts();
}

bool JACKInterface::GetPeak(size_t chan, float& peak, bool output)
{
	PortVector& p = output ? _outports : _inports;
	bool read = p[chan].peak_read;

	peak = p[chan].peak;
	p[chan].peak = 0.f;
	p[chan].peak_read = true;
	return !read;
}

void JACKInterface::FindPeaks(bool enable)
{
	_scan_peaks = enable;
}

JACKInterface::~JACKInterface()
{
	SetNumInputs(0);
	SetNumOutputs(0);
	jack_deactivate(_client);
	jack_client_close(_client);
//	printf("times locked: %i\n", times_locked);
}

