/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _jackinterface_h
#define _jackinterface_h

#include "AudioProcessor.hpp"
#include "Mutex.hpp"
#include <jack/jack.h>
#include <string>
#include <vector>

/**
 * \brief Class for interacting with JACK server and audio processing objects.
 */
class JACKInterface : public AudioInterface
{
	struct PortData
	{
		jack_port_t* port;
		float peak;
		bool peak_read;
		PortData() : peak(0), peak_read(false) { }
	};
	
	typedef std::vector<PortData> PortVector;
	typedef std::vector<float*> ArrayVector;

	PortVector _inports, _outports;
	ArrayVector _indata, _outdata;
	jack_client_t* _client;
	bool _cached_ports_valid;
	bool _scan_peaks;
	/**
	 * Actual static JACK processing callback. Chains directly to
	 * ProcessCallback.
	 * @param num number of samples to process.
	 * @param arg pointer to the instance of JACKInterface this applies to.
	 * @return Value of zero on success, non-zero on failure.
	 */
	static int GlobalProcessCallback(jack_nframes_t num, void* arg);
	/**
	 * JACK audio processing callback. Gets called whenever audio is to be
	 * processed. This again calls the AudioProcessor class specified with
	 * SetProcess().
	 * This method is static since JACK is a C based library, and has no notion
	 * of classes.
	 * @param num number of samples to process.
	 * @param arg pointer to the instance of JACKInterface this applies to.
	 * @return Value of zero on success, non-zero on failure.
	 */
	int ProcessCallback(jack_nframes_t num);
	/**
	 * JACK buffer size change callback. Gets called whenever JACK changes
	 * the buffer size. This again calls the SetBufferSize() method in the
	 * class specified with SetProcess().
	 * @param num new buffer size.
	 * @param arg pointer to the instance of JACKInterface this applies to.
	 * @return Value of zero on success, non-zero on failure.
	 */
	static int BufferSizeCallback(jack_nframes_t num, void* arg);
	/**
	 * Set number of input ports.
	 * @param num number of input ports.
	 */
	void SetNumInputs(size_t num);
	/**
	 * Set number of output ports.
	 * @param num number of output ports.
	 */
	void SetNumOutputs(size_t num);
	/**
	 * Set the number of ports, either input or output. Meant for internal use.
	 * @param num number of ports to use.
	 * @param pv vector in which to store the ports returned from JACK.
	 * @param av vector in which to store the address of the buffers used by the
	 * the ports.
	 * @param prefix name of ports to add. This name will be suffixed with the
	 * port number and used directly as the port name.
	 * @param flags flags to pass to JACK. Usually signifies whether to create
	 * input or output ports.
	 */
	void SetNumPorts(size_t num, PortVector& pv, ArrayVector& av,
	                 const char* prefix, unsigned long flags);
	/**
	 * Attempt to connect all available output or input ports to another range
	 * of ports belonging to another JACK application. Will only connect as many
	 * ports as is possible.
	 * @param appnamebase application name to which we will try to connect our
	 * ports. If multiple applications starting with this name is found, the
	 * first one found will be used.
	 * @param portname port name to which we will try to connect. Numerical
	 * order is assumed.
	 * @param ports vector of ports to use for connecting.
	 * @param flags what port type belonging to other program to look for.
	 * Should be either JackPortIsInput or JackPortIsOutput.
	 */
	void ConnectPorts(std::string name, PortVector& ports, unsigned long flags);
	/**
	 * Scan a channel looking for the sample with the maximum absolute value.
	 * @param port PortData entry to use store peak data in.
	 * @param x data array to scan in.
	 * @param num number of samples to scan.
	 */
	static void PeakScan(PortData& port, float* x, size_t num);
public:
	/**
	 * Class constructor. Throws exceptions of type std::runtime_error when
	 * errors occur.
	 * @param clientname name that will be used for registering with JACK.
	 */
	JACKInterface(char* clientname);
	void UpdatePorts();
	/**
	 * Attempts to connect current input ports to another JACK program specified
	 * by a name and port combination.
	 * @param name string describing name of program to connect to, plus port
	 * prefix. Example: decoder*:input_. Wildcards are only allowed prior to the
	 * colon sign. Ports are assumed to be named in numerical order starting at
	 * 1.
	 */
	void ConnectInputs(std::string name);
	/**
	 * Attempts to connect current output ports to another JACK program
	 * specified by a name and port combination.
	 * @see ConnectInputs
	 */
	void ConnectOutputs(std::string name);
	void SetProcess(AudioProcessor& o);
	/**
	 * Class destructor.
	 */
	void FindPeaks(bool enable = true);
	bool GetPeak(size_t chan, float& peak, bool output);
	~JACKInterface();
};

#endif

