/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "Ambisonics.hpp"
#include "LDC.hpp"
#include <algorithm>

#ifdef __SSE__
#include <xmmintrin.h>
#endif

LDC::LDC() : _is3d(false), _order(1), _R1(1.), _R2(1.)
{
#ifdef __SSE__
	// If there's a chance we'll be using SSE for calculations, let's disable
	// denormal processing.
	_MM_SET_FLUSH_ZERO_MODE(_MM_FLUSH_ZERO_ON);
#endif
}

void LDC::SetDistances(double R1, double R2)
{
	for (size_t i = 0; i < _filters.size(); ++i)
		_filters[i].SetDistances(R1, R2);
	_R1 = R1;
	_R2 = R2;
}

void LDC::SetIs3D(bool is3d)
{
	_is3d = is3d;
	SetOrder(_order); // Might need more/less filters now
	SetDistances(_R1, _R2);
}

void LDC::SetOrder(unsigned order)
{
	size_t i = 0;

	_order = order;
	// components needed for current order
	size_t num_comp = Ambisonics::NumComponents(_order, _is3d);

	// Start at order 1, order 0 needs no filter
	_filters.resize(num_comp - 1, NFCFilter(_sample_rate));
	for (unsigned cur_order = 1; cur_order < _order; ++cur_order) {
		// Find out how many components to use with cur_order order filter
		size_t new_comp = Ambisonics::OrderComponents(cur_order, _is3d);
		for (size_t j = 0; j < new_comp; ++j)
			_filters[i++].SetOrder(cur_order);
	}
	SetDistances(_R1, _R2);
}
	
void LDC::Process(float** inputs, float** outputs, unsigned num)
{
	// First components needs no modifying, pass through
	std::copy(inputs[0], inputs[0] + num, outputs[0]);
	for (size_t i = 0; i < _filters.size(); ++i)
		_filters[i].Process(inputs[i + 1], outputs[i + 1], num);
}

void LDC::SetSampleRate(unsigned sr)
{
	AudioProcessor::SetSampleRate(sr);
	for (size_t i = 0; i < _filters.size(); ++i) {
		_filters[i].SetSampleRate(sr);
		_filters[i].SetDistances(_R1, _R2);
	}
}

