/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "LDCGUI.hpp"
#include "Widgets.hpp"
#include <gtkmm/aboutdialog.h>
#include <gtkmm/box.h>
#include <gtkmm/table.h>

Glib::ustring LDCGUI::_menu_info =
	"<ui>"
	"  <menubar name='MenuBar'>"
	"    <menu action='FileMenu'>"
	"      <menuitem action='FileQuit'/>"
	"    </menu>"
	"    <menu action='RoutingMenu'>"
	"      <menuitem action='RoutingInputs'/>"
	"      <menuitem action='RoutingOutputs'/>"
	"    </menu>"
	"    <menu action='HelpMenu'>"
	"      <menuitem action='HelpAbout'/>"
	"    </menu>"
	"  </menubar>"
	"</ui>";

LDCGUI::LDCGUI(JACKInterface& ai) 
	: _adj_order(1., 0., 15., 1., 3.), _adj_radius1(2., 0.1, 20., 0.1, 1.),
	  _adj_radius2(2., 0.1, 20., 0.1, 1.),
	  _btn_order(_adj_order, 0., 0), _btn_3d("3D"),
	  _sld_radius1(_adj_radius1), _sld_radius2(_adj_radius2),
	  _btn_radius1(_adj_radius1, 0., 1), _btn_radius2(_adj_radius2, 0., 1),
	  _btn_inf1("Infinite"), _snd(ai)
{
	set_title("Loudspeaker distance compensator");
	set_default_size(400, 200);
	
	// Make containers
	Gtk::Table* table = Gtk::manage(new Gtk::Table(3, 5));
	Gtk::VBox* vbox1 = Gtk::manage(new Gtk::VBox());
	// Construct our menu. This corresponds to the XML string at the top.
	_actiongroup = Gtk::ActionGroup::create();
	// File menu
	_actiongroup->add(Gtk::Action::create("FileMenu", "File"));
	_actiongroup->add(Gtk::Action::create(
		"FileQuit", Gtk::Stock::QUIT, "_Quit", "Quit"),
		sigc::mem_fun(*this, &LDCGUI::on_quit));
	// Routing menu
	_actiongroup->add(Gtk::Action::create("RoutingMenu", "Routing"));
	_actiongroup->add(Gtk::Action::create(
		"RoutingInputs", "Auto-connect _inputs", "Auto-connect JACK inputs"),
		sigc::mem_fun(*this, &LDCGUI::on_route_inputs));
	_actiongroup->add(Gtk::Action::create(
		"RoutingOutputs", "Auto-connect _outputs", "Auto-connect JACK outputs"),
		sigc::mem_fun(*this, &LDCGUI::on_route_outputs));

	// Help menu
	_actiongroup->add(Gtk::Action::create("HelpMenu", "Help"));
	_actiongroup->add(Gtk::Action::create(
		"HelpAbout", "_About", "About program"),
		sigc::mem_fun(*this, &LDCGUI::on_view_aboutbox));

	_uimanager = Gtk::UIManager::create();
	_uimanager->insert_action_group(_actiongroup);
	add_accel_group(_uimanager->get_accel_group());
	_uimanager->add_ui_from_string(_menu_info);
	
	Gtk::Widget* menu = _uimanager->get_widget("/MenuBar");
	vbox1->pack_start(*menu, Gtk::PACK_SHRINK);

	// Create main GUI
	table->set_col_spacings(20);
	table->set_border_width(16);
	Gtk::Label* label1 = Gtk::manage(new Gtk::Label("Order:"));
	label1->set_alignment(Gtk::ALIGN_RIGHT);
	Gtk::Label* label2 = Gtk::manage(new Gtk::Label("Original\ndistance:"));
	label2->set_alignment(Gtk::ALIGN_RIGHT);
	Gtk::Label* label3 = Gtk::manage(new Gtk::Label("New\ndistance:"));
	label3->set_alignment(Gtk::ALIGN_RIGHT);
	table->attach(*label1, 0, 1, 0, 1, Gtk::FILL);
	table->attach(*label2, 0, 1, 1, 2, Gtk::FILL);
	table->attach(*label3, 0, 1, 2, 3, Gtk::FILL);
	table->attach(_btn_order, 1, 2, 0, 1, Gtk::EXPAND);
	table->attach(_btn_3d, 2, 3, 0, 1, Gtk::FILL);
	table->attach(_sld_radius1, 1, 3, 1, 2);
	table->attach(_btn_radius1, 3, 4, 1, 2, Gtk::FILL);
	table->attach(_btn_inf1, 4, 5, 1, 2, Gtk::FILL);
	table->attach(_sld_radius2, 1, 3, 2, 3);
	table->attach(_btn_radius2, 3, 4, 2, 3, Gtk::FILL);
	vbox1->pack_start(*table);

	_btn_radius1.set_digits(2);
	_btn_radius2.set_digits(2);
	_sld_radius1.set_draw_value(false);
	_sld_radius2.set_draw_value(false);
	
	// Connect signal handlers
	_btn_order.signal_value_changed().connect(sigc::mem_fun(
		*this, &LDCGUI::on_order_changed));
	_btn_3d.signal_clicked().connect(sigc::mem_fun(
		*this, &LDCGUI::on_3d_clicked));
	_adj_radius1.signal_value_changed().connect(sigc::mem_fun(
		*this, &LDCGUI::on_radius_changed));
	_adj_radius2.signal_value_changed().connect(sigc::mem_fun(
		*this, &LDCGUI::on_radius_changed));
	_btn_inf1.signal_clicked().connect(sigc::mem_fun(
		*this, &LDCGUI::on_inf_clicked));
	add(*vbox1);
	show_all_children();
	_snd.SetProcess(_ldc);
	
	// Call the signal handlers to have them update our LDC object
	// based on the default values in the widgets
	on_order_changed();
	on_3d_clicked();
	on_radius_changed();
	_snd.EnableProcessing();
}

void LDCGUI::on_order_changed()
{
	AudioInterface::Disable lock(_snd);

	_ldc.SetOrder(_btn_order.get_value_as_int());
	_snd.UpdatePorts();
}

void LDCGUI::on_3d_clicked()
{
	AudioInterface::Disable lock(_snd);

	_ldc.SetIs3D(_btn_3d.get_active());
	_snd.UpdatePorts();
}

void LDCGUI::on_radius_changed()
{
	AudioInterface::Disable lock(_snd);
	
	if (_btn_inf1.get_active())
		_ldc.SetDistances(NFCFilter::MaxDistance, _adj_radius2.get_value());
	else		
		_ldc.SetDistances(_adj_radius1.get_value(), _adj_radius2.get_value());
}

void LDCGUI::on_inf_clicked()
{
	bool inf = _btn_inf1.get_active();

	_sld_radius1.set_sensitive(!inf);
	_btn_radius1.set_sensitive(!inf);
	on_radius_changed();
}

void LDCGUI::on_quit()
{
	hide();
}

#include "LicenseText.hpp"

void LDCGUI::on_view_aboutbox()
{
	Gtk::AboutDialog dlg;
	std::vector<Glib::ustring> authors;

	dlg.set_skip_taskbar_hint();
	authors.push_back("Thom Johansen");
	authors.push_back("Bård støfringsdal");
	authors.push_back("Peter Svensson");
	dlg.set_comments("An HOA Ambisonics loudspeaker distance compensator for JACK");
	dlg.set_version("0.2");
	dlg.set_authors(authors);
	dlg.set_copyright("Copyright © 2006-2007 Q2S, "
	                  "Norwegian University of Science and Technology");
	dlg.set_website("http://www.q2s.ntnu.no");
	dlg.set_license(license_text);
	dlg.run();
}

void LDCGUI::on_route_inputs()
{
	const char* str[] = { "ambienc:output_", "alsa_pcm:capture_" };
	StringInputDialog input("Target ports", str, sizeof(str)/sizeof(char*));

	if (input.Run())
		_snd.ConnectInputs(input.GetString());
}	

void LDCGUI::on_route_outputs()
{
	const char* str[] = { "ambidec:input_" };
	StringInputDialog input("Target ports", str, sizeof(str)/sizeof(char*));

	if (input.Run())
		_snd.ConnectOutputs(input.GetString());
}

