/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "LevelMeter.hpp"
#include "iec268-scale.h"
#include <cmath>

#define METER_BITMAP_W 37
#define METER_BITMAP_H 210
#define METER_BAR_W 9
#define METER_BAR_X 7
#define METER_BAR_Y 205

#ifdef HAVE_CAIRO
Cairo::RefPtr<Cairo::ImageSurface> LevelMeter::_bitmap;
#else
Glib::RefPtr<Gdk::Pixbuf> LevelMeter::_bitmap;
#endif

LevelMeter::LevelMeter() : _level(0.), _peak(0.)
{
}

void LevelMeter::set_level(double l)
{
	const double db = 20.*log10(fabs(l));
	double def;

	if (db < -70.)
		def = 0.;
	else if (db < -60.)
		def = (db + 70.)*0.25;
	else if (db < -50.)
		def = (db + 60.)*0.5 + 5.;
	else if (db < -40.)
		def = (db + 50.)*0.75 + 7.5;
	else if (db < -30.)
		def = (db + 40.)*1.5 + 15.;
	else if (db < -20.)
		def = (db + 30.)*2. + 30.;
	else if (db < 0.)
		def = (db + 20.)*2.5 + 50.;
	else
		def = 100.;

    _level = def + def;
	// Peak hold time of two seconds
	if (_level >= _peak || _peak_timer.elapsed() - _peak_time > 2.) {
		_peak = _level;
		_peak_time = _peak_timer.elapsed();
	}
}

void LevelMeter::on_realize()
{
	Gtk::DrawingArea::on_realize();
	_window = get_window();

#ifndef HAVE_CAIRO
	Glib::RefPtr<Gdk::Colormap> colourmap = get_default_colormap();

	_fgcolour = Gdk::Color("red");
	_peakcolour = Gdk::Color("white");
	colourmap->alloc_color(_fgcolour);
	colourmap->alloc_color(_peakcolour);

	_gc = Gdk::GC::create(_window);
#endif
}

bool LevelMeter::on_expose_event(GdkEventExpose* e)
{
#ifdef HAVE_CAIRO
	_cr = _window->create_cairo_context();
	if (e) {
		// Only redraw what we have to
		_cr->rectangle(e->area.x, e->area.y, e->area.width, e->area.height);
		_cr->clip();
	}
	if (!_bitmap) {
		unsigned char* img = reinterpret_cast<unsigned char*>(meter_bitmap);

		_bitmap = Cairo::ImageSurface::create(img, Cairo::FORMAT_ARGB32,
		                                      METER_BITMAP_W, METER_BITMAP_H,
		                                      METER_BITMAP_W*4);
	}
	_cr->set_identity_matrix();	// Use device coordinates directly

	_cr->set_source(_bitmap, 0, 0);
	_cr->rectangle(0, 0, METER_BITMAP_W, METER_BITMAP_H);
	_cr->fill();
	_cr->set_line_width(1.);
	_cr->set_source_rgb(METER_FGCOLOUR);
	_cr->rectangle(METER_BAR_X, METER_BAR_Y, METER_BAR_W, -_level);
	_cr->fill();
	if (_peak >= 1.) {
		_cr->set_source_rgb(METER_PEAKCOLOUR);
		// Subtract 0.5 from bar origin to make Cairo draw a one pixel wide line
		_cr->move_to(METER_BAR_X, METER_BAR_Y + 0.5 - ceil(_peak));
		_cr->rel_line_to(METER_BAR_W, 0.);
		_cr->stroke();
	}
#else
	if (e) {
		Gdk::Rectangle r(e->area.x, e->area.y, e->area.width, e->area.height);

		_gc->set_clip_rectangle(r);
	}
	if (!_bitmap) {
		const guint8* img = reinterpret_cast<unsigned char*>(meter_bitmap);

		_bitmap = Gdk::Pixbuf::create_from_data(img, Gdk::COLORSPACE_RGB, true,
		                                        8, METER_BITMAP_W,
		                                        METER_BITMAP_H,
		                                        METER_BITMAP_W*4);
	}
	int llev = int(ceil(_level));

	_window->draw_pixbuf(_gc, _bitmap, 0, 0, 0, 0, METER_BITMAP_W,
	                     METER_BITMAP_H, Gdk::RGB_DITHER_NONE, 0, 0);
	_gc->set_foreground(_fgcolour);
	_gc->set_line_attributes(1, Gdk::LINE_SOLID, Gdk::CAP_NOT_LAST,
	                         Gdk::JOIN_MITER);
	_window->draw_rectangle(_gc, true, METER_BAR_X, METER_BAR_Y - llev,
	                        METER_BAR_W, llev);
	if (_peak >= 1.) {
		int y = METER_BAR_Y - int(ceil(_peak));

		_gc->set_foreground(_peakcolour);
		_window->draw_line(_gc, METER_BAR_X, y, METER_BAR_X + METER_BAR_W, y);
	}
#endif
	return true;
}

void LevelMeter::on_size_request(Gtk::Requisition* req)
{
	*req = Gtk::Requisition();
	req->height = METER_BITMAP_H;
	req->width = METER_BITMAP_W;
}

void LevelMeter::on_size_allocate(Gtk::Allocation& alloc)
{
	set_allocation(alloc);
	if (_window)
		_window->move_resize(alloc.get_x(), alloc.get_y(), alloc.get_width(),
		                     alloc.get_height());
}

