/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "MeterWindow.hpp"
#include <cstdio> // For snprintf()
#include <cmath>
#include <gtkmm/expander.h>

MeterWindow::MeterWindow(AudioInterface& snd) : _snd(snd), _num_imeters_disp(0),
                                                _num_ometers_disp(0)
{
	// Init window
	set_title("Level meters");
	set_default_size(600, 550);
	Gtk::ScrolledWindow* scrollwindow = Gtk::manage(new Gtk::ScrolledWindow());
	Gtk::VBox* vbox = Gtk::manage(new Gtk::VBox());
	Gtk::Expander* exp1 = Gtk::manage(new Gtk::Expander("<b>Inputs</b>"));
	Gtk::Expander* exp2 = Gtk::manage(new Gtk::Expander("<b>Outputs</b>"));

	exp1->set_expanded();
	exp1->set_use_markup();
	exp1->set_spacing(3);
	exp2->set_expanded();
	exp2->set_use_markup();
	exp2->set_spacing(3);

	scrollwindow->set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	scrollwindow->add(*vbox);
	vbox->pack_start(*exp1, Gtk::PACK_SHRINK);
	vbox->pack_start(*exp2, Gtk::PACK_SHRINK);
	exp1->add(_ibox);
	exp2->add(_obox);
	add(*scrollwindow);
	// The last parameter here controls the refresh time in ms
	Glib::signal_timeout().connect(
		sigc::mem_fun(*this, &MeterWindow::redraw_timer), 80);

	show_all_children();
}

bool MeterWindow::redraw_timer()
{
	float peak;

	for (size_t i = 0; i < _num_imeters_disp; ++i) {
		if (_snd.GetPeak(i, peak, false))
			_istrips[i]->queue_draw_level(peak);
	}
	for (size_t i = 0; i < _num_ometers_disp; ++i) {
		if (_snd.GetPeak(i, peak, true))
			_ostrips[i]->queue_draw_level(peak);
	}
	return true; // false stops timer
}

// Populate a table row with widgets
void MeterWindow::set_num_meters(std::vector<MeterStrip*>& strips,
                                 size_t& num_disp, Gtk::HBox& box, size_t num)
{
	size_t old_num = strips.size();

	if (num > old_num) {
		strips.resize(num);
		num_disp = num;
		// Create some new strip widgets
		for (size_t i = old_num; i < num; ++i) {
			char buf[8];

			snprintf(buf, 8, "%zu\n", i + 1);
			strips[i] = Gtk::manage(new MeterStrip(buf));
			box.pack_start(*strips[i], Gtk::PACK_SHRINK);
		}
	} else {
		num_disp = num;
	}

	// Now display only the ones we need
	for (size_t i = 0; i < num_disp; ++i)
		strips[i]->show();
	for (size_t i = num_disp; i < strips.size(); ++i)
		strips[i]->hide();
}

void MeterWindow::Notify(int what)
{
	// We won't ask the model what just happened, instead, we'll ask the audio
	// interface we were provided if the number of channels changed, which is
	// all the information we really care about here.
	(void)what;
	const size_t ochans = _snd.GetNumOutputs();
	const size_t ichans = _snd.GetNumInputs();

	if (ichans != _num_imeters_disp)
		set_num_meters(_istrips, _num_imeters_disp, _ibox, ichans);
	if (ochans != _num_ometers_disp)
		set_num_meters(_ostrips, _num_ometers_disp, _obox, ochans);
}

void MeterWindow::on_show()
{
	Gtk::Widget::on_show();
	_snd.FindPeaks(true);
}
 
void MeterWindow::on_hide()
{
	Gtk::Widget::on_hide();
	_snd.FindPeaks(false);
}
 
