/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _mutex_h
#define _mutex_h

#include <pthread.h>

/**
 * \brief Mutex class.
 *
 * Provides a single mutex (currently using the pthreads API, but should be
 * easy to port to other APIs as well).
 * Should be used with the ScopedLock and ScopedTryLock classes.
 */
class Mutex {
	pthread_mutex_t _mutex;
	pthread_mutexattr_t _mattr;
public:
	/**
	 * Describes a mutex type. Can either be Normal or Recursive, where a
	 * recursive mutex can be locked several times with no ill effects. Multiple
	 * locks of a normal mutex is undefined.
	 */
	enum Type { Normal, Recursive };
	/**
	 * Class constructor. Creates a single initialised mutex.
	 * @param type type of mutex.
	 */
	Mutex(Type type = Normal)
	{
		int t;

		pthread_mutexattr_init(&_mattr);
		if (type == Recursive)
			t = PTHREAD_MUTEX_RECURSIVE;
		else
			t = PTHREAD_MUTEX_NORMAL;
		pthread_mutexattr_settype(&_mattr, t);
		pthread_mutex_init(&_mutex, &_mattr);
	}

	/**
	 * Tries to lock the mutex and blocks until the mutex is free.
	 */
	void Lock()
	{
		pthread_mutex_lock(&_mutex);
	}

	/**
	 * Tries to lock the mutex and returns status of the mutex state. No
	 * blocking can occur.
	 * @return True if the mutex was free and is now locked by caller, returns
	 * false if mutex was taken and lock could not be done.
	 */
	bool Trylock()
	{
		return pthread_mutex_trylock(&_mutex) == 0;
	}

	/**
	 * Unlocks a previously locked mutex.
	 */
	void Unlock()
	{
		pthread_mutex_unlock(&_mutex);
	}

	/**
	 * Class destructor.
	 */
	~Mutex()
	{
		pthread_mutex_destroy(&_mutex);
		pthread_mutexattr_destroy(&_mattr);
	}
};

/**
 * \brief A scoped lock class for use with the Mutex class.
 *
 * Used in a scope, this class will lock a specified mutex, which stays locked
 * until the end of the scope, where the destructor will be called, the
 * mutex being unlocked as a consequence.
 * If the mutex is already locked upon construction, the constructor will block
 * until it is free.
 */
class ScopedLock {
	Mutex& _mutex;
public:
	/**
	 * Class constructor. Locks a mutex.
	 * @param m reference to Mutex object to be locked.
	 */
	ScopedLock(Mutex& m) : _mutex(m)
	{
		_mutex.Lock();
	}

	/**
	 * Class destructor. Unlocks mutex specified in constructor.
	 */
	~ScopedLock()
	{
		_mutex.Unlock();
	}
};

/**
 * \brief A scoped try-lock class for use with the Mutex class.
 *
 * Used in a scope, this class will lock a specified mutex, which stays locked
 * until the end of the scope, where the destructor will be called, the
 * mutex being unlocked as a consequence.
 * If the mutex is already locked upon construction, the constructor will 
 * immediately exit and the mutex will stay unlocked, which makes it necessary
 * to investigate if the lock succeeded before accessing protected data.
 * If executed in a context where any blocking would be undesirable, use this
 * class instead of ScopedLock.
 */
class ScopedTryLock {
	Mutex& _mutex;
	bool _locked;
public:
	/**
	 * Class constructor. Tries to lock a mutex.
	 * @param m reference to Mutex object to be locked.
	 */
	ScopedTryLock(Mutex& m) : _mutex(m)
	{
		_locked = _mutex.Trylock();
	}

	/**
	 * Check to see if lock succeeded.
	 * @returns true if lock succeeded.
	 */
	bool Locked() const
	{
		return _locked;
	}

	/**
	 * Class destructor. Unlocks mutex specified in constructor, if lock
	 * succeeded at construction time.
	 */
	~ScopedTryLock()
	{
		if (_locked)
			_mutex.Unlock();
	}
};

#endif

