/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _nfcfilter_h
#define _nfcfilter_h

#include "IIRFilter.hpp"

/**
 * \brief Ambisonics near-field compensation filter.
 *
 * This filter is used to either compensate for the finite distance of both
 * source and loudspeaker array, or to convert an ambisonics stream with one
 * reference loudspeaker array radius to another radius.
 * Based on equations 28 and 29 in:
 * Jérôme Daniel - Spatial Sound Encoding Including Near Field Effect:
 * Introducing Distance Coding Filters and a Viable, New Ambisonic Format.
 */ 
class NFCFilter : public IIRFilter<float> {
private:
	static const double Xmq[];
	std::vector<double> _rcp_a0_cache;
	unsigned _order;
	double _c; // speed of sound
	double _alpha_mult; // 4*Fs/c
public:
	/**
	 * Distance at which the NFC filter starts constructing a pure plane
	 * wave front curvature. All filters constructed with a radius beyond this
	 * will yield identical coefficients.
	 */
	enum { MaxDistance = 10000 };
	/**
	 * Constructs one near-field compensation filter.
	 * @param sr sample rate.
	 * @param c speed of sound in the current system. Defaults to 344 m/s.
	 */
	NFCFilter(unsigned sr, double c = 344.0);
	/**
	 * Sets the distances which will be used in near-field compensation.
	 * @param R1 radius used to calculate filter zeroes. Usually either source
	 * distance when used for near-field compensation, or old loudspeaker
	 * array radius when used for loudspeaker array radius adaptation. Assumed
	 * positive and non-zero.
	 * @param R2 radius used to calculate filter poles. Usually either current
	 * loudspeaker array radius when used for near-field correction, or new
	 * loudspeaker array radius when used for loudspeaker array radius
	 * adaptation. Assumed positive and non-zero.
	 */
	void SetDistances(double R1, double R2);
	/**
	 * Faster version of SetDistances which just sets the source distance,
	 * keeping the speaker distance the same. This depends on internal state
	 * set by SetDistances, so that must be called first.
	 * @param R1 source radius.
	 */
	void SetDistance(double R1);
	/**
	 * Sets the Ambisonics order to be used. Must always be called before
	 * the first use of SetDistances().
	 * @param m ambisonics order.
	 */
	void SetOrder(unsigned m);
	/**
	 * Sets the sample rate. Called whenever one wishes to change the sample
	 * rate from what was specified in the constructor. Note that a call to
	 * SetDistances() is needed to update the filter coefficients after this
	 * method has been called.
	 * @param sr sample rate.
	 */
	void SetSampleRate(unsigned sr);
};

#endif

