/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _positioninterpolator_h
#define _positioninterpolator_h

#include <algorithm>
#include <cmath>
#include "align16_allocator.h"
#include "Defs.hpp"
#include "ParameterFilter.hpp"
#include "PositionVector.hpp"

class PositionInterpolator {
	struct Filter : public ParameterFilter {
		size_t samples;
		double value;
		Filter(double scale) : ParameterFilter(scale), samples(0), value(0.)
		{
		}
	};

	struct AzimuthFilter : public Filter {
		int phasewraps;
		double wrappedphase;
		AzimuthFilter(double scale) : Filter(scale), phasewraps(0),
		                              wrappedphase(0.)
		{
		}
	};

	struct PositionFilters {
		AzimuthFilter azimuth;
		Filter elev, radius;
		PositionFilters(double scale) : azimuth(scale), elev(scale),
		                                radius(scale)
		{
		}
	};

	std::vector<PositionFilters> _filters;
	unsigned _samples_per_delta;
	double _scale;

	/**
	 * Wraps phase to -PI .. PI. Works better for this usage scenario than
	 * x - floor((x + pi)/(2*pi))*2*pi, which will also wrap PI to -PI, which we
	 * don't want.
	 * @param phase to wrap.
	 * @return Wrapped phase.
	 */
	double phasewrap(double wrappedval)
	{
		double a = wrappedval > M_PI ? -2.*M_PI : 2.*M_PI;

		while (wrappedval > M_PI || wrappedval <= -M_PI)
			wrappedval += a;
		return wrappedval;
	}
public:
	PositionInterpolator()
	{
		SetScale(10000.);
	}

	void SetScale(double scale)
	{
		_scale = scale;
		_samples_per_delta = unsigned(scale*4);
		for (size_t i = 0; i < _filters.size(); ++i) {
			_filters[i].azimuth.SetScale(scale);
			_filters[i].elev.SetScale(scale);
			_filters[i].radius.SetScale(scale);
		}
	}

	void SetAzimuth(size_t source, double azimuth, bool interp = true)
	{
		AzimuthFilter& afilter = _filters[source].azimuth;

		// Get current position wrapped to interval -pi .. pi
		// Ideally, we should use the current interpolated position here, but
		// this is much simpler and should work just fine
		double curpos = afilter.wrappedphase;
		if (interp) {
			afilter.samples = _samples_per_delta;
			// Check if we've done a transition across +/- pi. If this is the
			// case we want to unwrap the phase across so that the interpolated
			// trajectory can go behind the listener instead of having to go
			// around her again. This will yield an azimuth value bigger or
			// smaller than +/- pi, but we'll wrap it back inside this interval
			// again when the current interpolation is done
			// (in InterpolateAzimuth()).
			if (azimuth - curpos <= -M_PI)
				++afilter.phasewraps;
			else if (azimuth - curpos >= M_PI)
				--afilter.phasewraps;
			// Store the wrapped phase for when we'll need it
			afilter.wrappedphase = azimuth;
			azimuth += double(afilter.phasewraps)*2.*M_PI;
		} else {
			afilter.SetPos(azimuth);
		}
		afilter.value = azimuth;
	}

	void SetElevation(size_t source, double elev, bool interp = true)
	{
		Filter& efilter = _filters[source].elev;

		efilter.value = elev;
		if (interp)
			efilter.samples = _samples_per_delta;
		else
			efilter.SetPos(elev);
	}

	void SetRadius(size_t source, double radius, bool interp = true)
	{
		Filter& rfilter = _filters[source].radius;

		rfilter.value = radius;
		if (interp)
			rfilter.samples = _samples_per_delta;
		else
			rfilter.SetPos(radius);
	}

	void SetNumSources(size_t num)
	{	
		_filters.resize(num, PositionFilters(_scale));
	}

	bool Interpolate(Filter& f, unsigned num, FloatVector& res)
	{
		unsigned samples = f.samples;

		if (!samples)
			return false;

		f.Process(&res[0], f.value, num);
		f.samples -= std::min(samples, num);
		return true;
	}

	bool InterpolateAzimuth(size_t source, unsigned num, FloatVector& v)
	{
		AzimuthFilter& afilter =_filters[source].azimuth;
		const bool ret = Interpolate(afilter, num, v);

		if (!ret) {
			// We're done interpolating for now. Let's wrap the azimuth back
			// to within -pi .. pi bounds so it doesn't grow completely out of
			// proportion.
			double wrappedpos = afilter.wrappedphase;

			afilter.phasewraps = 0;
			afilter.SetPos(wrappedpos);
			afilter.value = wrappedpos;
		}
		return ret;	
	}

	bool InterpolateElevation(size_t source, unsigned num, FloatVector& v)
	{
		return Interpolate(_filters[source].elev, num, v);
	}

	bool InterpolateRadius(size_t source, unsigned num, FloatVector& v)
	{
		return Interpolate(_filters[source].radius, num, v);
	}

	double GetAzimuth(size_t source) const
	{
		return _filters[source].azimuth.GetPos();
	}

	double GetElevation(size_t source) const
	{
		return _filters[source].elev.GetPos();
	}

	double GetRadius(size_t source) const
	{
		return _filters[source].radius.GetPos();
	}
};

#endif

