/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "PositionMap.hpp"
#include <algorithm>
#include <cmath>

PositionMap::PositionMap() : 
#ifdef HAVE_CAIRO
                             _dashpat(4., 1),
#endif
                             _coords(1), _referencedist(1.),
                             _lbutton(false), _rbutton(false), _active(0),
                             _constrain_radius(false)
{
	add_events(Gdk::BUTTON_PRESS_MASK | Gdk::BUTTON_RELEASE_MASK
	           | Gdk::BUTTON_MOTION_MASK);
	_max_z = _max_radius = 1.;
	_range_x1 = _range_y1 = -1.;
	_range_x2 = _range_y2 = 1.;
}

void PositionMap::set_constrain_radius(bool c)
{
	_constrain_radius = c;
}

void PositionMap::set_num_pos(size_t num)
{
	_coords.resize(num);
	_lbutton = _rbutton = false;
	if (_active >= num)
		_active = num - 1;
	queue_draw();
}

void PositionMap::set_active_pos(size_t num)
{
	if (_active == num)
		return;
	_lbutton = _rbutton = false;
	_active = num;
	queue_draw();
}

void PositionMap::get_pos(size_t num, double& p1, double& p2, double& p3)
{
	const double x = _coords[num].x;
	const double y = _coords[num].y;
	const double z = _coords[num].z;

	p1 = atan2(x, y)*180./M_PI;
	p3 = sqrt(x*x + y*y + z*z);
	p2 = asin(z/p3)*180./M_PI;
}	

void PositionMap::get_posxyz(size_t num, double& p1, double& p2, double& p3)
{
	p1 = _coords[num].x;
	p2 = _coords[num].y;
	p3 = _coords[num].z;
}	

// Doesn't do any clipping
void PositionMap::set_pos(size_t num, double p1, double p2, double p3)
{
	double x, y, z;
	double xd, yd, zd;

	p1 *= M_PI/180.;
	p2 *= M_PI/180.;
	x = p3*cos(p2)*sin(p1);
	y = p3*cos(p1)*cos(p2);
	z = p3*sin(p2);
	// All the conversion is making it a bit hard to see if the position actually
	// changed. If we're getting the results from a get_pos() flung back at us,
	// there'll almost certainly be some roundoff error here making the converted
	// cartesian coords not match. We solve this by disregarding differences
	// under a certain threshold.
	xd = fabs(_coords[num].x - x);
	yd = fabs(_coords[num].y - y);
	zd = fabs(_coords[num].z - z);
	// Only bother doing anything if internal state changed enough
	if (xd > 1e-5 || yd > 1e-5 || zd > 1e-5) {
		_coords[num].x = x;
		_coords[num].y = y;
		_coords[num].z = z;
		queue_draw();
		_signal_pos_changed.emit(num);
	}
}

// We don't currently fix any positions that are outside this
void PositionMap::set_max_radius(double max_radius)
{
	_max_radius = max_radius;
}

void PositionMap::set_max_height(double max_height)
{
	_max_z = max_height;
}

void PositionMap::set_reference_dist(double d)
{
	_referencedist = d;
	queue_draw();
}

void PositionMap::set_rect(double x1, double x2, double y1, double y2)
{
	_range_x1 = x1;
	_range_x2 = x2;
	_range_y1 = y1;
	_range_y2 = y2;
	queue_draw();
}

void PositionMap::draw_pos(size_t num)
{
	char buf[8];
	int xd, yd;
	int rad;
	const double z = _coords[num].z;

	// Calculate radius of on-screen pos based on height. This radius value
	// will be stored away for later use in detecting if the mouse cursor has
	// hit a pos.
	if (z > 0.)
		rad = int(10 + 10*z/_max_z);
	else
		rad = int(5 + 5*(_max_z + z)/_max_z);
	_coords[num].radius = rad;

	map_to_win(_coords[num].x, _coords[num].y, xd, yd);
	// Draw pos 
	snprintf(buf, sizeof(buf), "%zu", num + 1);
#ifdef HAVE_CAIRO
	_cr->set_line_width(2.);
	if (num == _active)
		_cr->set_source_rgb(POSMAP_APOSCOLOUR);
	else
		_cr->set_source_rgb(POSMAP_POSCOLOUR);
	_cr->arc(xd, yd, rad, 0., 2.*M_PI);
	_cr->stroke();
	_cr->move_to(xd + rad, yd - 5. - rad);
	_cr->set_font_size(12.);
	_cr->show_text(buf);
	_cr->fill();
	_cr->arc(xd, yd, 1, 0, 2.*M_PI);
	_cr->fill();
#else
	_postextlayout->set_text(buf);
	_gc->set_foreground(num == _active ? _aposcolour : _poscolour);
	_gc->set_line_attributes(1, Gdk::LINE_SOLID, Gdk::CAP_NOT_LAST,
	                         Gdk::JOIN_MITER);
	_window->draw_arc(_gc, false, xd - rad, yd - rad, rad*2, rad*2,
	                  0, 360*64);
	_window->draw_layout(_gc, xd + rad, yd - 15 - rad, _postextlayout);
	_window->draw_rectangle(_gc, true, xd - 1, yd - 1, 2, 2);
#endif
	// Draw line from centre of space to centre of pos if we're moving it
	if ((_lbutton || _rbutton) && _active == num) {
		int xcentre, ycentre;

		map_to_win(0., 0., xcentre, ycentre);
#ifdef HAVE_CAIRO
		_cr->set_source_rgb(POSMAP_APOSCOLOUR);
		_cr->set_line_width(1.);
		_cr->set_dash(_dashpat, 0.);
		_cr->move_to(xcentre, ycentre);
		_cr->line_to(xd, yd);
		_cr->stroke();
		_cr->unset_dash();
#else
		_gc->set_foreground(_aposcolour);
		_gc->set_line_attributes(2, Gdk::LINE_DOUBLE_DASH, Gdk::CAP_NOT_LAST,
	                         Gdk::JOIN_MITER);
		_window->draw_line(_gc, xcentre, ycentre, xd, yd);
#endif
	}
}

void PositionMap::draw_axis()
{
	int xcentre, ycentre, xrad, yrad;
	int xleft, xright, ybottom, ytop;

	map_to_win(0., 0., xcentre, ycentre);
	map_to_win(_referencedist, xrad, yrad);
	map_to_win(_range_x1, _range_y1, xleft, ybottom);
	map_to_win(_range_x2, _range_y2, xright, ytop);
#ifdef HAVE_CAIRO
	_cr->set_line_width(2.);
	_cr->set_source_rgb(POSMAP_AXISCOLOUR);
	_cr->set_dash(_dashpat, 0.5);
	_cr->move_to(xleft, ycentre);
	_cr->line_to(xright, ycentre);
	_cr->move_to(xcentre, ybottom);
	_cr->line_to(xcentre, ytop);
	_cr->stroke();
	_cr->arc(xcentre, ycentre, xrad, 0, 2.*M_PI);
	_cr->stroke();
	_cr->unset_dash();
#else
	_gc->set_foreground(_axiscolour);
	_gc->set_line_attributes(2, Gdk::LINE_ON_OFF_DASH, Gdk::CAP_NOT_LAST,
	                         Gdk::JOIN_MITER);
	_window->draw_arc(_gc, false, xcentre - xrad, ycentre - yrad, xrad*2,
	                  yrad*2, 0, 360*64);
	_window->draw_line(_gc, xleft, ycentre, xright, ycentre);
	_window->draw_line(_gc, xcentre, ybottom, xcentre, ytop);
#endif
}

void PositionMap::on_realize()
{
	Gtk::DrawingArea::on_realize();
	_window = get_window();

#ifndef HAVE_CAIRO
	Glib::RefPtr<Gdk::Colormap> colourmap = get_default_colormap();

	_bgcolour = Gdk::Color("black");
	_aposcolour = Gdk::Color("white");
	_poscolour = Gdk::Color("yellow");
	_axiscolour = Gdk::Color("red");
	colourmap->alloc_color(_bgcolour);
	colourmap->alloc_color(_poscolour);
	colourmap->alloc_color(_aposcolour);
	colourmap->alloc_color(_axiscolour);

	_gc = Gdk::GC::create(_window);
	_postextlayout = create_pango_layout("");
#endif
}

bool PositionMap::on_expose_event(GdkEventExpose* e)
{
#ifdef HAVE_CAIRO
	_cr = _window->create_cairo_context();
	if (e) {
		// Only redraw what we have to
		_cr->rectangle(e->area.x, e->area.y, e->area.width, e->area.height);
		_cr->clip();
	}
	_cr->set_identity_matrix();	// Use device coordinates directly
	_cr->set_source_rgb(POSMAP_BGCOLOUR);
	_cr->paint();
#else
	if (e) {
		Gdk::Rectangle r(e->area.x, e->area.y, e->area.width, e->area.height);

		_gc->set_clip_rectangle(r);
	}
	_window->set_background(_bgcolour);
	_window->clear();
#endif
	draw_axis();
	for (size_t i = 0; i < _coords.size(); ++i)
		draw_pos(i);
	return true;
}

bool PositionMap::on_button_press_event(GdkEventButton* e)
{
	if ((e->button == 1 && !_rbutton) || (e->button == 3 && !_lbutton)
		|| (e->button == 2 && !(_rbutton || _lbutton))) {
		int x, y;

		// Find the point clicked on. The first one we find will be moved
		for (size_t i = 0; i < _coords.size(); ++i) {
			map_to_win(_coords[i].x, _coords[i].y, x, y);
			double x_dist = e->x - x;
			double y_dist = e->y - y;
			if (sqrt(x_dist*x_dist + y_dist*y_dist) <= _coords[i].radius) {
				_active = i;
				if (e->button == 1) {
					// Set flag for 2D positioning
					_lbutton = true;
				} else if (!_constrain_radius && e->button == 3) {
					// Set flag for height positioning
					_rbutton = true;
					_mouse_last_y = e->y;
				} else if (!_constrain_radius && e->button == 2) {
					// Reset height of pos
					_coords[i].z = 0.;
					_coords[i].radius = 10;
					_signal_pos_changed.emit(i);
				}
				_signal_pos_active.emit(i);
				queue_draw();
				break;
			}
		}
	}
	return Gtk::DrawingArea::on_button_press_event(e);
}

bool PositionMap::on_button_release_event(GdkEventButton* e)
{
	if (e->button == 1) {
		_lbutton = false;
		queue_draw();
	} else if (e->button == 3) {
		_rbutton = false;
		queue_draw();
	}
	return Gtk::DrawingArea::on_button_release_event(e);
}

bool PositionMap::on_motion_notify_event(GdkEventMotion* e)
{
	if (_lbutton) {
		// User is moving a pos around in the plane
		double x, y;

		map_from_win(e->x, e->y, x, y);
		x = _coords[_active].x = std::max(std::min(x, _range_x2), _range_x1);
		y = _coords[_active].y = std::max(std::min(y, _range_y2), _range_y1);
		const double z = _coords[_active].z;
		// Limit the radius to _max_radius
		const double length = sqrt(x*x + y*y + z*z);
		if (_constrain_radius || length > _max_radius) {
			double t = sqrt((_max_radius*_max_radius - z*z)/(x*x + y*y));
			// _max_radius*cos(asin(z/_max_radius))*sin(atan2(x, y));
			_coords[_active].x = x*t;
			//_max_radius*cos(atan2(x, y))*cos(asin(z/_max_radius));
			_coords[_active].y = y*t;
		}
		_signal_pos_changed.emit(_active);
		queue_draw();
	} else if (_rbutton) {
		// User is changing pos height
		double delta = (_mouse_last_y - e->y)/10.;
		double z = _coords[_active].z + delta;
		const double x = _coords[_active].x;
		const double y = _coords[_active].y;

		z = std::max(std::min(z, _max_z), -_max_z);
		// Limit the radius to _max_radius
		const double length = sqrt(x*x + y*y + z*z);
		if (length > _max_radius) {
			const double sign = z > 0. ? 1. : -1.;
			z = sqrt(_max_radius*_max_radius - x*x - y*y)*sign;
		}
		_coords[_active].z = z;
		_mouse_last_y = e->y;
		_signal_pos_changed.emit(_active);
		queue_draw();
	}
	return Gtk::DrawingArea::on_motion_notify_event(e);
}

void PositionMap::on_size_request(Gtk::Requisition* req)
{
	*req = Gtk::Requisition();
	req->height = 300;
	req->width = 300;
}

void PositionMap::on_size_allocate(Gtk::Allocation& alloc)
{
	set_allocation(alloc);

	if (_window)
		_window->move_resize(alloc.get_x(), alloc.get_y(), alloc.get_width(),
		                     alloc.get_height());
}

