/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _positionmap_h
#define _positionmap_h

#include "config.h"
#include <gdkmm/colormap.h>
#include <gtkmm/drawingarea.h>
#include <vector>

#ifdef HAVE_CAIRO
#define POSMAP_BGCOLOUR   0., 0., 0. // Black
#define POSMAP_POSCOLOUR  1., 1., 0. // Yellow
#define POSMAP_APOSCOLOUR 1., 1., 1. // White
#define POSMAP_AXISCOLOUR 1., 0., 0. // Red
#endif

class PositionMap : public Gtk::DrawingArea {
	struct Coordinate {
		double x, y, z;
		int radius;
		Coordinate() : x(0.), y(0.), z(0.)
		{
		}
	};

	Glib::RefPtr<Gdk::Window> _window;
#ifdef HAVE_CAIRO
	Cairo::RefPtr<Cairo::Context> _cr;
	std::valarray<double> _dashpat;
#else
	Gdk::Color _bgcolour, _poscolour, _aposcolour, _axiscolour;
	Glib::RefPtr<Gdk::GC> _gc;
	Glib::RefPtr<Pango::Layout> _postextlayout;
#endif
	std::vector<Coordinate> _coords;
	double _max_radius;
	double _range_x1, _range_x2;
	double _range_y1, _range_y2;
	double _max_z;
	double _referencedist;
	bool _lbutton, _rbutton;
	double _mouse_last_y;
	size_t _active;
	bool _constrain_radius;

	sigc::signal<void, size_t> _signal_pos_changed;
	sigc::signal<void, size_t> _signal_pos_active;

	void map_to_win(double x, double y, int& xd, int& yd)
	{
		const Gtk::Allocation& a = get_allocation();
		const double range_w = _range_x2 - _range_x1;
		const double range_h = _range_y2 - _range_y1;
		const double norm_x = (x - _range_x1)/range_w;
		const double norm_y = 1. - (y - _range_y1)/range_h;

		xd = int(double(a.get_width())*norm_x);
		yd = int(double(a.get_height())*norm_y);
	}

	void map_to_win(double radius, int& xrad, int& yrad)
	{
		const Gtk::Allocation& a = get_allocation();
		const double range_w = _range_x2 - _range_x1;
		const double range_h = _range_y2 - _range_y1;

		xrad = int(radius/range_w*double(a.get_width()));
		yrad = int(radius/range_h*double(a.get_height()));
	}

	void map_from_win(double x, double y, double& xr, double& yr)
	{
		const Gtk::Allocation& a = get_allocation();
		const double range_w = _range_x2 - _range_x1;
		const double range_h = _range_y2 - _range_y1;

		xr = double(x)/double(a.get_width())*range_w + _range_x1;
		yr = range_h*(1. - double(y)/double(a.get_height())) + _range_y1;
	}

	void draw_pos(size_t num);
	void draw_axis();
protected:
	virtual void on_realize();
	virtual bool on_expose_event(GdkEventExpose* e);
	virtual bool on_button_press_event(GdkEventButton* e);
	virtual bool on_button_release_event(GdkEventButton* e);
	virtual bool on_motion_notify_event(GdkEventMotion* e);
	virtual void on_size_request(Gtk::Requisition* req);
	virtual void on_size_allocate(Gtk::Allocation& alloc);
public:
	enum CircleStyle { Position, Reference };
	PositionMap();
	void set_constrain_radius(bool c = true);
	void set_num_pos(size_t num);
	void set_active_pos(size_t num);
	void set_max_radius(double max_radius);
	void set_max_height(double max_height);
	size_t get_active_pos()
	{
		return _active;
	}
	void get_pos(size_t num, double& p1, double& p2, double& p3);
	void get_posxyz(size_t num, double& p1, double& p2, double& p3);
	void set_pos(size_t num, double p1, double p2, double p3);
	void set_reference_dist(double d);
	void set_rect(double x1, double x2, double y1, double y2);
	sigc::signal<void, size_t> signal_pos_changed()
	{
		return _signal_pos_changed;
	}
	
	sigc::signal<void, size_t> signal_pos_active()
	{
		return _signal_pos_active;
	}
};

#endif

