/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "SIMDMath.hpp"
#include "config.h"

SIMDMathPtr SIMDMath::_instance;

void SIMDMath::cpuid(unsigned* caps)
{
	unsigned cap1;
	unsigned cap2;

	unsigned unused; // clobbered input registers must be defined as outputs
	asm volatile (
		"cpuid"
		: "=d" (cap1), "=a" (unused)
		: "a" (0x00000001)
		: "ebx", "ecx");
	asm volatile (
		"cpuid"
		: "=a" (cap2)
		: "a" (0x80000000) 
		: "ebx", "ecx", "edx");	
	if (cap2 > 0x80000000)
		asm volatile (
			"cpuid"
			: "=d" (cap2), "=a" (unused)
			: "a" (0x80000001)
			: "ebx", "ecx");
	else
		cap2 = 0;
	caps[0] = cap1;
	caps[1] = cap2;
}

SIMDMathPtr SIMDMath::Create()
{
	unsigned caps[2];
	
	enum Caps {	
		_MMXExt = 0x00400000,	// in caps[1]
		_MMX = 0x00800000,		// in caps[0]
		_SSE = 0x02000000,		// in caps[0]
		_SSE2 = 0x04000000,		// in caps[0]
		_3DNowExt = 0x40000000,	// in caps[1]
		_3DNow = 0x80000000		// in caps[1]
	};
	
	// Check if we've already constructed a derived class.
	if (_instance) return _instance;
	cpuid(caps); 
#ifdef COMPILE_SSE
	if (caps[0] & _SSE)
		_instance.reset(new SSEMath());
	else
#endif
		_instance.reset(new CMath());
	return _instance;
}

void CMath::VVAcc(float* out, float* a, unsigned num)
{
	for (unsigned i = 0; i < num; ++i)
		out[i] += a[i];
}

void CMath::VSMul(float* out, float* a, float b, unsigned num)
{
	for (unsigned i = 0; i < num; ++i)
		out[i] = b*a[i];
}

void CMath::VVMul(float* out, float* a, float* b, unsigned num)
{
	for (unsigned i = 0; i < num; ++i)
		out[i] = a[i]*b[i];
}

void CMath::VVMulAcc(float* out, float* a, float* b, unsigned num)
{
	for (unsigned i = 0; i < num; ++i)
		out[i] += a[i]*b[i];
}

void CMath::VSMulAcc(float* out, float* a, float b, unsigned num)
{
	for (unsigned i = 0; i < num; ++i)
		out[i] += b*a[i];
}

void CMath::VVDiv(float* out, float* a, float* b, unsigned num)
{
	for (unsigned i = 0; i < num; ++i)
		out[i] = a[i]/b[i];
}

void CMath::SpectrumMul(float* out, float* a, float* b, unsigned num)
{
	out[0] = a[0]*b[0];
	for (unsigned i = 1; i < num/2; ++i) {
		out[i] = a[i]*b[i] - a[num/2 + i]*b[num/2 + i];
		out[num/2 + i] = a[i]*b[num/2 + i] + a[num/2 + i]*b[i];
	}
	out[num/2] = a[num/2]*b[num/2];
}

void CMath::SpectrumMulAcc(float* out, float* a, float* b, unsigned num)
{
	out[0] += a[0]*b[0];
	for (unsigned i = 1; i < num/2; ++i) {
		out[i] += a[i]*b[i] - a[num/2 + i]*b[num/2 + i];
		out[num/2 + i] += a[i]*b[num/2 + i] + a[num/2 + i]*b[i];
	}
	out[num/2] += a[num/2]*b[num/2];
}

void CMath::MatrixMix(float** inputs, float** outputs, Matrix& m, unsigned num,
                      unsigned ooffset)
{
	for (unsigned n = 0; n < num; ++n) {
		for (unsigned r = 0; r < m.size1(); ++r) {
			float* coef = &m(r, 0);
			float acc = 0.f;

			// Iterate over channels.
			for (unsigned chan = 0; chan < m.size2(); ++chan)
				acc += inputs[chan][n]*coef[chan];
			outputs[r][n + ooffset] = acc;
		}
	}
}

