/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "SIMDMath.hpp"
#include <xmmintrin.h>

void SSEMath::VVAcc(float* out, float* a, unsigned num)
{
	for (unsigned n = 0; n < num; n += 4) {
		__m128 av = _mm_load_ps(out + n);
		__m128 bv = _mm_load_ps(a + n);
		_mm_store_ps(out + n, _mm_add_ps(av, bv));
	}
}

void SSEMath::VSMul(float* out, float* a, float b, unsigned num)
{
	__m128 bv = _mm_set1_ps(b);
	
	for (unsigned n = 0; n < num; n += 4) {
		__m128 av = _mm_load_ps(a + n);
		
		_mm_store_ps(out + n, _mm_mul_ps(av, bv));
	}
}

void SSEMath::VVMul(float* out, float* a, float* b, unsigned num)
{
	for (unsigned n = 0; n < num; n += 4) {
		__m128 av = _mm_load_ps(a + n);
		__m128 bv = _mm_load_ps(b + n);
	
		_mm_store_ps(out + n, _mm_mul_ps(av, bv));
	}
}

void SSEMath::VVMulAcc(float* out, float* a, float* b, unsigned num)
{
	for (unsigned n = 0; n < num; n += 4) {
		__m128 av = _mm_load_ps(a + n);
		__m128 bv = _mm_load_ps(b + n);
		__m128 outv = _mm_load_ps(out + n);
		
		_mm_store_ps(out + n, _mm_add_ps(outv, _mm_mul_ps(av, bv)));
	}
}

void SSEMath::VSMulAcc(float* out, float* a, float b, unsigned num)
{
	__m128 bv = _mm_set1_ps(b);
	
	for (unsigned n = 0; n < num; n += 4) {
		__m128 av = _mm_load_ps(a + n);
		__m128 outv = _mm_load_ps(out + n);
	
		_mm_store_ps(out + n, _mm_add_ps(outv, _mm_mul_ps(av, bv)));
	}
}

void SSEMath::VVDiv(float* out, float* a, float* b, unsigned num)
{
	for (unsigned n = 0; n < num; n += 4) {
		__m128 av = _mm_load_ps(a + n);
		__m128 bv = _mm_load_ps(b + n);
	
		_mm_store_ps(out + n, _mm_div_ps(av, bv));
	}
}

void SSEMath::SpectrumMul(float* out, float* a, float* b, unsigned num)
{
	// One standard complex multiplication
	__m128 ar = _mm_load_ps(a);
	__m128 ai = _mm_load_ps(a + num/2);
	__m128 br = _mm_load_ps(b);
	__m128 bi = _mm_load_ps(b + num/2);
	__m128 arbr = _mm_mul_ps(ar, br);
	__m128 aibi = _mm_mul_ps(ai, bi);
	__m128 cr = _mm_sub_ps(arbr, aibi);
	__m128 ci = _mm_add_ps(_mm_mul_ps(ar, bi), _mm_mul_ps(br, ai));

	// Overwrite the first values with special cases for DC and Nyquist
#if defined(__GNUC__) && __GNUC__ < 4
	// GCC 3.x has a bug in the handling of movss, which is fixed in 4.x.
	// We'll just use inline asm for GCC 3.x.
	asm (
		"movss %[arbr], %[cr]\n\t"
		"movss %[aibi], %[ci]\n\t"
		: [cr] "+x" (cr), [ci] "+x" (ci)
		: [arbr] "x" (arbr), [aibi] "x" (aibi)
	);
#else
	cr = _mm_move_ss(cr, arbr);
	ci = _mm_move_ss(ci, aibi);
#endif
	_mm_store_ps(out, cr);
	_mm_store_ps(out + num/2, ci);

	// Complex multiply the rest 
	for (unsigned n = 4; n < num/2; n += 4) {
		ar = _mm_load_ps(a + n);
		ai = _mm_load_ps(a + n + num/2);
		br = _mm_load_ps(b + n);
		bi = _mm_load_ps(b + n + num/2);
		cr = _mm_sub_ps(_mm_mul_ps(ar, br), _mm_mul_ps(ai, bi));
		ci = _mm_add_ps(_mm_mul_ps(ar, bi), _mm_mul_ps(br, ai));
		_mm_store_ps(out + n, cr);
		_mm_store_ps(out + n + num/2, ci);	
	}
}

void SSEMath::SpectrumMulAcc(float* out, float* a, float* b, unsigned num)
{
	// One standard complex multiplication
	__m128 ar = _mm_load_ps(a);
	__m128 ai = _mm_load_ps(a + num/2);
	__m128 br = _mm_load_ps(b);
	__m128 bi = _mm_load_ps(b + num/2);
	__m128 arbr = _mm_mul_ps(ar, br);
	__m128 aibi = _mm_mul_ps(ai, bi);
	__m128 cr = _mm_sub_ps(arbr, aibi);
	__m128 ci = _mm_add_ps(_mm_mul_ps(ar, bi), _mm_mul_ps(br, ai));

	// Overwrite the first values with special cases for DC and Nyquist
#if defined(__GNUC__) && __GNUC__ < 4
	asm (
		"movss %[arbr], %[cr]\n\t"
		"movss %[aibi], %[ci]\n\t"
		: [cr] "+x" (cr), [ci] "+x" (ci)
		: [arbr] "x" (arbr), [aibi] "x" (aibi)
	);
#else
	cr = _mm_move_ss(cr, arbr);
	ci = _mm_move_ss(ci, aibi);
#endif
	__m128 outr = _mm_load_ps(out);
	__m128 outi = _mm_load_ps(out + num/2);
	_mm_store_ps(out, _mm_add_ps(outr, cr));
	_mm_store_ps(out + num/2, _mm_add_ps(outi, ci));

	// Complex multiply the rest 
	for (unsigned n = 4; n < num/2; n += 4) {
		ar = _mm_load_ps(a + n);
		ai = _mm_load_ps(a + n + num/2);
		br = _mm_load_ps(b + n);
		bi = _mm_load_ps(b + n + num/2);
		cr = _mm_sub_ps(_mm_mul_ps(ar, br), _mm_mul_ps(ai, bi));
		ci = _mm_add_ps(_mm_mul_ps(ar, bi), _mm_mul_ps(br, ai));
		outr = _mm_load_ps(out + n);
		outi = _mm_load_ps(out + n + num/2);
		_mm_store_ps(out + n, _mm_add_ps(outr, cr));
		_mm_store_ps(out + n + num/2, _mm_add_ps(outi, ci));	
	}
}

void SSEMath::MatrixMix(float** inputs, float** outputs, Matrix& m,
                        unsigned num, unsigned ooffset)
{
	__m128 in, coef;

	for (unsigned row = 0; row < m.size1(); ++row) {
		// Do four samples at a time as far as we can
		for (unsigned n = 0; n < (num & ~3); n += 4) {
			float* coefs = &m(row, 0);
			__m128 acc = _mm_setzero_ps();
			
			// Multiply and accumulate rightwards in the matrix
			for (unsigned chan = 0; chan < m.size2(); ++chan) {
				in = _mm_load_ps(&inputs[chan][n]);
				coef = _mm_load1_ps(coefs++);
				acc = _mm_add_ps(acc, _mm_mul_ps(in, coef));
			}
			_mm_store_ps(&outputs[row][n + ooffset], acc);
		}
// This isn't needed yet
#if 0
		// Process last few samples separately
		for (unsigned n = (num & ~3); n < num; ++n) {
			float* coefs = &m(row, 0);
			__m128 acc = _mm_setzero_ps();
			
			for (unsigned chan = 0; chan < m.size2(); ++chan) {
				in = _mm_load_ss(&inputs[chan][n]);
				coef = _mm_load_ss(coefs++);
				acc = _mm_add_ss(acc, _mm_mul_ss(in, coef));
			}
			_mm_store_ss(&outputs[row][n + ooffset], acc);
		}
#endif
	}
}

