/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "SourceWindow.hpp"
#include <cstdio> //For sprintf()
#include <cmath>
#include <gtkmm/alignment.h>
#include <gtkmm/aspectframe.h>
#include <gtkmm/box.h>
#include <gtkmm/label.h>
#include <gtkmm/scrolledwindow.h>

SourceWindow::SourceWindow(AmbisonicsEncoderModel& enc)
	: _btn_pwave("Plane wave"), _sld_gain(-60., 12.), _enc(enc)
{
	set_title("Source positions");
	set_default_size(800, 600);
	set_border_width(5);
	// GUI consists of a Notebook with two tabs, we'll construct the first tab
	// first.
	Gtk::ScrolledWindow* scrollwindow = Gtk::manage(new Gtk::ScrolledWindow());
	scrollwindow->set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	Gtk::Label* la1 = Gtk::manage(new Gtk::Label("Azimuth"));
	Gtk::Label* la2 = Gtk::manage(new Gtk::Label("Elevation"));
	Gtk::Label* la3 = Gtk::manage(new Gtk::Label("Radius"));
	Gtk::Label* la4 = Gtk::manage(new Gtk::Label("Gain"));

	_table.set_row_spacings(5);
	_table.set_col_spacings(10);
	_table.set_border_width(10);
	// Set up space for the first row (captions)
	_table.resize(1, 6);
	_table.attach(*la1, 1, 2, 0, 1,
	              Gtk::FILL | Gtk::EXPAND, Gtk::SHRINK);
	_table.attach(*la2, 2, 3, 0, 1,
	              Gtk::FILL | Gtk::EXPAND, Gtk::SHRINK);
	_table.attach(*la3, 3, 4, 0, 1, 
	              Gtk::FILL | Gtk::EXPAND, Gtk::SHRINK);
	_table.attach(*la4, 4, 5, 0, 1, 
	              Gtk::FILL | Gtk::EXPAND, Gtk::SHRINK);
	scrollwindow->add(_table);
	
	// Now create the second tab.
	// Create all the containers
	Gtk::Table* table1 = Gtk::manage(new Gtk::Table());
	Gtk::HBox* hbox1 = Gtk::manage(new Gtk::HBox());
	Gtk::HBox* hbox2 = Gtk::manage(new Gtk::HBox());
	Gtk::HBox* hbox3 = Gtk::manage(new Gtk::HBox());
	Gtk::VBox* vbox1 = Gtk::manage(new Gtk::VBox());
	Gtk::VBox* padbox = Gtk::manage(new Gtk::VBox());
	Gtk::AspectFrame* aspectframe = Gtk::manage(
		new Gtk::AspectFrame("Source positions"));
	Gtk::Frame* frame1 = Gtk::manage(
		new Gtk::Frame("Source settings"));
	Gtk::Alignment* align1 = Gtk::manage(
		new Gtk::Alignment(Gtk::ALIGN_LEFT, Gtk::ALIGN_CENTER, 0., 0.));

	// Create labels
	Gtk::Label* lb1 = Gtk::manage(new Gtk::Label("Azimuth:", Gtk::ALIGN_LEFT));
	Gtk::Label* lb2 = Gtk::manage(new Gtk::Label("Elevation:", Gtk::ALIGN_LEFT));
	Gtk::Label* lb3 = Gtk::manage(new Gtk::Label("Radius:", Gtk::ALIGN_LEFT));
	Gtk::Label* lb4 = Gtk::manage(new Gtk::Label("Scale:", Gtk::ALIGN_LEFT));
	Gtk::Label* lb5 = Gtk::manage(new Gtk::Label("Source:"));

	// Set up the graphical source positions box
	padbox->set_border_width(5);
	padbox->add(_posmap);
	aspectframe->add(*padbox);

	// Table containing the position controls for current source
	table1->set_border_width(10);
	table1->set_row_spacings(8);
	table1->set_col_spacings(10);
	table1->attach(*lb1, 0, 1, 0, 1);
	table1->attach(*lb2, 0, 1, 1, 2);
	table1->attach(*lb3, 0, 1, 2, 3);
	table1->attach(_btn_azimuth, 1, 2, 0, 1);
	table1->attach(_btn_elevation, 1, 2, 1, 2);
	table1->attach(_btn_radius, 1, 2, 2, 3);
	table1->attach(_btn_pwave, 1, 2, 3, 4);
	table1->attach(_sld_gain, 2, 3, 0, 4);
	table1->attach(_posxyz_label, 0, 2, 5, 6);
	// Put the table in a frame
	frame1->add(*table1);

	// Pack source number select and source map scale controls
	hbox2->pack_start(*lb5, Gtk::PACK_SHRINK, 10);
	hbox2->pack_start(_btn_source, Gtk::PACK_EXPAND_PADDING);
	hbox3->pack_start(*lb4, Gtk::PACK_SHRINK, 10);
	hbox3->pack_start(_sld_scale);

	// Keep the source controls in an alignment
	vbox1->pack_start(*hbox2, Gtk::PACK_SHRINK, 20);
	vbox1->pack_start(*frame1, Gtk::PACK_SHRINK);
	vbox1->pack_start(*hbox3, Gtk::PACK_SHRINK, 20);
	align1->add(*vbox1);

	// GUI is the graphical view on the left, and control widgets on the right
	hbox1->set_border_width(5);
	hbox1->pack_start(*aspectframe, Gtk::PACK_EXPAND_WIDGET, 10);
	hbox1->pack_start(*align1, Gtk::PACK_SHRINK);
	
	// Put it all in a Notebook.
	_tabs.append_page(*scrollwindow, "Controls");
	_tabs.append_page(*hbox1, "Graphical");
	add(_tabs);

	// Set up widgets	
	_btn_source.set_range(1., 1.);
	_btn_source.set_digits(0);
	_btn_source.set_increments(1., 3.);

	_btn_azimuth.set_range(-180., 180.);
	_btn_azimuth.set_digits(1);
	_btn_azimuth.set_increments(1., 15.);
	_btn_azimuth.set_wrap();

	_btn_elevation.set_range(-90., 90.);
	_btn_elevation.set_digits(1);
	_btn_elevation.set_increments(1., 10.);

	_btn_radius.set_range(0.1, 50.);
	_btn_radius.set_digits(2);
	_btn_radius.set_increments(0.1, 1.);

	_sld_scale.set_range(1., 50.);
	_sld_scale.set_increments(0.5, 5.);
	_sld_scale.set_digits(1);
	_sld_scale.set_value(20.);

	_posmap.set_rect(-20., 20., -20., 20.);
	_posmap.set_max_height(40.);
	_posmap.set_max_radius(enc.MaxRadius());

	_posmap.signal_pos_changed().connect(sigc::mem_fun(
		*this, &SourceWindow::on_map_pos_changed));
	_posmap.signal_pos_active().connect(sigc::mem_fun(
		*this, &SourceWindow::on_map_pos_active));
	_btn_source.signal_value_changed().connect(sigc::mem_fun(
		*this, &SourceWindow::on_active_source_changed));
	_btn_azimuth.signal_value_changed().connect(sigc::bind<size_t>(
		sigc::mem_fun(*this, &SourceWindow::on_azimuth_changed), 0));
	_btn_elevation.signal_value_changed().connect(sigc::bind<size_t>(
		sigc::mem_fun(*this, &SourceWindow::on_elevation_changed), 0));
	_btn_radius.signal_value_changed().connect(sigc::bind<size_t>(
		sigc::mem_fun(*this, &SourceWindow::on_radius_changed), 0));
	_btn_pwave.signal_toggled().connect(sigc::bind<size_t>(
		sigc::mem_fun(*this, &SourceWindow::on_pwave_toggled), 0));
	_sld_gain.signal_value_changed().connect(sigc::bind<size_t>(
		sigc::mem_fun(*this, &SourceWindow::on_gain_changed), 0));
	_sld_scale.signal_value_changed().connect(sigc::mem_fun(
		*this, &SourceWindow::on_scale_changed));
	_tabs.signal_switch_page().connect(sigc::mem_fun(
		*this, &SourceWindow::on_switch_tab));
	show_all_children();
	// Set our actual notification callback. We need to use GtkView since we
	// get notified from several threads (GUI and MIDI), which GTK does not
	// like.
	SetNotify(sigc::mem_fun(*this, &SourceWindow::notify));
}

/* Populate a table row with widgets */
void SourceWindow::CreateRow(size_t pos)
{
	char buf[16];
	
	snprintf(buf, sizeof(buf), "Source %zu:", pos + 1);
	Gtk::Label* label = Gtk::manage(new Gtk::Label(buf));
	Gtk::HScale* azimuth = Gtk::manage(new Gtk::HScale());
	Gtk::HScale* elev = Gtk::manage(new Gtk::HScale());
	Gtk::HScale* radius = Gtk::manage(new Gtk::HScale());
	Gtk::HScale* gain = Gtk::manage(new Gtk::HScale());
	Gtk::CheckButton* pwave = Gtk::manage(new Gtk::CheckButton("Plane wave"));
	
	azimuth->set_range(-180., 180.);
	azimuth->set_increments(1., 15.);
	azimuth->set_digits(1);
	elev->set_range(-90., 90.);
	elev->set_increments(0.5, 10.);
	elev->set_digits(1);
	radius->set_range(0.1, 50.);
	radius->set_increments(0.1, 1);
	radius->set_digits(2);
	gain->set_range(-60., 12.);
	gain->set_increments(1., 6.);
	gain->set_digits(1);

	_table.attach(*label, 0, 1, pos + 1, pos + 2,
	              Gtk::FILL, Gtk::SHRINK);
	_table.attach(*azimuth, 1, 2, pos + 1, pos + 2,
	              Gtk::FILL | Gtk::EXPAND, Gtk::SHRINK);
	_table.attach(*elev, 2, 3, pos + 1, pos + 2,
	              Gtk::FILL | Gtk::EXPAND, Gtk::SHRINK);
	_table.attach(*radius, 3, 4, pos + 1, pos + 2,
	              Gtk::FILL | Gtk::EXPAND, Gtk::SHRINK);
	_table.attach(*gain, 4, 5, pos + 1, pos + 2,
	              Gtk::FILL | Gtk::EXPAND, Gtk::SHRINK);
	_table.attach(*pwave, 5, 6, pos + 1, pos + 2,
	              Gtk::FILL, Gtk::SHRINK);
	azimuth->signal_value_changed().connect(sigc::bind<size_t>(
		sigc::mem_fun(*this, &SourceWindow::on_azimuth_changed), pos));
	elev->signal_value_changed().connect(sigc::bind<size_t>(
		sigc::mem_fun(*this, &SourceWindow::on_elevation_changed), pos));
	radius->signal_value_changed().connect(sigc::bind<size_t>(
		sigc::mem_fun(*this, &SourceWindow::on_radius_changed), pos));
	gain->signal_value_changed().connect(sigc::bind<size_t>(
		sigc::mem_fun(*this, &SourceWindow::on_gain_changed), pos));
	pwave->signal_toggled().connect(sigc::bind<size_t>(
		sigc::mem_fun(*this, &SourceWindow::on_pwave_toggled), pos));
	_azimuthsliders[pos] = azimuth;
	_elevationsliders[pos] = elev;
	_radiussliders[pos] = radius;
	_gainsliders[pos] = gain;
	_pwavebuttons[pos] = pwave;
}

// Set up the table structure needed for a give set of speaker positions.
// We allow a table to grow as big as it needs to be, then just hide the
// parts of it that are uneeded later. This saves us having to worry about
// deleting and recreating the widgets all the time.
void SourceWindow::MakeRows(size_t num)
{
	guint table_h = _table.property_n_rows().get_value() - 1;
	Gtk::Table_Helpers::TableList::iterator i;

	if (num > table_h) {
		_table.resize(num + 1, 6);
		_azimuthsliders.resize(num);
		_elevationsliders.resize(num);
		_radiussliders.resize(num);
		_gainsliders.resize(num);
		_pwavebuttons.resize(num);
		for(size_t y = table_h; y < num; y++)
			CreateRow(y);
	}
	
	for (i = _table.children().begin(); i != _table.children().end(); ++i) {
		if (i->get_top_attach() < num + 1)
			i->get_widget()->show();
		else
			i->get_widget()->hide();
	}
	UpdateControls();
}

/* Update all text entry widgets with current values */
void SourceWindow::UpdateControls()
{
	if (_tabs.get_current_page() == 0) {
		for (size_t i = 0; i < _enc.NumSources(); ++i) {
			_azimuthsliders[i]->set_value(_enc.SourceAzimuth(i));
			_elevationsliders[i]->set_value(_enc.SourceElevation(i));
			// Set elevation slider unsensitive if 2D Ambisonics
			_elevationsliders[i]->set_sensitive(_enc.Is3D());
			_radiussliders[i]->set_value(_enc.SourceRadius(i));
			// Set radius slider unsensitive if plane wave source
			_radiussliders[i]->set_sensitive(!_enc.PlaneWave(i));
			_gainsliders[i]->set_value(_enc.SourceGain(i));
			_pwavebuttons[i]->set_active(_enc.PlaneWave(i));
			// NOTE: There's a bug in GTK+ that prevents the Scale numeric value
			// to be redrawn unless the graphical representation of the slider
			// itself doesn't move. We need the value to always be redrawn
			// thanks to our MIDI support.
			//_azimuthsliders[i]->queue_draw();
			//_elevationsliders[i]->queue_draw();
			//_radiussliders[i]->queue_draw();
		}
	} else if (_tabs.get_current_page() == 1) {
		size_t num = _btn_source.get_value_as_int() - 1;

		// Azimuth is negated compared to PositionMap thanks to Ambisonics
		// convention being a counter-clockwise azimuth
		for (size_t i = 0; i < _enc.NumSources(); ++i)
			_posmap.set_pos(i, -_enc.SourceAzimuth(i), _enc.SourceElevation(i),
			                _enc.SourceRadius(i));
		_posmap.queue_draw();
		_btn_azimuth.set_value(_enc.SourceAzimuth(num));
		_btn_elevation.set_value(_enc.SourceElevation(num));
		_btn_elevation.set_sensitive(_enc.Is3D());
		_btn_radius.set_value(_enc.SourceRadius(num));
		_sld_gain.set_value(_enc.SourceGain(num));
		_btn_pwave.set_active(_enc.PlaneWave(num));
	}
}

void SourceWindow::notify()
{
	switch (_viewparameter) {
	case AmbisonicsEncoderModel::ChangedNumSources:
		_posmap.set_num_pos(_enc.NumSources());
		_btn_source.set_range(1., _enc.NumSources());
		MakeRows(_enc.NumSources());
		// NOTE: fall through to redraw
	case AmbisonicsEncoderModel::ChangedSourcePosition:
	case AmbisonicsEncoderModel::ChangedSourceGain:
	case AmbisonicsEncoderModel::ChangedPlaneWave:
	case AmbisonicsEncoderModel::ChangedIs3D:
		UpdateControls();
		break;
	case AmbisonicsEncoderModel::ChangedLoudspeakerRadius:
		_posmap.set_reference_dist(_enc.LoudspeakerRadius());
		break;
	case AmbisonicsEncoderModel::ChangedAll:
		_posmap.set_num_pos(_enc.NumSources());
		_btn_source.set_range(1., _enc.NumSources());
		MakeRows(_enc.NumSources());
		_posmap.set_reference_dist(_enc.LoudspeakerRadius());
		UpdateControls();
		break;
	}
}

void SourceWindow::on_azimuth_changed(size_t i)
{
	if (_tabs.get_current_page() == 0) {
		_enc.SourceAzimuth(i, _azimuthsliders[i]->get_value());
	} else if (_tabs.get_current_page() == 1) {
		size_t num = _btn_source.get_value_as_int() - 1;

		_enc.SourceAzimuth(num, _btn_azimuth.get_value());
	}
}

void SourceWindow::on_elevation_changed(size_t i)
{
	if (_tabs.get_current_page() == 0) {
		_enc.SourceElevation(i, _elevationsliders[i]->get_value());
	} else if (_tabs.get_current_page() == 1) {
		size_t num = _btn_source.get_value_as_int() - 1;

		_enc.SourceElevation(num, _btn_elevation.get_value());
	}
}

void SourceWindow::on_radius_changed(size_t i)
{
	if (_tabs.get_current_page() == 0) {
		_enc.SourceRadius(i, _radiussliders[i]->get_value());
	} else if (_tabs.get_current_page() == 1) {
		size_t num = _btn_source.get_value_as_int() - 1;

		_enc.SourceRadius(num, _btn_radius.get_value());
	}
}

void SourceWindow::on_pwave_toggled(size_t i)
{
	if (_tabs.get_current_page() == 0) {		
		_enc.PlaneWave(i, _pwavebuttons[i]->get_active());
	} else if (_tabs.get_current_page() == 1) {
		size_t num = _btn_source.get_value_as_int() - 1;

		_enc.PlaneWave(num, _btn_pwave.get_active());
	}
}

void SourceWindow::on_gain_changed(size_t i)
{
	if (_tabs.get_current_page() == 0) {
		_enc.SourceGain(i, _gainsliders[i]->get_value());
	} else if (_tabs.get_current_page() == 1) {
		size_t num = _btn_source.get_value_as_int() - 1;

		_enc.SourceGain(num, _sld_gain.get_value());
	}
}

void SourceWindow::update_xyz_label(size_t num)
{
	double x, y, z;
	char buf[128];

	_posmap.get_posxyz(num, x, y, z);
	// Ambisonics customarily uses weird X and Y axis directions
	snprintf(buf, sizeof(buf), "X: %.1f, Y: %.1f, Z: %.1f", y, -x, z);
	_posxyz_label.set_text(buf);
}

void SourceWindow::on_map_pos_active(size_t num)
{
	_btn_source.set_value(num + 1);
}

void SourceWindow::on_map_pos_changed(size_t num)
{
	double a, b, c;

	if (num != _btn_source.get_value() - 1)
		return;
	_posmap.get_pos(num, a, b, c);
	_btn_azimuth.set_value(-a);
	_btn_elevation.set_value(b);
	_btn_radius.set_value(c);
	update_xyz_label(num);
}

void SourceWindow::on_active_source_changed()
{
	const size_t num = _btn_source.get_value_as_int() - 1;

	_posmap.set_active_pos(num);
	UpdateControls();
	update_xyz_label(num);
}

void SourceWindow::on_scale_changed()
{
	double val = _sld_scale.get_value();

	_posmap.set_rect(-val, val, -val, val);
}

void SourceWindow::on_switch_tab(GtkNotebookPage* page, guint num)
{
	UpdateControls();
}

