/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "SpeakerWindow.hpp"
#include <cstdio> //For sprintf()
#include <cmath>
#include <gtkmm/aspectframe.h>
#include <gtkmm/box.h>
#include <gtkmm/radiobutton.h>

SpeakerWindow::SpeakerWindow(AmbisonicsDecoderModel& dec)
	: _azimuthwidgets(1), _elevationwidgets(1),
	  _btn_custom("Custom"), _btn_circle("Circle"),
	  _adj_azimuth(0., -180., 180., 1., 10.),
	  _btn_azimuth(_adj_azimuth, 0., 1), _dec(dec)
{
	// Init window
	set_title("Loudspeaker layout");
	set_default_size(500, 500);
	// Construct the following structure:
	// 1. VBox containing a Table and a ScrolledWindow
	// 2. Table will contain a RadioButton group, two Labels and a SpinButton
	// 3. Scrolledwindow contains another Table
	// 4. Table will contain widgets constructed on demand later
	Gtk::VBox* vbox = Gtk::manage(new Gtk::VBox());
	Gtk::Table* toptable = Gtk::manage(new Gtk::Table(1, 3));

	// Build the top table
	toptable->set_col_spacings(20);
	toptable->set_border_width(16);

	Gtk::Label* layout_label = Gtk::manage(new Gtk::Label());
	layout_label->set_markup("<b>Speaker layout</b>");
	layout_label->set_alignment(Gtk::ALIGN_LEFT);
	Gtk::RadioButton::Group group = _btn_custom.get_group();
	_btn_circle.set_group(group);
	_btn_circle.set_active();

	Gtk::VBox* vbox2 = Gtk::manage(new Gtk::VBox());
	vbox2->pack_start(*layout_label, false, true, 5);
	vbox2->pack_start(_btn_custom, false, true);
	vbox2->pack_start(_btn_circle, false, true);
	toptable->attach(*vbox2, 0, 1, 0, 1);
	Gtk::Label* azimuth_label = Gtk::manage(
		new Gtk::Label("First loudspeaker azimuth"));
	toptable->attach(*azimuth_label, 1, 2, 0, 1);
	toptable->attach(_btn_azimuth, 2, 3, 0, 1);
	vbox->pack_start(*toptable, false, true, 20);
	// Build the ScrolledWindow
	_scrolledwindow.add(_table);
	_scrolledwindow.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	vbox->pack_start(_scrolledwindow);

	// Build the last table	
	_table.set_row_spacings(5);
	_table.set_col_spacings(20);
	_table.set_border_width(10);
	// Set up space for the first row (captions)
	_table.resize(1, 3);
	Gtk::Label* label1 = Gtk::manage(new Gtk::Label("Azimuth"));
	Gtk::Label* label2 = Gtk::manage(new Gtk::Label("Elevation"));
	_table.attach(*label1, 1, 2, 0, 1,
	              Gtk::FILL | Gtk::EXPAND, Gtk::SHRINK);
	_table.attach(*label2, 2, 3, 0, 1,
	              Gtk::FILL | Gtk::EXPAND, Gtk::SHRINK);

	_posmap.set_rect(-1.2, 1.2, -1.2, 1.2);
	_posmap.set_max_radius(1.);
	// Only allow speakers to be placed on the unit circle/sphere
	_posmap.set_constrain_radius();
	Gtk::AspectFrame* aspectframe = Gtk::manage(
		new Gtk::AspectFrame("Speaker positions"));
	Gtk::VBox* padbox = Gtk::manage(new Gtk::VBox());
	padbox->set_border_width(5);
	padbox->add(_posmap);
	aspectframe->add(*padbox);

	// Put it all in a Notebook.
	_tabs.append_page(*vbox, "Controls");
	_tabs.append_page(*aspectframe, "Graphical");
	add(_tabs);

	// Connect signals to signal handlers
	_btn_custom.signal_clicked().connect(sigc::mem_fun(
		*this, &SpeakerWindow::on_layout_clicked));
	_btn_circle.signal_clicked().connect(sigc::mem_fun(
		*this, &SpeakerWindow::on_layout_clicked));
	_btn_azimuth.signal_value_changed().connect(sigc::mem_fun(
		*this, &SpeakerWindow::on_firstazimuth_changed));
	_posmap.signal_pos_changed().connect(sigc::mem_fun(
		*this, &SpeakerWindow::on_map_pos_changed));
	_tabs.signal_switch_page().connect(sigc::mem_fun(
		*this, &SpeakerWindow::on_switch_tab));

	show_all_children();
}

/* Populate a table row with widgets */
void SpeakerWindow::CreateRow(size_t pos)
{
	char buf[64];

	// Construct and set up a Label and two SpinButtons
	snprintf(buf, sizeof(buf), "Speaker %zu:", pos + 1);
	Gtk::Label* label = Gtk::manage(new Gtk::Label(buf));
	Gtk::SpinButton* azimuth = Gtk::manage(new Gtk::SpinButton(0., 2));
	Gtk::SpinButton* elev = Gtk::manage(new Gtk::SpinButton(0., 2));
	azimuth->set_range(-360.0, 360.0);
	azimuth->set_increments(1, 10);
	azimuth->set_wrap();
	elev->set_range(-90.0, 90.0);
	elev->set_increments(1, 10);
	// Fix them in the Table
	_table.attach(*label, 0, 1, pos + 1, pos + 2,
	              Gtk::FILL, Gtk::SHRINK);
	_table.attach(*azimuth, 1, 2, pos + 1, pos + 2,
	              Gtk::FILL | Gtk::EXPAND, Gtk::SHRINK);
	_table.attach(*elev, 2, 3, pos + 1, pos + 2,
	              Gtk::FILL | Gtk::EXPAND, Gtk::SHRINK);

	// Connect signal handlers
	azimuth->signal_value_changed().connect(sigc::bind<size_t>(
		sigc::mem_fun(*this, &SpeakerWindow::on_azimuth_changed), pos));
	elev->signal_value_changed().connect(sigc::bind<size_t>(
		sigc::mem_fun(*this, &SpeakerWindow::on_elev_changed), pos));
	_azimuthwidgets[pos] = azimuth;
	_elevationwidgets[pos] = elev;
}

/* Set up the table structure needed for a given set of speaker positions.
 * We allow a table to grow as big as it needs to be, then just hide the
 * parts of it that are uneeded later. This saves us having to worry about
 * deleting and recreating the widgets all the time.
 */
void SpeakerWindow::CreateRows()
{
	size_t num = _dec.NumSpeakers();
	guint table_h = _table.property_n_rows().get_value() - 1;
	Gtk::Table_Helpers::TableList::iterator i;

	if (num > table_h) {
		_azimuthwidgets.resize(num);
		_elevationwidgets.resize(num);
		_table.resize(num + 1, 3);
		for(size_t y = table_h; y < num; y++)
			CreateRow(y);
	}

	for (i = _table.children().begin(); i != _table.children().end(); ++i) {
		if (i->get_top_attach() < num + 1)
			i->get_widget()->show();
		else
			i->get_widget()->hide();
	}
}

/* Update all widgets with current values */
void SpeakerWindow::UpdateControls()
{
	if (_tabs.get_current_page() == 0) {
		for (size_t i = 0; i < _dec.NumSpeakers(); ++i) {
			_azimuthwidgets[i]->set_value(_dec.SpeakerAzimuth(i));
			_elevationwidgets[i]->set_value(_dec.SpeakerElevation(i));
			// Make elevation controls unsensitive if 2D Ambisonics
			_elevationwidgets[i]->set_sensitive(_dec.Is3D());
		}
	} else if (_tabs.get_current_page() == 1) {
		for (size_t i = 0; i < _dec.NumSpeakers(); ++i) {
			_posmap.set_pos(i, -_dec.SpeakerAzimuth(i),
			                _dec.SpeakerElevation(i), 1.);
		}
	}
}

void SpeakerWindow::Notify(int what)
{
	switch (what) {
	case AmbisonicsDecoderModel::ChangedSpeakerLayout:
	{
		bool custom = _dec.SpeakerLayout()
		              == AmbisonicsDecoderModel::CustomLayout;

		// Enable or disable azimuth button and speaker position boxes depending
		// depending on layout buttons status
		if (custom || _dec.Is3D())
			_btn_custom.set_active();
		else
			_btn_circle.set_active();
		_btn_azimuth.set_sensitive(!custom);
		_table.set_sensitive(custom);
		_posmap.set_sensitive(custom);
		UpdateControls();
		break;
	}
	case AmbisonicsDecoderModel::ChangedFirstSpeakerAzimuth:
		_btn_azimuth.set_value(_dec.FirstSpeakerAzimuth());
		UpdateControls();
		break;
	case AmbisonicsDecoderModel::ChangedNumSpeakers:
		// Add/remove components from the widget table
		CreateRows();
		_posmap.set_num_pos(_dec.NumSpeakers());
		// NOTE: fall through to redraw here
	case AmbisonicsDecoderModel::ChangedSpeakerPosition:
		UpdateControls();
		break;
	case AmbisonicsDecoderModel::ChangedIs3D:
		// Deactiveate and disable circle layout button if 3D mode. It doesn't
		// currently make sense.
		if (_dec.Is3D())
			_dec.SpeakerLayout(AmbisonicsDecoderModel::CustomLayout);
		_btn_circle.set_sensitive(!_dec.Is3D());
		UpdateControls();
		break;
	case AmbisonicsDecoderModel::ChangedAll:
	{
		bool custom = _dec.SpeakerLayout()
		              == AmbisonicsDecoderModel::CustomLayout;

		if (custom || _dec.Is3D())
			_btn_custom.set_active();
		else
			_btn_circle.set_active();
		_btn_circle.set_sensitive(!_dec.Is3D());
		_btn_azimuth.set_sensitive(!custom);
		_table.set_sensitive(custom);
		_posmap.set_sensitive(custom);
		_btn_azimuth.set_value(_dec.FirstSpeakerAzimuth());
		CreateRows();
		_posmap.set_num_pos(_dec.NumSpeakers());
		UpdateControls();
		break;
	}
	};
}

void SpeakerWindow::on_azimuth_changed(size_t i)
{
	Gtk::SpinButton* e = _azimuthwidgets[i];
	
	_dec.SpeakerAzimuth(i, e->get_value());
}

void SpeakerWindow::on_elev_changed(size_t i)
{
	Gtk::SpinButton* e = _elevationwidgets[i];
	
	_dec.SpeakerElevation(i, e->get_value());
}

void SpeakerWindow::on_layout_clicked()
{
	AmbisonicsDecoderModel::SpeakerLayoutPreset layout;

	if (_btn_custom.get_active())
		layout = AmbisonicsDecoderModel::CustomLayout;
	else 
		layout = AmbisonicsDecoderModel::CircleLayout;
	_dec.SpeakerLayout(layout);
}

void SpeakerWindow::on_firstazimuth_changed()
{
	_dec.FirstSpeakerAzimuth(_btn_azimuth.get_value());
}

void SpeakerWindow::on_map_pos_changed(size_t num)
{
	double a, e, r;

	_posmap.get_pos(num, a, e, r);
	_dec.SpeakerAzimuth(num, -a);
	_dec.SpeakerElevation(num, e);
}

void SpeakerWindow::on_switch_tab(GtkNotebookPage* page, guint num)
{
	UpdateControls();
}
