/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _view_h
#define _view_h

#include <sigc++/class_slot.h>
#include <glibmm.h>

/**
 * \brief Used to present or process information updates from a Model class.
 */
class View {
public:
	/**
	 * Is called every time a Model object has updated it's state.
	 * @param what describes what part of the Model state changed. Value is Model
	 * dependent.
	 */
	virtual void Notify(int what) = 0;
	virtual ~View()
	{
	}
};

/**
 * \brief View class for use when several threads need to update one single
 * GTK view.
 *
 * Instead of overriding the Notify method, derived classes should instead
 * register their own notifiy method by using the SetNotify method.
 * This stems from the fact that the notification is done through a pipe to
 * make sure that the GUI thread is always the thread that gets the
 * notification.
 * The notification event type is stored in the protected member variable
 * _viewparameter.
 */
class GtkView : public View {
protected:
	Glib::Dispatcher _dispatcher;
	int _viewparameter;
public:
	void Notify(int what)
	{
		_viewparameter = what;
		_dispatcher.emit();
	}

	/**
	 * Sets the sigc++ slot to which notifications will be delivered.
	 * @param slot a sigc++ slot (like sigc::mem_fun).
	 */
	void SetNotify(const sigc::slot<void>& slot)
	{
		_dispatcher.connect(slot);
	}

	virtual ~GtkView()
	{
	}
};

#endif

