/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _widgets_hpp
#define _widgets_hpp

#include <string>
#include <gtkmm/comboboxentrytext.h>
#include <gtkmm/comboboxtext.h>
#include <gtkmm/dialog.h>
#include <gtkmm/scale.h>
#include <gtkmm/stock.h>
#include "Ambisonics.hpp"

class NormalizationComboBox : public Gtk::ComboBoxText {
	static const Glib::ustring _norm_strings[4];
	size_t _index;
public:
	NormalizationComboBox(size_t def = 0) : _index(def)
	{
		for (size_t i = 0; i < 4; ++i)
			append_text(_norm_strings[i]);
		set_active_text(_norm_strings[_index]);
	}

	Ambisonics::Normalization get_norm() const
	{
		return Ambisonics::Normalization(_index);
	}

	void set_norm(Ambisonics::Normalization n)
	{
		_index = size_t(n);
		set_active_text(_norm_strings[_index]);
	}

	virtual void on_changed()
	{
		for (_index = 0; _index < 4; ++_index) {
			if (_norm_strings[_index].compare(get_active_text()) == 0)
				break;
		}
	}

	virtual ~NormalizationComboBox()
	{
	}
};

/**
 * \brief Convenience class which creates a dialog which accepts a string of
 * input and also offers a some pre-defined options.
 */
class StringInputDialog : public Gtk::Dialog {
	Gtk::ComboBoxEntryText _combo;
public:
	/**
	 * Class constructor.
	 * @param title title of input dialog.
	 * @param strings pre-defined strings to fill the entry box with.
	 * @param num number of pre-defined strings.
	 */
	StringInputDialog(const char* title, const char** strings, int num)
		: Gtk::Dialog(title, true)
	{
		set_skip_taskbar_hint();
		// Add all the strings to the combobox
		for (int i = 0; i < num; ++i)
			_combo.append_text(strings[i]);
		// Make first string default
		if (num) _combo.set_active_text(strings[0]);
		// Add combo box to the dialog
		get_vbox()->pack_start(_combo);
		// Add buttons
		add_button(Gtk::Stock::OK, Gtk::RESPONSE_OK);
		add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
		set_resizable(false);
		show_all_children();
	}

	/**
	 * Shows dialog and block until user is finished.
	 * @returns True if OK pressed, false if exited in other ways.
	 */
	bool Run()
	{
		if (run() == Gtk::RESPONSE_OK)
			return true;
		else
			return false;
	}
	
	/**
	 * Fetches string displayed in input box.
	 * @returns String currently in input box.
	 */
	std::string GetString()
	{
		return _combo.get_entry()->get_text();
	}
};

class GainSlider : public Gtk::VScale {
	double _default;
public:
	GainSlider(double min, double max, double def = 0.) : _default(def)
	{
		set_range(min, max);
		set_inverted();
		set_value_pos(Gtk::POS_BOTTOM);
		set_increments(1., 6.);
	}

	virtual Glib::ustring on_format_value(double value)
	{
		char buf[64];

		snprintf(buf, sizeof(buf), "%.1f dB", value);
		return buf;
	}

	virtual bool on_button_press_event(GdkEventButton* e)
	{
		if (e->button == 2) {
			// If middle button is pressed, reset gain to default
			set_value(_default);
			return true;
		} else {
			// If not, let parent handle it
			return Gtk::VScale::on_button_press_event(e);
		}
	}

	virtual void on_size_request(Gtk::Requisition* req)
	{
		// GTK has a tendency to ignore the fact that the label might need more
		// space than the default widget width, so we hijack the size request
		// and force some more width to be allocated.
		Gtk::VScale::on_size_request(req);
		req->width += 40;
	}

	virtual ~GainSlider()
	{
	}
};

#endif

