// Allocator that wraps operator new -*- C++ -*-

// Copyright (C) 2001, 2002, 2003, 2004 Free Software Foundation, Inc.
//
// This file is part of the GNU ISO C++ Library.  This library is free
// software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the
// Free Software Foundation; either version 2, or (at your option)
// any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License along
// with this library; see the file COPYING.  If not, write to the Free
// Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.

// As a special exception, you may use this file as part of a free software
// library without restriction.  Specifically, if other files instantiate
// templates or use macros or inline functions from this file, or you compile
// this file and link it with other files to produce an executable, this
// file does not by itself cause the resulting executable to be covered by
// the GNU General Public License.  This exception does not however
// invalidate any other reasons why the executable file might be covered by
// the GNU General Public License.

// modified for alignment by 16 by Thom Johansen.

#ifndef _ALIGN16_ALLOCATOR_H
#define _ALIGN16_ALLOCATOR_H

#include <new>

  /**
   *  @brief  An allocator that uses global new, as per [20.4], to allocate a buffer aligned with 16.
   *
   *  This is precisely the allocator defined in the C++ Standard. 
   *    - all allocation calls operator new
   *    - all deallocation calls operator delete
   *
   *  (See @link Allocators allocators info @endlink for more.)
   */
  template<typename _Tp>
    class align16_allocator
    {
    public:
      typedef size_t     size_type;
      typedef ptrdiff_t  difference_type;
      typedef _Tp*       pointer;
      typedef const _Tp* const_pointer;
      typedef _Tp&       reference;
      typedef const _Tp& const_reference;
      typedef _Tp        value_type;

      template<typename _Tp1>
        struct rebind
        { typedef align16_allocator<_Tp1> other; };

      align16_allocator() throw() { }

      align16_allocator(const align16_allocator&) throw() { }

      template<typename _Tp1>
        align16_allocator(const align16_allocator<_Tp1>&) throw() { }

      ~align16_allocator() throw() { }

      pointer
      address(reference __x) const { return &__x; }

      const_pointer
      address(const_reference __x) const { return &__x; }

      // NB: __n is permitted to be 0.  The C++ standard says nothing
      // about what the return value is when __n == 0.
      // We allocate 15 + sizeof(pointer) bytes extra, then align the pointer
      // to 16, and save the original pointer returned from new() in the
      // address below the new pointer. deallocate() then accesses this value
      // when it comes time to delete() the buffer.
      pointer
      allocate(size_type __n, const void* = 0)
      { 
        _Tp* p = static_cast<_Tp*>(::operator new(__n*sizeof(_Tp) + 15 + sizeof(pointer)));
        size_t t = ((reinterpret_cast<size_t>(p) + 15 + sizeof(pointer)) & ~15);
        reinterpret_cast<_Tp**>(t)[-1] = p;
        _Tp* newp = reinterpret_cast<_Tp*>(t);
        return newp;
      }

      // __p is not permitted to be a null pointer.
      void
      deallocate(pointer __p, size_type)
      { 
        ::operator delete(reinterpret_cast<_Tp**>(__p)[-1]);
      }

      size_type
      max_size() const throw() 
      { return size_t(-1) / sizeof(_Tp); }

      // _GLIBCXX_RESOLVE_LIB_DEFECTS
      // 402. wrong new expression in [some_] allocator::construct
      void 
      construct(pointer __p, const _Tp& __val) 
      { ::new(__p) _Tp(__val); }

      void 
      destroy(pointer __p) { __p->~_Tp(); }
    };

  template<typename _Tp>
    inline bool
    operator==(const align16_allocator<_Tp>&, const align16_allocator<_Tp>&)
    { return true; }
  
  template<typename _Tp>
    inline bool
    operator!=(const align16_allocator<_Tp>&, const align16_allocator<_Tp>&)
    { return false; }

#endif
