/* 
 * This file is part of ambitools.
 * Copyright (C) 2006-2007 Centre for Quantifiable Quality of Service in
 * Communication Systems, Norwegian University of Science and Technology.
 *
 * ambitools is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ambitools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ambitools; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _simdmath_h
#define _simdmath_h

#include "config.h"
#include <boost/shared_ptr.hpp>
#include "Matrix.hpp"

class SIMDMath;

typedef boost::shared_ptr<SIMDMath> SIMDMathPtr;

/**
 * \brief Class to perform vector math operations that are easily optimisable
 * by SIMD instructions where these are available.
 *
 * All input arrays should have
 * addresses aligned by 16, since some SIMD instruction sets do not support
 * (fast) accesses to memory that does not satisfy this constraint.
 * Only arrays that have sizes that are both a power of two and greater or equal
 * than four are supported.
 * Class is a singleton and needs to be constructed using Create().
 */
class SIMDMath {
public:
	enum { None = 0x00, Default = 0x01, ThreeDNow = 0x02, SSE = 0x04};
private:
	static SIMDMathPtr _instance;
	static void cpuid(unsigned* caps);
public:
	/**
	 * Detects what SIMD instruction sets, if any, the current CPU has, and
	 * returns a pointer to a class taking advantage of that instruction set.
	 * If none are found, the default class using ordinary C math routines
	 * is returned.
	 * @return A SIMDMathPtr (reference counting pointer) to a class best suited
	 * to exploiting the current CPU. 
	 */
	static SIMDMathPtr Create();
	/**
	 * Method to be overriden by any inheriting classes.
	 * @return Value describing what kind of SIMD instruction set is being used.
	 */
	virtual unsigned GetMethod() = 0;
	/**
	 * Accumulates one vector in another vector.
	 * @param out result vector.
	 * @param a vector to be added to output vector.
	 * @param num number of samples in vectors.
	 */
	virtual void VVAcc(float* out, float* a, unsigned num) = 0;
	/**
	 * Multiply a vector by a scalar and save result in result vector.
	 * @param out result vector.
	 * @param a vector to be multiplied by a scalar.
	 * @param b multiplicative constant.
	 * @param num number of samples in vectors.
	 */
	virtual void VSMul(float* out, float* a, float b, unsigned num) = 0;
	/**
	 * Multiply a vector by a vector, element by element,  and save result in
	 * result vector.
	 * @param out result vector.
	 * @param a first vector to be multiplied.
	 * @param b second vector to be multiplied.
	 * @param num number of samples in vectors.
	 */
	virtual void VVMul(float* out, float* a, float* b, unsigned num) = 0;
	/**
	 * Multiply a vector by a scalar and accumulate result to output vector.
	 * @param out result vector.
	 * @param a vector to be multiplied by a scalar.
	 * @param b multiplicative constant.
	 * @param num number of samples in vectors.
	 */
	virtual void VSMulAcc(float* out, float* a, float b, unsigned num) = 0;
	/**
	 * Multiply a vector by another vector, element by element, and accumulate
	 * result to output vector.
	 * @param out result vector.
	 * @param a first vector to be multiplied.
	 * @param b second vector to be multiplied.
	 * @param num number of samples in vectors.
	 */
	virtual void VVMulAcc(float* out, float* a, float* b, unsigned num) = 0;
	/**
	 * Divide a vector by another vector and save result in result vector.
	 * @param out result vector.
	 * @param a dividend vector.
	 * @param b divisor vector.
	 * @param num number of samples in vectors.
	 */
	virtual void VVDiv(float* out, float* a, float* b, unsigned num) = 0;
	/**
	 * Multiply two frequency spectrums.
	 * @param out result vector.
	 * @param a first frequency spectrum.
	 * @param b second frequency spectrum.
	 * @param num number of elements in spectrum data.
	 */
	virtual void SpectrumMul(float* out, float* a, float* b, unsigned num) = 0;
	/**
	 * Multiply two frequency spectrums and accumulate result to output.
	 * @param out result vector.
	 * @param a first frequency spectrum.
	 * @param b second frequency spectrum.
	 * @param num number of elements in spectrum data.
	 */
	virtual void SpectrumMulAcc(float* out, float* a, float* b, unsigned num) = 0;
	/**
	 * Mix a set of inputs with mixing factors decided by a coefficient matrix.
	 * A good way to visualise how it works is to see inputs as connected at the
	 * top of the matrix, and outputs connected at the right. The mixer will mix
	 * inputs according to mixing factors specified in the matrix and output the
	 * result to the respective output.
	 * @param inputs array of pointers to the input buffers. Must contain as
	 * many pointers as there are columns in the coefficient matrix.
	 * @param outputs array of pointers to the output buffers. Must contain as
	 * many pointers as there are rows in the coefficient matrix.
	 * @param m coefficient matrix deciding the mixing factors.
	 * @param num number of samples to mix.
	 * @param ooffset offset in the output arrays from which to start
	 * processing. Defaults to 0.
	 */
	virtual void MatrixMix(float** inputs, float** outputs, Matrix& m,
	                       unsigned num, unsigned ooffset = 0) = 0;
	/**
	 * Class destructor.
	 */
	virtual ~SIMDMath()
	{
	}
};

/**
 * \brief Default SIMDMath derived class.
 *
 * Based on standard C math routines that will work everywhere.
 * @see SIMDMath
 */
class CMath : public SIMDMath {
public:
	CMath()
	{
	}
	virtual unsigned GetMethod()
	{
		return Default;
	}
	virtual void VVAcc(float* out, float* a, unsigned num);
	virtual void VSMul(float* out, float* a, float b, unsigned num);
	virtual void VVMul(float* out, float* a, float* b, unsigned num);
	virtual void VVMulAcc(float* out, float* a, float* b, unsigned num);
	virtual void VSMulAcc(float* out, float* a, float b, unsigned num);
	virtual void VVDiv(float* out, float* a, float* b, unsigned num);
	virtual void SpectrumMul(float* out, float* a, float* b, unsigned num);
	virtual void SpectrumMulAcc(float* out, float* a, float* b, unsigned num);
	virtual void MatrixMix(float** inputs, float** outputs, Matrix& m,
	                       unsigned num, unsigned ooffset = 0);
	virtual ~CMath()
	{
	}
};

#ifdef COMPILE_SSE
/**
 * \brief SIMDMath derived class that uses the SSE instruction set available on
 * processors from Pentium 3/Athlon XP and on.
 *
 * Uses SSE intrinsics for implementation, so no inline assembly is used.
 * IMPORTANT: all arrays to be used here must have addresses aligned by 16.
 * This restriction can be loosened, but we have no need for it as of yet.
 * @see SIMDMath
 */
class SSEMath : public SIMDMath {
public:
	SSEMath()
	{
	}
	virtual unsigned GetMethod()
	{
		return SSE;
	}
	virtual void VVAcc(float* out, float* a, unsigned num);
	virtual void VSMul(float* out, float* a, float b, unsigned num);
	virtual void VVMul(float* out, float* a, float* b, unsigned num);
	virtual void VSMulAcc(float* out, float* a, float b, unsigned num);
	virtual void VVMulAcc(float* out, float* a, float* b, unsigned num);
	virtual void VVDiv(float* out, float* a, float* b, unsigned num);
	virtual void SpectrumMul(float* out, float* a, float* b, unsigned num);
	virtual void SpectrumMulAcc(float* out, float* a, float* b, unsigned num);
	virtual void MatrixMix(float** inputs, float** outputs, Matrix& m,
	                       unsigned num, unsigned ooffset = 0);
	virtual ~SSEMath()
	{
	}
};
#endif // COMPILE_SSE

#endif
